// File: sgfpars.h

//   Copyright (C) 2001-4 Ulrich Goertz (u@g0ertz.de)

//   This is part of Kombilo, a go database program.

//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.

//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with this program (see doc/license.txt); if not, write to the Free Software
//   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//   The GNU GPL is also currently available at
//   http://www.gnu.org/copyleft/gpl.html

#include "Python.h"

class SGFError {
public:
  SGFError();
};

char* SGFescape(char* s);

class Node {
 public:
  Node* previous;
  Node* next;
  Node* up;
  Node* down;
  int numChildren;
  int level;
  char* SGFstring;
  int parsed;
  PyObject* data;

  int posyD;

  Node(Node* prev, char* SGFst, int lvl) throw(SGFError);
  ~Node();
  PyObject* pathToNode();
  void parseNode() throw(SGFError);
  PyObject* getData();

  static int sloppy;
};

class Cursor {
 public:
  Cursor(char* sgf, int sloppy) throw(SGFError);
  ~Cursor();
  
  int atStart;
  int atEnd;
  int height;
  int width;
  Node* root;
  Node* currentN;
  int posx;
  int posy;

  char* encoding;

  int noChildren();
  PyObject* currentNode();
  void parse(char* s) throw(SGFError);
  void game(int n) throw(SGFError);
  PyObject* next(int n=0) throw(SGFError);
  PyObject* previous() throw(SGFError);
  PyObject* getRootNode(int n) throw(SGFError);
  void updateCurrentNode() throw(SGFError);
  void updateRootNode(PyObject* data, int n) throw(SGFError);
  char* rootNodeToString(PyObject* data);
  char* nodeToString(PyObject* data) throw(SGFError);
  char* outputVar(Node* node);
  char* output();
  void add(char* st);
  void delVariation(Node* node);
  void setFlags();  

 protected:
  void delVar(Node* node);
  void deltree(Node* node);

};

class intN {
 public:
  int data;
  intN* prev;
  intN(int i, intN* p); 
};

class IntStack {
 public:
  intN* root; 
  IntStack();
  void push(int i);
  void pop();
  int top();
  bool nonempty();
};

class nodeN {
 public:
  Node* data;
  nodeN* prev;
  nodeN(Node* i, nodeN* p);
};

class NodeStack {
 public:
  nodeN* root; 
  NodeStack();
  void push(Node* i);
  void pop();
  Node* top();
  bool nonempty();
};

