#!/usr/bin/python
# File: v.py

##   Copyright (C) 2001-4 Ulrich Goertz (u@g0ertz.de)

##   This is a simple SGF viewer; it comes with the go database program
##   Kombilo.

##   This program is free software; you can redistribute it and/or modify
##   it under the terms of the GNU General Public License as published by
##   the Free Software Foundation; either version 2 of the License, or
##   (at your option) any later version.

##   This program is distributed in the hope that it will be useful,
##   but WITHOUT ANY WARRANTY; without even the implied warranty of
##   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##   GNU General Public License for more details.

##   You should have received a copy of the GNU General Public License
##   along with this program (gpl.txt); if not, write to the Free Software
##   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##   The GNU GPL is also currently available at
##   http://www.gnu.org/copyleft/gpl.html


from Tkinter import *
from tkMessageBox import *
from ScrolledText import ScrolledText
import tkFileDialog
import Pmw

import cPickle
import os
import sys

try:
    import cjkcodecs.aliases
    cjkcodecs_available = 1
except:
    cjkcodecs_available = 0
    
from string import split, find, replace, join, strip
from copy import copy, deepcopy
from math import sqrt
from random import randint

try:
    from sgfpars import Node, Cursor, SGFError, SGFescape
except:
    from sgfparser import Node, Cursor, SGFError, SGFescape
from board1 import *

# ---------------------------------------------------------------------------------------

class BunchTkVar:
    """ This class is used to collect the Tk variables where the options
        are stored. """
    
    def saveToDisk(self, filename, onFailure = lambda:None):
        d = {}
        for x in self.__dict__.keys():
            d[x] = self.__dict__[x].get()
        try:
            f = open(filename, 'w')
            cPickle.dump(d,f)
            f.close()
        except IOError:
            onFailure()
            

    def loadFromDisk(self, filename, onFailure = lambda:None):
        try:
            f = open(filename)
            d = cPickle.load(f)
            f.close()
        except IOError:
            onFailure()
        else:
            for x in self.__dict__.keys():
                if d.has_key(x): self.__dict__[x].set(d[x])


# ---------------------------------------------------------------------------------------


class ScrolledText_HSB(ScrolledText):
    """ A ScrolledText which hides the scroll bar when it is not needed. """

    def __init__(self, parent, **args):
        if not args:
            args = {}
        apply(ScrolledText.__init__, (self, parent,) , args)
        self.bind('<Configure>', self.checkScrollbars)
        self.bind('<Key>', self.checkScrollbars)
        self.vbarPackinfo = self.vbar.pack_info()
        self.vbar.pack_forget()

    def insert(self, pos, text, tags = None):
        ScrolledText.insert(self, pos, text, tags)
        if self.yview() != (0.0, 1.0):
            apply(self.vbar.pack, (), self.vbarPackinfo)

    def delete(self, pos1, pos2):
        ScrolledText.delete(self, pos1, pos2)
        if self.yview() == (0.0, 1.0):
            self.vbar.pack_forget()

    def checkScrollbars(self, event):
        if self.yview() == (0.0, 1.0):
            self.vbar.pack_forget()
        else:
            apply(self.vbar.pack, (), self.vbarPackinfo)


# ---------------------------------------------------------------------------------------

class ScrolledList(Frame):
    """ A listbox with dynamic vertical and horizontal scrollbars. """
    
    def __init__(self, parent, **kw):
        Frame.__init__(self, parent)

        self.sbar = Scrollbar(self)
        self.sbar1 = Scrollbar(self)
        self.checking = 0
        
        if not kw: kw = {}

        for var, value in [('height', 12), ('width', 40), ('relief', SUNKEN),
                           ('selectmode', SINGLE), ('takefocus', 1), ('exportselection', 0)]:
            if not kw.has_key(var): kw[var] = value
        
        self.list = Listbox(self, kw)
        # self.rowconfigure(0, weight=1)
        # self.columnconfigure(0, weight=1)
        self.sbar.config(command = self.list.yview)
        self.sbar1.config(command = self.list.xview, orient='horizontal')
        self.list.config(xscrollcommand = self.sbar1.set, yscrollcommand = self.sbar.set)
        self.list.grid(row=0, column=0, sticky=NSEW)
        self.sbar.grid(row=0, column=1, sticky=NSEW)
        self.sbar1.grid(row=1, column=0, sticky=NSEW)

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        self.focus_force()

        self.onSelectionChange = None

        self.bind('<Up>', self.up)
        self.bind('<Down>', self.down)
        self.bind('<Prior>', self.pgup)
        self.bind('<Next>', self.pgdown)

        # self.unbind('<Configure>')
        self.bind('<Configure>', self.checkScrollbars)
        
        self.sbar.grid_forget()
        self.sbar1.grid_forget()

    def insert(self, index, data):
        if type(data) == type([]):
            apply(self.list.insert, [index] + data)
        else:
            self.list.insert(index, data)
        
        if self.list.yview() != (0.0, 1.0):
            self.sbar.grid(row=0, column=1, sticky=NSEW)
        if self.list.xview() != (0.0, 1.0):
            self.sbar1.grid(row=1, column=0, sticky=NSEW)

    def delete(self, index, data=None):
        if data:
            self.list.delete(index, data)
        else:
            self.list.delete(index)

        if self.list.yview() == (0.0, 1.0):
            self.sbar.grid_forget()

        if self.list.xview() == (0.0, 1.0):
            self.sbar1.grid_forget()


    def checkScrollbars(self, event=None):
        if self.checking:
            if self.list.yview() != (0.0, 1.0):
                self.sbar.grid(row=0, column=1, sticky=NSEW)
            else:
                self.sbar.grid_forget()
            if self.list.xview() != (0.0, 1.0):
                self.sbar1.grid(row=1, column=0, sticky=NSEW)
            else:
                self.sbar1.grid_forget()
        else:
            self.after(100, self.checkScrollbars)
        self.checking = 1-self.checking

            
    def up(self, event):
        if not self.list.curselection() or len(self.list.curselection())>1: return
        index = int(self.list.curselection()[0])
        if index != 0:
            self.list.select_clear(index)
            self.list.select_set(index-1)
            self.list.see(index-1)
            if self.onSelectionChange:
                self.onSelectionChange(None, index-1)

    def down(self, event):
        if not self.list.curselection() or len(self.list.curselection())>1: return
        index = int(self.list.curselection()[0])
        if index != self.list.size()-1:
            self.list.select_clear(index)
            self.list.select_set(index+1)
            self.list.see(index+1)
            if self.onSelectionChange:
                self.onSelectionChange(None, index+1)

    def pgup(self, event):
        if not self.list.curselection() or len(self.list.curselection())>1: return
        index = int(self.list.curselection()[0])
        if index >= 10:
            self.list.select_clear(index)
            self.list.select_set(index-10)
            self.list.see(index-10)
            if self.onSelectionChange:
                self.onSelectionChange(None, index-10)
        elif self.list.size():
            self.list.select_clear(index)
            self.list.select_set(0)
            self.list.see(0)
            if self.onSelectionChange:
                self.onSelectionChange(None, 0)

    def pgdown(self, event):
        if not self.list.curselection() or len(self.list.curselection())>1: return
        index = int(self.list.curselection()[0])
        if index <= self.list.size()-10:
            self.list.select_clear(index)
            self.list.select_set(index+10)
            self.list.see(index+10)
            if self.onSelectionChange:
                self.onSelectionChange(None, index+10)
        elif self.list.size():
            self.list.select_clear(index)
            self.list.select_set(self.list.size()-1)
            self.list.see(END)
            if self.onSelectionChange:
                self.onSelectionChange(None, self.list.size()-1)



# ---------------------------------------------------------------------------------------

class SGFtreeCanvas(Frame):
    """ The canvas (in the data window) displaying the tree structure of the current game."""

    def __init__(self, parent, **args):

        Frame.__init__(self, parent)
        
        if not args: args = {}

        for var, value in [('height', 100), ('width', 150), ('relief', SUNKEN)]:
            if not args.has_key(var): args[var] = value
        
        self.canvas = Canvas(self, background='#FFBA59')
        apply(self.canvas.config, (), args)
        self.canvas.config(scrollregion=(0,0,1000,30))

        self.sbar_vert = Scrollbar(self)
        self.sbar_hor = Scrollbar(self)

        self.sbar_vert.config(command = self.yview)
        self.sbar_hor.config(command = self.xview, orient='horizontal')
        self.canvas.config(xscrollcommand = self.sbar_hor.set, yscrollcommand = self.sbar_vert.set)
        
        self.movenoCanvas = Canvas(self, width=150, height=15, background='white')
        self.movenoCanvas.config(scrollregion=(0,0,1000,15))
        # self.movenoCanvas.config(xscrollcommand = self.sbar_hor.set)

        self.updateMovenoCanvas()
     
        self.movenoCanvas.grid(row=0, column=0, sticky=NSEW)
        self.sbar_vert.grid(row=1, column=1, sticky=NSEW)
        self.sbar_hor.grid(row=2, column=0, sticky=NSEW)
        self.canvas.grid(row=1, column=0, sticky=NSEW)

        self.rowconfigure(1, weight=1)
        self.columnconfigure(0, weight=1)

        # self.config(height=100, width=200)
        # self.canvSize = 200,100
        self.drawn = []

        self.lastbind = None

    def updateMovenoCanvas(self):
        """ Put numbers on the small white canvas above the SGF tree canvas, to indicate move numbers."""
        
        self.movenoCanvas.delete(ALL)
        for i in range(90):
            self.movenoCanvas.create_text(int(SGFtreeCanvas.UNIT*.75)+i*5*SGFtreeCanvas.UNIT,7,text=`5*i`,
                                          font=('Helvetica', 8))
   

    def xview(self, a1, a2=None, a3=None):
        if a1 == MOVETO:
            self.movenoCanvas.xview(a1, a2)
            self.canvas.xview(a1, a2)
        elif a1 == SCROLL:
            self.movenoCanvas.xview(a1, a2, a3)
            self.canvas.xview(a1, a2, a3)
        else:
            return
            # raise Exception()


    def yview(self, a1, a2=None, a3=None):

        if a1 == MOVETO: self.canvas.yview(a1, a2)
        elif a1 == SCROLL: self.canvas.yview(a1, a2, a3)
        elif a1 == 'refresh': pass
        else: return # raise Exception()

        self.canvas.update_idletasks()

        vert = self.sbar_vert.get()
        y0 = int(vert[0] * self.canvSize[1])
        y1 = int(vert[1] * self.canvSize[1])

        u = max(0, (y0-300)/SGFtreeCanvas.UNIT)
        l = (y1+500)/SGFtreeCanvas.UNIT

        # print 'yview', u,l 

        nodelist = [(self.rootnode, 0,0)]

        while nodelist:
            c, posx, posy = nodelist[-1]
            del nodelist[-1]

            # print posx, posy

            if u <= posy <= l and not posy in self.drawn:
                self.mark(c, posx, posy)
                
                if c.previous:
                    if c.up and c.up.up:
                        self.link(posx, posy, c.posyD+1)
                    else:
                        self.link(posx, posy, c.posyD)

                    if c.down and posy+c.down.posyD>l:
                        if c.up:
                            self.link(posx, posy+c.down.posyD, c.down.posyD+1)
                        else:
                            self.link(posx, posy+c.down.posyD, c.down.posyD)

            if c != self.rootnode and c.down and posy + c.down.posyD <= l:
                d = c.down
                py = posy + d.posyD
                while d.down and py + d.down.posyD < u:
                    d = d.down
                    py += d.posyD
                nodelist.append((d, posx, py))

            if c.down and posy + c.down.posyD < u: continue

            while c.next:
                c = c.next
                posx += 1
                if u <= posy <= l and not posy in self.drawn:

                    self.mark(c, posx, posy)

                    if c.previous:
                        if c.up and c.up.up:
                            self.link(posx, posy, c.posyD+1)
                        else:
                            self.link(posx, posy, c.posyD)

                        if c.down and posy+c.down.posyD>l:
                            if c.up: delta2 = 1
                            else: delta2 = 0
                            self.link(posx, posy + c.down.posyD, c.down.posyD + delta2)

                if c.down and posy + c.down.posyD <= l:
                    d = c.down
                    py = posy + d.posyD
                    while d.down and py + d.down.posyD < u:
                        d = d.down
                        py += d.posyD
                    nodelist.append((d, posx, py))

        self.canvas.lower('lines')
            
        for i in range(u, l):
             if not i in self.drawn: self.drawn.append(i)

        
    def mark(self, c, posx, posy):
        try:
            n = c.getData()
            if n.has_key('W'): color = 'white'
            elif n.has_key('B'): color = 'black'
            else:
                color = 'red'
        except: color = 'yellow'

        self.canvas.create_oval(SGFtreeCanvas.UNIT*posx+SGFtreeCanvas.UNIT/2,
                                SGFtreeCanvas.UNIT*posy+SGFtreeCanvas.UNIT/2,
                                SGFtreeCanvas.UNIT*posx+SGFtreeCanvas.UNIT,
                                SGFtreeCanvas.UNIT*posy+SGFtreeCanvas.UNIT, fill=color)

        try:
            if n.has_key('C') or n.has_key('TR') or n.has_key('SQ') or n.has_key('CR') or n.has_key('LB') or n.has_key('MA'):
                self.canvas.create_oval(SGFtreeCanvas.UNIT*posx+SGFtreeCanvas.UNIT*2/3,
                                        SGFtreeCanvas.UNIT*posy+SGFtreeCanvas.UNIT*2/3,
                                        SGFtreeCanvas.UNIT*posx+SGFtreeCanvas.UNIT*5/6,
                                        SGFtreeCanvas.UNIT*posy+SGFtreeCanvas.UNIT*5/6, fill='blue')
        except: pass
        

    def link(self, posx, posy, delta):

        s4 = SGFtreeCanvas.UNIT/4
        s34 = 3*SGFtreeCanvas.UNIT/4
        
        if delta == 0:
            self.canvas.create_line(SGFtreeCanvas.UNIT*posx-s4,
                                    SGFtreeCanvas.UNIT*posy+s34,
                                    SGFtreeCanvas.UNIT*posx+s34,
                                    SGFtreeCanvas.UNIT*posy+s34,
                                    fill='blue', tags='lines', width=2)        
        else:
            self.canvas.create_line(SGFtreeCanvas.UNIT*posx-s4,
                                    SGFtreeCanvas.UNIT*(posy-1)+s34,
                                    SGFtreeCanvas.UNIT*posx+s34,
                                    SGFtreeCanvas.UNIT*posy+s34,
                                    fill='blue', tags='lines', width=2)        
            if delta > 1:
                self.canvas.create_line(SGFtreeCanvas.UNIT*posx-s4,
                                        SGFtreeCanvas.UNIT*(posy-delta)+s34,
                                        SGFtreeCanvas.UNIT*posx-s4,
                                        SGFtreeCanvas.UNIT*(posy-1)+s34,
                                        fill='blue', tags='lines', width=2)        


SGFtreeCanvas.UNIT = 25



# ---------------------------------------------------------------------------------------

class DataWindow:

    def __init__(self, master):
        self.mster = master
        
        window = Toplevel()
        window.title('Data window')
        window.protocol('WM_DELETE_WINDOW', self.quit)
        # window.withdraw()
        
        win = Pmw.PanedWidget(window, orient='vertical')

        self.win = win
        win.pack(expand=YES, fill=BOTH)
        
        self.initPanes()
        
        self.SGFtreeC = SGFtreeCanvas(self.gametreeF)
        
        self.SGFtreeC.pack(side=LEFT, expand=YES, fill=BOTH)

        self.guessModeCanvas = Canvas(self.gametreeF, width=150, height=80, background='#FFBA59')

        
        
        self.filelistV = IntVar()
        self.filelistV.set(1)
        b1 = Checkbutton(self.toolbarF, text = 'Files', variable = self.filelistV,
                         command = self.toggleFilelist, indicatoron=0)
        b1.grid(row=0, column=0)
        
        self.gamelistV = IntVar()
        self.gamelistV.set(1)
        b1 = Checkbutton(self.toolbarF, text = 'Games', variable = self.gamelistV,
                         command = self.toggleGamelist, indicatoron=0)
        b1.grid(row=0, column=1)

        self.gameinfoV = IntVar()
        self.gameinfoV.set(1)
        b1 = Checkbutton(self.toolbarF, text = 'Info', variable = self.gameinfoV,
                         command = self.toggleGameinfo, indicatoron=0)
        b1.grid(row=0, column=2)

        self.editToolsV = IntVar()
        self.editToolsV.set(1)
        b1 = Checkbutton(self.toolbarF, text = 'Edit tools', variable = self.editToolsV,
                         command = self.toggleEditTools, indicatoron=0)
        b1.grid(row=0, column=3)

        self.gametreeV = IntVar()
        self.gametreeV.set(1)
        b1 = Checkbutton(self.toolbarF, text = 'Game tree', variable = self.gametreeV,
                         command = self.toggleGametree, indicatoron=0)
        b1.grid(row=0, column=4)

        self.commentsV = IntVar()
        self.commentsV.set(1)
        b1 = Checkbutton(self.toolbarF, text = 'Comments', variable = self.commentsV,
                         command = self.toggleComments, indicatoron=0)
        b1.grid(row=0, column=5)

        self.filelistF.rowconfigure(0, weight=1)
        self.filelistF.columnconfigure(0, weight=1)

        self.filelist = ScrolledList(self.filelistF, height = 4)
        self.filelist.grid(row=0, column=0, rowspan=4, sticky=NSEW)

        self.filelistB1 = Button(self.filelistF, text = 'NEW', command = self.mster.newFile)
        self.filelistB1.grid(row=0, column=1, sticky=S)
        self.filelistB2 = Button(self.filelistF, text= 'OPEN', command = self.mster.openFile)
        self.filelistB2.grid(row=1, column=1, sticky=S)
        self.filelistB3 = Button(self.filelistF, text = 'DEL', command = self.mster.delFile)
        self.filelistB3.grid(row=0, column=2, sticky=S)
        self.filelistB4 = Button(self.filelistF, text = 'split', command=self.mster.splitCollection)
        self.filelistB4.grid(row=1, column=2, sticky=S)

        self.tkImages = []
        for button, filename in [(self.filelistB1, 'new'), (self.filelistB2, 'open'), (self.filelistB3, 'trash'),
                                 (self.filelistB4, 'split')]:
            try:
                im = PhotoImage(file=os.path.join(self.mster.basepath, 'gifs/', filename+'.gif'))
                button.config(image=im)
                self.tkImages.append(im)
            except:
                pass

        self.filelist.onSelectionChange = self.mster.changeCurrentFile
        self.filelist.list.bind('<1>', self.mster.changeCurrentFile)

        self.gamelistF.rowconfigure(0, weight=1)
        self.gamelistF.columnconfigure(0, weight=1)

        self.gamelist = ScrolledList(self.gamelistF, height=4)
        self.gamelist.grid(row=0, column=0, rowspan=4, sticky=NSEW)

        self.gamelistB1 = Button(self.gamelistF, text = 'NEW', command = self.mster.newGame)
        self.gamelistB1.grid(row=0, column=1, sticky=S)
        self.gamelistB2 = Button(self.gamelistF, text = 'DEL', command = self.mster.delGame)
        self.gamelistB2.grid(row=1, column=1, sticky=S)

        for button, filename in [(self.gamelistB1, 'new'), (self.gamelistB2, 'trash')]:
            try:
                im = PhotoImage(file=os.path.join(self.mster.basepath, 'gifs/', filename+'.gif'))
                button.config(image=im)
                self.tkImages.append(im)
            except:
                pass
        
        self.gamelist.onSelectionChange = self.mster.changeCurrentGame
        self.gamelist.list.bind('<1>', self.gamelistClick)
        self.gamelist.list.bind('<B1-Motion>', self.gamelistDrag)
        self.gamelist.list.bind('<ButtonRelease-1>', self.gamelistRelease)

        self.gamelist.clickedLast = -1
        self.gamelist.dragLast = -1

        self.gameinfo = ScrolledText_HSB(self.gameinfoF, height = 4, width = 20, wrap=WORD, 
                                         relief=SUNKEN)
        self.gameinfo.config(state=DISABLED)
        self.gameinfo.pack(fill=BOTH, expand=NO)

        lab = Label(self.editToolsF, text='Ctrl-Click:')
        
        self.labelType = StringVar()
        self.labelType.set('DEL ST')

        self.removeStoneButton = Radiobutton(self.editToolsF, text='DEL ST', indicatoron=0,
                                             variable=self.labelType, value='DEL ST')
        self.triangleButton = Radiobutton(self.editToolsF, text='TR', indicatoron=0,
                                          variable=self.labelType, value='TR')
        self.squareButton = Radiobutton(self.editToolsF, text='SQ', indicatoron=0,
                                        variable=self.labelType, value='SQ')
        self.letterUButton = Radiobutton(self.editToolsF, text='ABC', indicatoron=0,
                                         variable=self.labelType, value='ABC')
        self.letterLButton = Radiobutton(self.editToolsF, text='abc', indicatoron=0,
                                         variable=self.labelType, value='abc')
        self.numberButton = Radiobutton(self.editToolsF, text='123', indicatoron=0,
                                        variable=self.labelType, value='123')

        ca1 = Canvas(self.editToolsF, height=30, width=4, highlightthickness=0)
        ca1.create_line(3,0,3,30, fill='black', width=2)
        ca2 = Canvas(self.editToolsF, height=30, width=4, highlightthickness=0)
        ca2.create_line(3,0,3,30, fill='black', width=2)
        ca3 = Canvas(self.editToolsF, height=30, width=4, highlightthickness=0)
        ca3.create_line(3,0,3,30, fill='black', width=2)

        self.delButton = Button(self.editToolsF, text='DEL', command = lambda self=self: self.mster.delVar())

        self.guessMode = IntVar()
        self.SNM = 0
        self.guessRightWrong = [0,0]
        self.guessModeButton = Checkbutton(self.editToolsF, text = 'Guess mode', indicatoron=0,
                                           variable = self.guessMode, command = self.toggleGuessMode)
        self.mirrorButton1 = Button(self.editToolsF, text = 'M |',
                                    command=lambda self=self, flip=flip_mirror1: self.mster.mirrorSGF(flip))
        self.mirrorButton2 = Button(self.editToolsF, text = 'M /',
                                    command=lambda self=self, flip=flip_mirror2: self.mster.mirrorSGF(flip))
        self.mirrorButton3 = Button(self.editToolsF, text = 'R',
                                    command=lambda self=self, flip=flip_rotate: self.mster.mirrorSGF(flip))


        self.images = []
        for button, image in [(self.removeStoneButton, 'delst.gif'), (self.triangleButton, 'tr.gif'),
                              (self.squareButton, 'sq.gif'), (self.letterUButton, 'abc-u.gif'),
                              (self.letterLButton, 'abc-l.gif'),
                              (self.numberButton, '123.gif'), (self.delButton, 'delvar.gif'),
                              (self.guessModeButton, 'guess.gif'),
                              (self.mirrorButton1, 'mirrorv.gif'), (self.mirrorButton2, 'mirrord.gif'),
                              (self.mirrorButton3, 'rotate.gif') ]:
            try:
                im = PhotoImage(file=os.path.join(self.mster.basepath, 'gifs/', image))
                button.config(image=im)
                self.images.append(im)
            except:
                pass

        lab.pack(side=LEFT)
        self.removeStoneButton.pack(side=LEFT, fill=Y)
        self.triangleButton.pack(side=LEFT, fill=Y)
        self.squareButton.pack(side=LEFT, fill=Y)
        self.letterUButton.pack(side=LEFT, fill=Y)
        self.letterLButton.pack(side=LEFT, fill=Y)
        self.numberButton.pack(side=LEFT, fill=Y)
        ca1.pack(side=LEFT, fill=Y)
        self.delButton.pack(side=LEFT, fill=Y)
        ca2.pack(side=LEFT, fill=Y)
        self.mirrorButton1.pack(side=LEFT, fill=Y)
        self.mirrorButton2.pack(side=LEFT, fill=Y)
        self.mirrorButton3.pack(side=LEFT, fill=Y)
        ca3.pack(side=LEFT, fill=Y)
        self.guessModeButton.pack(side=RIGHT, fill=Y)
        self.comments = ScrolledText_HSB(self.commentsF, height = 5, width = 20, wrap=WORD, 
                                         relief=SUNKEN)
        self.comments.pack(expand=YES, fill=BOTH)
        self.window = window


    def initPanes(self):
        self.toolbarF = self.win.add(name='toolb', min=28, max=28)
        self.filelistF = self.win.add(name='filel', min=1, max=500, size=65)
        self.gamelistF = self.win.add(name='gamel', min=1, max=500, size=65)
        self.gameinfoF = self.win.add(name='gamei', min=1, max=500, size=75)
        self.editToolsF = self.win.add(name='editt', min=1, max=30, size=30) 
        self.gametreeF = self.win.add(name='gamet', min=1, max=500, size=90) 
        self.commentsF = self.win.add(name='comm', min=1, max=500, size=80) 

    def quit(self):
        self.mster.options.datawindowVisible.set(0)
        self.window.withdraw()

    def toggleGuessMode(self):
        if self.guessMode.get():
            self.guessRightWrong = [0,0]
            self.guessModeCanvas.pack(side=RIGHT, expand=NO, fill=BOTH)
            self.guessModeCanvas.delete(ALL)
            self.guessModeCanvas.create_rectangle(1,1,79,79, fill='', outline='black')

            if self.mster.options.showNextMoveVar.get():
                self.SNM = 1
                self.mster.options.showNextMoveVar.set(0)
                self.mster.showNextMove()
            else:
                self.SNM = 0
        else:
            self.guessModeCanvas.pack_forget()
            if self.SNM: self.mster.options.showNextMoveVar.set(1)
            self.mster.showNextMove()

    def gamelistClick(self, event):
        self.gamelist.clickedLast = self.gamelist.list.nearest(event.y)
        self.gamelist.dragLast = -1
        self.mster.changeCurrentGame(event)

    def gamelistDrag(self, event):

        i = self.gamelist.list.nearest(event.y)
        if self.gamelist.dragLast == -1:
            if self.gamelist.clickedLast == i: return
            else: self.gamelist.dragLast = self.gamelist.clickedLast
        
        if self.gamelist.dragLast != i:
            s = self.gamelist.list.get(self.gamelist.dragLast)
            self.gamelist.delete(self.gamelist.dragLast)
            self.gamelist.insert(i, s)
            self.gamelist.list.select_set(i)
            self.gamelist.dragLast = i 
        return 'break'

    def gamelistRelease(self, event):

        if self.gamelist.dragLast == -1:
            return None, None
        
        i = self.gamelist.list.nearest(event.y)

        if self.gamelist.dragLast != i:
            s = self.gamelist.list.get(self.gamelist.dragLast)
            self.gamelist.delete(self.gamelist.dragLast)
            self.gamelist.insert(i, s)
            self.gamelist.list.select_set(i)
            self.gamelist.dragLast = i

        if self.gamelist.clickedLast != i:

            try:
                n = self.mster.cursor.root.next
                for j in range(self.gamelist.clickedLast): n = n.down

                m = self.mster.cursor.root.next
                for j in range(i): m = m.down

                if n.up: n.up.down = n.down
                else: self.mster.cursor.root.next = n.down
                if n.down: n.down.up = n.up

                if i < self.gamelist.clickedLast: # insert n above m
                    n.up = m.up
                    if m.up: m.up.down = n
                    else: self.mster.cursor.root.next = n
                    m.up = n
                    n.down = m
                else:                             # insert n below m
                    n.down = m.down
                    if m.down: m.down.up = n
                    m.down = n
                    n.up = m

                self.mster.cursor.updateGamelist(0)
                self.mster.cursor.currentGame = i
                self.gamelist.list.select_set(i)
                self.gamelist.list.see(i)
                self.updateGameInfo(self.mster.cursor)

                filelistIndex = self.mster.currentFileNum
                d = self.mster.filelist[filelistIndex][4]

                if i < self.gamelist.clickedLast:

                    if d.has_key(self.gamelist.clickedLast):
                        p_last = d[self.gamelist.clickedLast]
                    else: p_last = ()

                    jj = self.gamelist.clickedLast
                    while jj > i:
                        if d.has_key(jj-1):
                            d[jj] = d[jj-1]
                        else:
                            if d.has_key(jj): del d[jj]
                        jj -= 1

                    d[i] = p_last

                else:
                    if d.has_key(i):
                        p_i = d[i]
                    else: p_i = ()

                    jj = self.gamelist.clickedLast
                    while jj < i:
                        if d.has_key(jj+1):
                            d[jj] = d[jj+1]
                        else:
                            if d.has_key(jj):
                                del d[jj]
                        jj += 1

                    d[self.gamelist.clickedLast] = p_i

                return self.gamelist.clickedLast, i

            except SGFError: showwarning('Error', 'SGF error')
            except: showwarning('Error', 'An error occured, please send a bug report.')
            
        return None, None

    def get_geometry(self):
        l = [ self.filelistV.get(), self.win._size['filel'],
              self.gamelistV.get(), self.win._size['gamel'],
              self.gameinfoV.get(), self.win._size['gamei'],
              self.editToolsV.get(), self.win._size['editt'],
              self.gametreeV.get(), self.win._size['gamet'],
              self.commentsV.get(), self.win._size['comm'] ]
        l1 = [ `x` for x in l ]
        l1.append(self.window.geometry())

        return join(l1, '|%')
        
    def set_geometry(self, s):
        l = split(s, '|%')

        l1 = [ self.filelistV, 
               self.gamelistV, 
               self.gameinfoV, 
               self.editToolsV,
               self.gametreeV, 
               self.commentsV ]
               
        for i in range(len(l)/2):
            l1[i].set(int(l[2*i]))
            if int(l[2*i]):
                self.win.configurepane(i+1, min=1, max=500, size = int(l[2*i+1]))
            else:
                self.win.configurepane(i+1, min=1, max=1, size = 1)
        self.win.updatelayout()
        self.window.geometry(l[-1])

    def toggleFilelist(self):
        if self.filelistV.get():
            s = self.window.geometry()
            x1, x2, x3 = split(s, '+')
            x, y = split(x1, 'x')
            y = `int(y)+65`
            self.window.geometry('%sx%s+%s+%s' % (x, y, x2, x3))
            self.win.configurepane(1, min=10, max=1.0, size=65)
        else:
            s = self.window.geometry()
            x1, x2, x3 = split(s, '+')
            x, y = split(x1, 'x')
            y = `int(y)-self.win._size['filel']`
            self.window.geometry('%sx%s+%s+%s' % (x, y, x2, x3))
            self.win.configurepane(1, min=0.0, max=0.0, size=0.0)
        self.win.updatelayout()
        
    def toggleGamelist(self):
        if self.gamelistV.get():
            s = self.window.geometry()
            x1, x2, x3 = split(s, '+')
            x, y = split(x1, 'x')
            y = `int(y)+65`
            self.window.geometry('%sx%s+%s+%s' % (x, y, x2, x3))
            self.win.configurepane(2, min=10, max=1.0, size=65)
        else:
            s = self.window.geometry()
            x1, x2, x3 = split(s, '+')
            x, y = split(x1, 'x')
            y = `int(y)-self.win._size['gamel']`
            self.window.geometry('%sx%s+%s+%s' % (x, y, x2, x3))
            self.win.configurepane(2, min=0.0, max=0.0, size=0.0)
        self.win.updatelayout()
        
    def toggleGameinfo(self):
        if self.gameinfoV.get():
            s = self.window.geometry()
            x1, x2, x3 = split(s, '+')
            x, y = split(x1, 'x')
            y = `int(y)+75`
            self.window.geometry('%sx%s+%s+%s' % (x, y, x2, x3))
            self.win.configurepane(3, min=10, max=1.0, size=75)
        else:
            s = self.window.geometry()
            x1, x2, x3 = split(s, '+')
            x, y = split(x1, 'x')
            y = `int(y)-self.win._size['gamei']`
            self.window.geometry('%sx%s+%s+%s' % (x, y, x2, x3))
            self.win.configurepane(3, min=0.0, max=0.0, size=0.0)

            
    def toggleEditTools(self):
        if self.editToolsV.get():
            s = self.window.geometry()
            x1, x2, x3 = split(s, '+')
            x, y = split(x1, 'x')
            y = `int(y)+30`
            self.window.geometry('%sx%s+%s+%s' % (x, y, x2, x3))
            self.win.configurepane(4, min=10, max=1.0, size=30)
        else:
            s = self.window.geometry()
            x1, x2, x3 = split(s, '+')
            x, y = split(x1, 'x')
            y = `int(y)-self.win._size['editt']`
            self.window.geometry('%sx%s+%s+%s' % (x, y, x2, x3))
            self.win.configurepane(4, min=0.0, max=0.0, size=0.0)

            
    def toggleGametree(self):
        if self.gametreeV.get():
            s = self.window.geometry()
            x1, x2, x3 = split(s, '+')
            x, y = split(x1, 'x')
            y = `int(y)+90`
            self.window.geometry('%sx%s+%s+%s' % (x, y, x2, x3))
            self.win.configurepane(5, min=10, max=1.0, size=90)
        else:
            s = self.window.geometry()
            x1, x2, x3 = split(s, '+')
            x, y = split(x1, 'x')
            y = `int(y)-self.win._size['gamet']`
            self.window.geometry('%sx%s+%s+%s' % (x, y, x2, x3))
            self.win.configurepane(5, min=0.0, max=0.0, size=0.0)

            
    def toggleComments(self):
        if self.commentsV.get():
            s = self.window.geometry()
            x1, x2, x3 = split(s, '+')
            x, y = split(x1, 'x')
            y = `int(y)+80`
            self.window.geometry('%sx%s+%s+%s' % (x, y, x2, x3))
            self.win.configurepane(6, min=10, max=1.0, size=80)
        else:
            s = self.window.geometry()
            x1, x2, x3 = split(s, '+')
            x, y = split(x1, 'x')
            y = `int(y)-self.win._size['comm']`
            self.window.geometry('%sx%s+%s+%s' % (x, y, x2, x3))
            self.win.configurepane(6, min=0.0, max=0.0, size=0.0)


    def updateGameInfo(self, cursor):

        currentGame = cursor.currentGame
        
        try:
            node = cursor.getRootNode(currentGame)
        except: return
        
        t = []

        s1 = cursor.transcode('PW',node)
        if s1: t.append(s1)
        else: t.append(' ?')

        s1 = cursor.transcode('WR',node)
        if s1:
            t.append(', ')
            t.append(s1)

        t.append(' - ')

        s1 = cursor.transcode('PB',node)
        if s1: t.append(s1)
        else: t.append(' ?')

        s1 = cursor.transcode('BR',node)
        if s1:
            t.append(', ')
            t.append(s1)

        s1 = cursor.transcode('RE',node)
        if s1: t.append(', ' + s1)
        s1 = cursor.transcode('KM',node)
        if s1: t.append(' (Komi ' + s1 + ')')
        s1 = cursor.transcode('HA',node)
        if s1: t.append(' (Hcp ' + s1 + ')')

        t.append('\n')

        if node.has_key('EV'): t.append(cursor.transcode('EV', node) + ', ')
        if node.has_key('RO'): t.append(cursor.transcode('RO', node) + ', ')
        if node.has_key('DT'): t.append(cursor.transcode('DT', node) + '\n')

        s1 = cursor.transcode('GC',node)
        if s1:
            s1 = replace(s1, '\n\r', ' ')
            s1 = replace(s1, '\r\n', ' ')
            s1 = replace(s1, '\r', ' ')
            s1 = replace(s1, '\n', ' ')
            
            t.append(s1)

        self.gameinfo.config(state=NORMAL)
        self.gameinfo.delete('1.0', END)
        self.gameinfo.insert('1.0', join(t, ''))
        self.gameinfo.config(state=DISABLED)



# -----------------------------------------------------------------------------------


def flip_mirror1(pos):

    if not pos: return ''

    x = ord(pos[0]) - ord('a')
    y = ord(pos[1]) - ord('a')
    x = 18 - x
    return chr(x+ord('a')) + chr(y+ord('a'))


def flip_mirror2(pos):

    if not pos: return ''

    x = ord(pos[0]) - ord('a')
    y = ord(pos[1]) - ord('a')
    help = 18 - y
    y = 18 - x
    x = help
    return chr(x+ord('a')) + chr(y+ord('a'))

def flip_rotate(pos):

    if not pos: return ''
    
    x = ord(pos[0]) - ord('a')
    y = ord(pos[1]) - ord('a')
    help = 18 - x
    x = y
    y = help
    return chr(x+ord('a')) + chr(y+ord('a'))



class EnhancedCursor(Cursor):
    """
    This integrates the cursor of an SGF file with an SGFtreeCanvas. 
    Adds a snapshot/restore feature to Cursor. """
    
    def __init__(self, sgf, comments, SGFtreeC, master):

        try:
            Cursor.__init__(self, sgf, 1)
        except:
            raise SGFError()
        self.mster = master
        
        self.currentGame = 0
        self.updateGamelist()
        
        self.comments = comments
        self.SGFtreeCanv = SGFtreeC

        self.SGFtreeCanv.canvSize = 1, 1
        self.SGFtreeCanv.drawn = []

        self.SGFtreeCanv.rootnode = self.root.next
        self.updateTree()


    def transcode(self, prop, node = None):
        if type(node) == type({}):
            d = node
        elif node:
            d = node.getData()
        else: d = self.currentNode()
        if not d.has_key(prop) or not d[prop]: return ''
        s = d[prop][0]
        if self.encoding:
            try:
                s = s.decode(self.encoding.lower(), 'ignore').encode('utf-8', 'ignore')
            except:
                pass        
        return s


    def updateTree(self, update=1):
        self.SGFtreeCanv.canvas.delete('all')
        self.SGFtreeCanv.drawn = []
        
        width = SGFtreeCanvas.UNIT * (self.width+1) + 30 

        if self.SGFtreeCanv.rootnode.down: h = self.SGFtreeCanv.rootnode.down.posyD
        else:
            n = self.root.next
            h = 0
            while n.down:
                n = n.down
                h += n.posyD
            h = self.height - h

        height = SGFtreeCanvas.UNIT * (h+1) + 30

        self.SGFtreeCanv.canvSize = width, height
        self.SGFtreeCanv.canvas.config(scrollregion=(0,0,width,height))
        self.SGFtreeCanv.movenoCanvas.config(scrollregion=(0,0,width,15))
        self.SGFtreeCanv.updateMovenoCanvas()

        if self.SGFtreeCanv.lastbind is not None: self.SGFtreeCanv.canvas.unbind('<1>', self.SGFtreeCanv.lastbind)
        self.SGFtreeCanv.lastbind = self.SGFtreeCanv.canvas.bind('<1>', self.onButton1)
        
        self.SGFtreeCanv.yview('refresh')

        if update:
            self.SGFtreeCanv.update_idletasks()
            self.seeCurrent()



    def updateGamelist(self, select = 1):
        """Update the game list in the data window (i.e. the list of games in the SGF collection of self."""
        
        self.mster.dataWindow.gamelist.delete(0, END)
        i = 0
        n = self.root.next
        while n:
            s = '[' + `i` + '] '

            
            d = n.getData()
            s += self.transcode('PW',n)
            s += ' - '
            s += self.transcode('PB',n)
            s1 = self.transcode('RE',n)
            if s1: s += ' (' + s1 + ')'

            self.mster.dataWindow.gamelist.insert(END, s)
            n = n.down
            i += 1

        if select:
            self.mster.dataWindow.gamelist.list.select_set(0)
            self.mster.dataWindow.gamelist.list.see(0)
            self.mster.dataWindow.updateGameInfo(self)



    def onButton1(self, event):
        try:
            x1 = self.SGFtreeCanv.canvas.canvasx(event.x)
            y1 = self.SGFtreeCanv.canvas.canvasy(event.y)
        
            x = int((x1-SGFtreeCanvas.UNIT/2)/SGFtreeCanvas.UNIT)
            y = int((y1-SGFtreeCanvas.UNIT/2)/SGFtreeCanvas.UNIT)

            p1 = self.currentN.pathToNode()

            c = self.root.next
            for i in range(self.currentGame): c = c.down

            posx, posy = 0,0
        
            while (posx, posy) != (x,y):
                if not c.next:
                    return
                c = c.next
                posx += 1
            
                if posx > x: break 
                while posy < y and c.down and posy + c.down.posyD <= y:
                    c = c.down
                    posy += c.posyD

            p2 = c.pathToNode()

            i = 0

            while i < len(p1) and i < len(p2) and p1[i] == p2[i]:
                i += 1

            for j in range(len(p1)-i):
                self.mster.prev(0)
            for j in range(i, len(p2)):
                self.mster.next(p2[j],0)

            self.seeCurrent()

        except: showwarning('Error', 'SGF error')


    def next(self, n=0, markCurrent=1):

        res = Cursor.next(self, n)

        if markCurrent:
            self.seeCurrent()

        return res

    def previous(self, markCurrent=1):
        res = Cursor.previous(self)

        if markCurrent:
            self.seeCurrent()
            
        return res

    def add(self, st):
        w = self.width
        Cursor.add(self, st)

        if not self.currentN.up and not self.width > w:
            self.SGFtreeCanv.mark(self.currentN, self.posx, self.posy)
            self.SGFtreeCanv.link(self.posx, self.posy, 0)
            self.SGFtreeCanv.canvas.lower('lines');
            self.seeCurrent()
        else:
            self.updateTree()

        

    def game(self, n=0, update = 1):
        Cursor.game(self, n)
        self.currentGame = n
        self.SGFtreeCanv.rootnode = self.currentN
        self.updateTree(update)
        

    def symmetry(self, flip):

        nodelist = [self.root.next]

        while nodelist:

            node = nodelist[-1]
            del nodelist[-1]

            while node:
                try:
                    d = node.getData()
                    for prop in ['B', 'W', 'AB', 'AW', 'AE', 'CR', 'SQ', 'TR', 'MA', 'TB', 'TW', 'DD', 'L']:
                        if d.has_key(prop):
                            d[prop] = [ flip(x) for x in d[prop] ]
                    
                    for prop in ['LB', 'VW']:
                        if d.has_key(prop):
                            d[prop] = [ flip(split(x, ':')[0]) + ':' + split(x, ':')[1] for x in d[prop] ]
                    for prop in ['LN', 'AR']:
                        if d.has_key(prop):
                            d[prop] = [ flip(split(x, ':')[0]) + ':' + flip(split(x, ':')[1]) for x in d[prop]]
                    if node.previous:
                        node.SGFstring = self.nodeToString(d)
                    else:
                        node.SGFstring = self.rootNodeToString(d)
                except:
                    showwarning('Error', 'SGF error')

                try:
                    if node.down: nodelist.append(node.down)
                    node = node.next
                except:
                    break
                

    def seeCurrent(self):

        x,y = self.posx, self.posy

        self.SGFtreeCanv.canvas.delete('curr')    
        self.SGFtreeCanv.canvas.create_oval(x*SGFtreeCanvas.UNIT+SGFtreeCanvas.UNIT/2,
                                            y*SGFtreeCanvas.UNIT+SGFtreeCanvas.UNIT/2,
                                            (x+1)*SGFtreeCanvas.UNIT,
                                            (y+1)*SGFtreeCanvas.UNIT, fill='', outline='green',
                                            width=4, tags='curr')

        x = (x+.5)*SGFtreeCanvas.UNIT
        y = y*SGFtreeCanvas.UNIT

        # print 'seecurrent', x, y

        self.SGFtreeCanv.update_idletasks()
        hor = self.SGFtreeCanv.sbar_hor.get()
        vert = self.SGFtreeCanv.sbar_vert.get()

        cv0, cv1 = self.SGFtreeCanv.canvSize

        # print 'x: %1.3f, [%1.3f, %1.3f], y: %1.3f, [%1.3f, %1.3f]' % (x*1.0/cv0, hor[0], hor[1], \
        #                                                               y*1.0/cv1, vert[0], vert[1])

        x1 = (hor[1]-hor[0]) * cv0
        y1 = (vert[1]-vert[0]) * cv1

        if x-40 < hor[0]*cv0: self.SGFtreeCanv.xview('moveto', (x-50)*1.0/cv0)
        elif x+40 > hor[1]*cv0: self.SGFtreeCanv.xview('moveto', (x+50 - x1)*1.0/cv0)

        if y1 < SGFtreeCanvas.UNIT*2:
            self.SGFtreeCanv.yview('moveto', y*1.0/cv1)
        else:
            if y-15 < vert[0]*cv1:
                self.SGFtreeCanv.yview('moveto', (y-25)*1.0/cv1)
            elif y+15 > vert[1]*cv1:
                self.SGFtreeCanv.yview('moveto', (y+25 - y1)*1.0/cv1)
        
        # print 'x: %1.3f, [%1.3f, %1.3f], y: %1.3f, [%1.3f, %1.3f]' % (x*1.0/self.cv0, hor[0], hor[1], \
        #                                                               y*1.0/self.cv1, vert[0], vert[1])
   

# ---------------------------------------------------------------------------------------

class Viewer:
    """ This is the main class of v.py. """

    def convCoord(self, x):
        """ This takes coordinates in SGF style (aa - ss), 
            and returns the corresponding
            integer coordinates (between 1 and 19). """

        try:
            p, q =  ord(x[0])-ord('a')+1, ord(x[1])-ord('a')+1
            if 1 <= p <= 19 and 1 <= q <= 19: return (p,q)
            else: return 0
        except:
            return 0
        

    def splitCollection(self):

        if self.cursor.root.numChildren <= 1: return

        filename = tkFileDialog.asksaveasfilename(filetypes=[('SGF files', '*.sgf'), ('All files', '*')],
                                                  initialdir = self.sgfpath)
        if not filename: return
        self.sgfpath = os.path.split(filename)[0]

        if filename[-4:] == '.sgf': filename = filename[:-4]
        
        i = 0

        n = self.cursor.root.next

        while n:
            try:
                f = open(filename + `i` + '.sgf', 'w')
                try:
                    f.write('(' + self.cursor.outputVar(n) + ')')
                except SGFError:
                    showwarning('Error', 'SGF error in game ' + `i`)
                f.close()
            except IOError:
                showwarning('Error', 'I/O Error when writing ' + filename + `i` + '.sgf')
                break
            n = n.down
            i += 1


    def mirrorSGF(self, flip):
        self.leaveNode()
        self.currentFileChanged()
        self.cursor.symmetry(flip)
        l = self.cursor.currentN.pathToNode()
        self.start(0)
        for i in l:
            self.next(i, 0)

        self.cursor.seeCurrent()


    def setup(self, gameNo = 0, update = 1):
        """ Set up initial position (clear board, put handi stones etc.). """

        if update:
            self.board.clear()
        self.board.delMarks()
        self.board.delLabels()

        if not self.cursor:
            return

        try:
            self.cursor.game(gameNo, update)
        except SGFError:
            showwarning('Error', 'SGF Error in game ' + `gameNo` + ' in current file.')
            return

        try:
            si = self.cursor.getRootNode(gameNo)['SZ'][0]
            if strip(si) != '19':
                showwarning('Error', 'The board size of this game is not 19x19.')
                return
        except:
            pass

        # display game name
        
        gameName = self.currentFile[:15]
        if self.cursor.root.numChildren > 1:
            gameName += '[' + `gameNo` + ']'
        self.gameName.set(gameName)

        self.moveno.set('0')
        self.capW = 0
        self.capB = 0
        self.capVar.set('Cap - B: ' + str(self.capB) + ', W: ' + str(self.capW))

        try:

            if not self.options.doNotChangeWindowTitle.get():
                pw = self.cursor.transcode('PW').replace('\r','').replace('\n','')
                pb = self.cursor.transcode('PB').replace('\r','').replace('\n','')
                self.master.title(pw + ' - ' + pb)

            if self.cursor.currentNode().has_key('AB'):
                for x in self.cursor.currentNode()['AB']:
                    self.board.play(self.convCoord(x), 'black')
            if self.cursor.currentNode().has_key('AW'):
                for x in self.cursor.currentNode()['AW']:
                    self.board.play(self.convCoord(x), 'white')

            if self.cursor.currentNode().has_key('B'):
                self.board.play(self.convCoord(self.cursor.currentNode()['B'][0]), 'black')
            elif self.cursor.currentNode().has_key('W'):
                self.board.play(self.convCoord(self.cursor.currentNode()['B'][0]), 'white')

        except SGFError:
            showwarning('SGF Error', 'SGF Error')
            self.gameName.set('')
            self.currentFile = ''
            self.cursor = None
            return 0
        else:
            self.board.currentColor = 'black'
            self.modeVar.set('blackwhite')
            
        try:
            self.displayLabels(self.cursor.currentNode())
            self.markAll()
        except:
            pass
        return 1


    def markAll(self):
        """ Mark all variations for the next move. """

        if not self.cursor: return
        
        passV = 0

        try:
            if self.options.showCurrMoveVar.get():
                c = self.cursor.currentNode()
                for color in ['B', 'W']:
                    if c.has_key(color) and c[color][0] and self.convCoord(c[color][0]):
                        if color == 'B': self.board.placeMark(self.convCoord(c[color][0]),'', 'white', 'small')
                        else: self.board.placeMark(self.convCoord(c[color][0]), '', 'black', 'small')

            if self.cursor.atEnd: return
        
            for i in range(self.cursor.noChildren()):
                c = self.cursor.next(i,0)

                for color in ['B', 'W']:
                    if c.has_key(color):
                        if c[color][0] and self.convCoord(c[color][0]):
                            if self.options.showNextMoveVar.get():
                                self.board.placeMark(self.convCoord(c[color][0]),'', 'black')
                        else:
                            passV = 1
                        if color == 'B':
                            self.board.currentColor = 'black'
                        else:
                            self.board.currentColor = 'white'

                self.cursor.previous(0)

        except SGFError:
            showwarning('SGF Error', 'SGF Error')
            self.board.delMarks()

        if self.board.currentColor == 'black':
            if self.modeVar.get() == 'white': self.board.currentColor = 'white'
            if self.modeVar.get() == 'whiteblack': self.modeVar.set('blackwhite')
        elif self.board.currentColor == 'white':
            if self.modeVar.get() == 'black': self.board.currentColor = 'black'
            if self.modeVar.get() == 'blackwhite': self.modeVar.set('whiteblack')

    
    def showNextMove(self):
        """ Toggle 'show next move' option. """

        self.board.delMarks()
        self.markAll()
        

    def passFct(self):
        """ React to pass button: choose the 'pass variation' in the SGF file."""

        if not self.cursor: return
        self.leaveNode()

        if self.modeVar.get()[:5] == 'black': nM = 'B'
        else:                                 nM = 'W'

        for i in range(self.cursor.noChildren()):             
            try:
                c = self.cursor.next(i)
            except SGFError:
                continue
            if (c.has_key('B') and not self.convCoord(c['B'][0])) or (c.has_key('W') and not self.convCoord(c['W'][0])):  # found
                self.board.delMarks()
                self.board.delLabels()
                self.moveno.set(str(int(self.moveno.get())+1))
                if self.dataWindow.guessMode.get(): self.guessSuccess()
                self.markAll()
                self.displayNode(c)
                return 1
            else:
                try: self.cursor.previous()
                except SGFError:
                    showwarning('Error', 'Error in SGF file')
                    break
                
        # not found
        
        if self.dataWindow.guessMode.get():
            self.guessFailure('', '')
            return 0

        if self.modeVar.get() == 'blackwhite':
            self.modeVar.set('whiteblack')
        elif self.modeVar.get() == 'whiteblack':
            self.modeVar.set('blackwhite')
                    
        s = ';' + nM + '[]'

        try:
            self.cursor.add(s)
            c = self.cursor.currentNode()
            self.board.delMarks()
            self.board.delLabels()
            self.moveno.set(str(int(self.moveno.get())+1))
            self.displayNode(c)
            self.capVar.set('Cap - B: ' + str(self.capB) + ', W: ' + str(self.capW))
            self.currentFileChanged()
            self.board.currentColor = self.modeVar.get()[:5]
        except SGFError:
            pass

        return 0
        

    def gotoMove(self, event):

        x,y = self.board.getBoardCoord((event.x, event.y), self.board.shadedStoneVar.get())
        if (not x*y): return
        pos = chr(x-1+ord('a')) + chr(y-1+ord('a'))

        try:
            c = self.cursor.currentNode()
        except SGFError:
            showwarning('Error', 'SGF Error')
            return
        
        if (c.has_key('B') and c['B'][0] == pos) or (c.has_key('W') and c['W'][0] == pos):
            return

        found = 0

        if not self.board.status.has_key((x,y)):
            i = 0
            n = self.cursor.currentN

            while n.next:
                i += 1
                n = n.next
                try:
                    c = n.getData()
                except:
                    showwarning('Error', 'SGF Error')
                    return
                if (c.has_key('B') and c['B'][0] == pos) or (c.has_key('W') and c['W'][0] == pos):
                    found = 1
                    break

            if found:
                for j in range(i):
                    self.next(0,0)

                self.cursor.seeCurrent()
                return

        l = self.cursor.currentN.pathToNode()

        i = 0
        n = self.cursor.currentN

        while n.previous:
            i += 1
            n = n.previous
            try:
                c = n.getData()
            except SGFError:
                showwarning('Error', 'SGF Error')
                return
            if (c.has_key('B') and c['B'][0] == pos) or (c.has_key('W') and c['W'][0] == pos):
                found = 1
                break

        if found:
            for j in range(i): self.prev(0)
            self.cursor.seeCurrent()


    def guessSuccess(self):
        self.dataWindow.guessModeCanvas.delete(ALL)
        self.dataWindow.guessModeCanvas.create_rectangle(1,1,79,79, fill='', outline='black')

        self.dataWindow.guessModeCanvas.create_rectangle(10,10,70,70, fill='green')

        self.dataWindow.guessRightWrong[0] += 1

        perc = int(self.dataWindow.guessRightWrong[0] *
                   100.0/(self.dataWindow.guessRightWrong[0] + self.dataWindow.guessRightWrong[1]))
        self.dataWindow.guessModeCanvas.create_text(110, 20, text = `self.dataWindow.guessRightWrong[0]` + '/' + \
                                                    `self.dataWindow.guessRightWrong[0] + self.dataWindow.guessRightWrong[1]`)
        self.dataWindow.guessModeCanvas.create_text(110, 40, text = `perc` + '%')
        
        if self.cursor.atEnd:
            self.dataWindow.guessModeCanvas.create_text(40, 40, text = 'END', font=('Helvetica', 16))


    def guessFailure(self, right, pos):
        self.dataWindow.guessModeCanvas.delete(ALL)
        self.dataWindow.guessModeCanvas.create_rectangle(1,1,79,79, fill='', outline='black')

        if self.cursor.atEnd:
            self.dataWindow.guessModeCanvas.create_rectangle(10,10,70,70, fill='green')
            
            self.dataWindow.guessModeCanvas.create_text(110, 20, text = `self.dataWindow.guessRightWrong[0]` + '/' + \
                                                        `self.dataWindow.guessRightWrong[0] + self.dataWindow.guessRightWrong[1]`)

            try:
                perc = int(self.dataWindow.guessRightWrong[0] *
                           100.0/(self.dataWindow.guessRightWrong[0] + self.dataWindow.guessRightWrong[1]))
                self.dataWindow.guessModeCanvas.create_text(110, 40, text = `perc` + '%')
            except ZeroDivisionError: pass
            self.dataWindow.guessModeCanvas.create_text(40, 40, text = 'END', font=('Helvetica', 16))
            return

        if not right or not pos:
            dx = 50
            dy = 50
            p0 = 40
            p1 = 40
        else:
            dist = int(4 * sqrt((right[0]-pos[0])*(right[0]-pos[0]) + (right[1]-pos[1])*(right[1]-pos[1])))

            dx = max(2, dist/3 + randint(0, dist/3))
            dy = max(2, dist/3 + randint(0, dist/3))

            p0 = right[0] * 4 + 2
            p1 = right[1] * 4 + 2

        self.dataWindow.guessModeCanvas.create_rectangle(p0-dx, p1-dy, min(80, p0+dx), min(80,p1+dx), fill='red', outline = '')

        self.dataWindow.guessRightWrong[1] += 1
        
        perc = int(self.dataWindow.guessRightWrong[0] *
                   100.0/(self.dataWindow.guessRightWrong[0] + self.dataWindow.guessRightWrong[1]))
        self.dataWindow.guessModeCanvas.create_text(110, 20, text = `self.dataWindow.guessRightWrong[0]` + '/' + \
                                                    `self.dataWindow.guessRightWrong[0] + self.dataWindow.guessRightWrong[1]`)
        self.dataWindow.guessModeCanvas.create_text(110, 40, text = `perc` + '%')
     
        
    def nextMove(self, p):
        """ React to mouse-click on the board"""

        self.leaveNode()

        x, y = p

        if self.modeVar.get() in ['black', 'white'] and not self.dataWindow.guessMode.get():
            if self.modeVar.get() == 'black':
                nM = 'AB'
            else:
                nM = 'AW'

            pos = chr(x+ord('a')-1) + chr(y+ord('a')-1)

            try:
                if self.cursor.currentNode().has_key('AB') or self.cursor.currentNode().has_key('AW') \
                   or self.cursor.currentNode().has_key('AE'):
                    if not self.cursor.currentNode().has_key(nM):
                        self.cursor.currentNode()[nM] = []
                        
                    l = self.cursor.currentNode()[nM]
                    if not pos in l:
                        l.append(pos)
                        self.cursor.updateCurrentNode()

                    if not self.board.play(p, self.modeVar.get()):
                        self.board.undostack.append(((20,20), '' ,[]))
                    self.board.currentColor = self.modeVar.get()
                
                    if nM == 'AB':
                        self.capB = self.capB + len(self.board.undostack[len(self.board.undostack)-1][2])
                    if nM == 'AW':
                        self.capW = self.capW + len(self.board.undostack[len(self.board.undostack)-1][2])
                    self.capVar.set('Cap - B: ' + str(self.capB) + ', W: ' + str(self.capW))

                else:
                    s = ';' + nM + '[' + pos + ']'
                    self.cursor.add(s)
                    c = self.cursor.currentNode()
            
                    self.board.delMarks()
                    self.board.delLabels()
            
                    self.moveno.set(str(int(self.moveno.get())+1))

                    self.displayNode(c)

                    if nM == 'AB':
                        self.capB = self.capB + len(self.board.undostack[len(self.board.undostack)-1][2])
                    if nM == 'AW':
                        self.capW = self.capW + len(self.board.undostack[len(self.board.undostack)-1][2])
                    self.capVar.set('Cap - B: ' + str(self.capB) + ', W: ' + str(self.capW))

            except SGFError:
                showwarning('Error', 'SGF error')

            self.currentFileChanged()
                    
            return 0



        if self.modeVar.get() == 'blackwhite':
            nM = 'B'
        elif self.modeVar.get() == 'whiteblack':
            nM = 'W'

        done = 0
        right_pos = '' # (for guess mode)
        for i in range(self.cursor.noChildren()):             # look for the move in the SGF file
            if (not done):
                try:
                    c = self.cursor.next(i,0)
                    if i==0:
                        if c.has_key(nM): right_pos = self.convCoord(c[nM][0])
                except:
                    continue
                if c.has_key(nM) and self.convCoord(c[nM][0])==p:  # found
                    self.cursor.seeCurrent()
                    self.board.delMarks()
                    self.board.delLabels()
                    done = 1
                    if self.dataWindow.guessMode.get(): self.guessSuccess()

                    if self.modeVar.get() == 'blackwhite':
                        self.modeVar.set('whiteblack')
                    elif self.modeVar.get() == 'whiteblack':
                        self.modeVar.set('blackwhite')
                    
                    self.moveno.set(str(int(self.moveno.get())+1))

                    self.displayNode(c)
                    if c.has_key('B'):
                        self.capB = self.capB + len(self.board.undostack[len(self.board.undostack)-1][2])
                    if c.has_key('W'):
                        self.capW = self.capW + len(self.board.undostack[len(self.board.undostack)-1][2])
                    self.capVar.set('Cap - B: ' + str(self.capB) + ', W: ' + str(self.capW))

                else:
                    try:
                        self.cursor.previous(0)
                    except SGFError:
                        showwarning('Error', 'Error in SGF file')
                        break

        

        if not done:
            if self.dataWindow.guessMode.get():
                self.guessFailure(right_pos, p)
                return 0
            
            if self.modeVar.get() == 'blackwhite':
                self.modeVar.set('whiteblack')
            elif self.modeVar.get() == 'whiteblack':
                self.modeVar.set('blackwhite')
                    
            s = ';' + nM + '[' + chr(x+ord('a')-1) + chr(y+ord('a')-1) + ']'

            try:
                self.cursor.add(s)
                c = self.cursor.currentNode()
            
                self.board.delMarks()
                self.board.delLabels()

                self.moveno.set(str(int(self.moveno.get())+1))

                self.displayNode(c)
                
                if nM == 'B':
                    self.capB = self.capB + len(self.board.undostack[len(self.board.undostack)-1][2])
                if nM == 'W':
                    self.capW = self.capW + len(self.board.undostack[len(self.board.undostack)-1][2])
                self.capVar.set('Cap - B: ' + str(self.capB) + ', W: ' + str(self.capW))
            except SGFError:
                showwarning('Error', 'SGF error')

            self.currentFileChanged()

        self.markAll()
                    
        return done


    def prev(self, markCurrent=1):
        """ Go back one move. """

        if not self.cursor.atStart:
            self.leaveNode()

            try:
                c = self.cursor.currentNode()
                if (c.has_key('B') and c['B'][0]) or (c.has_key('W') and c['W'][0]):
                    if c.has_key('B'):
                        self.capB = self.capB - len(self.board.undostack[len(self.board.undostack)-1][2])
                    if c.has_key('W'):
                        self.capW = self.capW - len(self.board.undostack[len(self.board.undostack)-1][2])
                    self.capVar.set('Cap - B: ' + str(self.capB) + ', W: ' + str(self.capW))

                    self.board.undo()

                if c.has_key('AE') and c['AE'][0]:
                    for p in c['AE']:
                        self.board.undo(1, 0)
                if c.has_key('AW') and c['AW'][0]:
                    for p in c['AW']:
                        self.board.undo(1, 0)
                if c.has_key('AB') and c['AB'][0]:
                    for p in c['AB']:
                        self.board.undo(1, 0)

                c = self.cursor.previous(markCurrent)
                self.moveno.set(str(int(self.moveno.get())-1))

                self.board.delLabels()    
                self.board.delMarks()
            
                self.markAll()

                if self.board.currentColor == 'black':
                    self.modeVar.set('blackwhite')
                else:
                    self.modeVar.set('whiteblack')
                
                self.displayLabels(c)
            except: pass
        
    def next(self, n=0, markCurrent=1):
        """Go to the next move."""

        if not self.cursor.atEnd:
            self.leaveNode()

            try:
                c = self.cursor.next(n, markCurrent)
            except SGFError:
                return 0 # failure
            
            self.moveno.set(str(int(self.moveno.get())+1))

            self.board.delMarks()
            self.board.delLabels()

            self.displayNode(c)
            if c.has_key('B') and c['B'][0]:
                self.capB = self.capB + len(self.board.undostack[len(self.board.undostack)-1][2])
            if c.has_key('W') and c['W'][0]:
                self.capW = self.capW + len(self.board.undostack[len(self.board.undostack)-1][2])
            self.capVar.set('Cap - B: ' + str(self.capB) + ', W: ' + str(self.capW))

            self.markAll()

            if self.board.currentColor == 'black':
                self.modeVar.set('blackwhite')
            else:
                self.modeVar.set('whiteblack')
                
            return 1 # success


    def upVariation(self, event):
        if self.cursor.currentN.up:
            i = self.cursor.currentN.level
            self.prev(0)
            self.next(i-1)

    def downVariation(self, event):
        if self.cursor.currentN.down:
            i = self.cursor.currentN.level
            self.prev(0)
            self.next(i+1)


        
    def displayNode(self, c):
        """Display the stones played in the current node,
           and call displayLabels(). """

        if c.has_key('AB') and c['AB'][0]:
            for p in c['AB']:
                if not self.board.play(self.convCoord(p), 'black'):
                    self.board.undostack.append(((20,20), '' ,[]))
                else: self.board.currentColor = self.board.invert(self.board.currentColor)
        if c.has_key('AW') and c['AW'][0]:
            for p in c['AW']:
                if not self.board.play(self.convCoord(p), 'white'):
                    self.board.undostack.append(((20,20), '' ,[]))
                else: self.board.currentColor = self.board.invert(self.board.currentColor)
        if c.has_key('AE') and c['AE'][0]:
            for p in c['AE']:
                if not self.board.remove(self.convCoord(p)):
                    self.board.undostack.append(((20,20), '' ,[]))

        if c.has_key('B') and c['B'][0]:
            p = self.convCoord(c['B'][0])
            if not p or not self.board.play(p, 'black'):
                self.board.undostack.append(((20,20), '' ,[]))
        elif c.has_key('W') and c['W'][0]:
            p = self.convCoord(c['W'][0])
            if not p or not self.board.play(p, 'white'):
                self.board.undostack.append(((20,20), '' ,[]))

        self.displayLabels(c)


    def displayLabels(self, c):
        """ Display the labels in the current node."""

        self.comments.delete('1.0', END)
        if c.has_key('C'):
            self.comments.insert('1.0', replace(self.cursor.transcode('C',c), '\r', ''))

        for type in ['CR', 'MA', 'SQ', 'TR']:
            if c.has_key(type) and c[type][0]:
                for p in c[type]:
                    self.board.placeLabel(self.convCoord(p), type)

        if c.has_key('LB') and c['LB'][0]:
            for p1 in c['LB']:
                p, text = split(p1, ':')
                self.board.placeLabel(self.convCoord(p), 'LB', text)

    def next10(self):
        for i in range(10): self.next(0,0)
        self.cursor.seeCurrent()

    def prev10(self):
        for i in range(10): self.prev(0)
        self.cursor.seeCurrent()
        
    def end(self):
        """ Go to end of game. """
        if not self.cursor: return
        while not self.cursor.atEnd and self.next(0, 0): pass
        self.cursor.seeCurrent()
        
    def start(self, update=1):
        """ Go to beginning of game."""
        if not self.cursor: return
        if update: self.leaveNode()
        self.comments.delete('1.0', END)
        self.board.delMarks()
        self.board.delLabels()
        self.board.currentColor = 'black'
        self.setup(self.cursor.currentGame)
        self.modeVar.set('blackwhite')
        self.cursor.seeCurrent()
        
    def labelClick(self, event):
        x, y = self.board.getBoardCoord((event.x, event.y), 1)
        if not x*y: return
        pos = chr(x+ord('a')-1) + chr(y+ord('a')-1)

        t = self.dataWindow.labelType.get()

        # delete stones

        try:

            if t == 'DEL ST':
                if not self.board.status.has_key((x,y)): return
                if self.cursor.currentNode().has_key('AB') or self.cursor.currentNode().has_key('AW') \
                   or self.cursor.currentNode().has_key('AE'):
                    removed = 0
                    for key in ['AW', 'AB']:
                        if self.cursor.currentNode().has_key(key) and pos in self.cursor.currentNode()[key]:
                            self.cursor.currentNode()[key].remove(pos)
                            removed = 1
                            if not self.cursor.currentNode()[key]: del self.cursor.currentNode()[key]
                    if not removed:
                        if not self.cursor.currentNode().has_key('AE'):
                            self.cursor.currentNode()['AE'] = []
                    
                        self.cursor.currentNode()['AE'].append(pos)

                    self.cursor.updateCurrentNode()
                    
                    self.board.remove((x,y), removed)

                    self.board.currentColor = self.modeVar.get()[:5]
                
                else:
                    s = ';AE[' + pos + ']'
                    self.cursor.add(s)
                    c = self.cursor.currentNode()
                    self.moveno.set(str(int(self.moveno.get())+1))
                    self.displayNode(c)
                
                return

            # labels

            # check if there is a label at current position (and delete it, if so)

            for type in ['TR', 'SQ', 'CR', 'MA']:
                if self.cursor.currentNode().has_key(type) and \
                   pos in self.cursor.currentNode()[type]:
                    self.cursor.currentNode()[type].remove(pos)
                    if not self.cursor.currentNode()[type]: del self.cursor.currentNode()[type]
                    self.cursor.updateCurrentNode()
                             
                    self.board.placeLabel((x,y), type)
                    return
            
            if self.cursor.currentNode().has_key('LB'):
                for item in self.cursor.currentNode()['LB']:
                    if split(item, ':')[0] == pos:
                        self.cursor.currentNode()['LB'].remove(item)
                        if not self.cursor.currentNode()['LB']: del self.cursor.currentNode()['LB'] 
                        self.cursor.updateCurrentNode()
                        self.board.placeLabel((x,y), 'LB', '')
                        return

            if t == '123':
                if self.cursor.currentNode().has_key('LB'):
                    max = '0'
                    for item in self.cursor.currentNode()['LB']:
                        p, t = split(item, ':')
                        if max < t < '9': max = t                
                    text = chr(ord(max)+1)
                    self.cursor.currentNode()['LB'].append(pos+':'+text)
                else:
                    text = '1'
                    self.cursor.currentNode()['LB'] = [pos+':1']

                self.cursor.updateCurrentNode()
                self.board.placeLabel((x,y), 'LB', text)

            
            elif t == 'ABC':
                if self.cursor.currentNode().has_key('LB'):
                    max = '@'
                    for item in self.cursor.currentNode()['LB']:
                        p, t = split(item, ':')
                        if p == pos:
                            self.cursor.currentNode()['LB'].remove(item)
                            self.cursor.updateCurrentNode()
                            self.board.placeLabel((x,y), 'LB', '')
                            return
                        if max < t < 'Z': max = t
                    text = chr(ord(max)+1)
                    self.cursor.currentNode()['LB'].append(pos+':'+text)
                else:
                    text = 'A'
                    self.cursor.currentNode()['LB'] = [pos+':A']

                self.cursor.updateCurrentNode()
                self.board.placeLabel((x,y), 'LB', text)

            elif t == 'abc':
                if self.cursor.currentNode().has_key('LB'):
                    max = chr(ord('a')-1)
                    for item in self.cursor.currentNode()['LB']:
                        p, t = split(item, ':')
                        if p == pos:
                            self.cursor.currentNode()['LB'].remove(item)
                            self.cursor.updateCurrentNode()
                            self.board.placeLabel((x,y), 'LB', '')
                            return
                        if max < t < 'z': max = t                
                    text = chr(ord(max)+1)
                    self.cursor.currentNode()['LB'].append(pos+':'+text)
                else:
                    text = 'a'
                    self.cursor.currentNode()['LB'] = [pos+':a']

                self.cursor.updateCurrentNode()
                self.board.placeLabel((x,y), 'LB', text)

            
            else:
                text = ''

                if self.cursor.currentNode().has_key(t):
                    self.cursor.currentNode()[t].append(pos)
                else:
                    self.cursor.currentNode()[t] = [pos]

                self.cursor.updateCurrentNode()
                self.board.placeLabel((x,y), t, text)

        except SGFError:
            showwarning('Error', 'SGF Error')

        self.currentFileChanged()

        

    def delVar(self):
        try:
            n = self.cursor.currentN
            self.prev()
            self.cursor.delVariation(n)
            
            self.board.delMarks()
            self.markAll()

            self.cursor.updateTree()
        except SGFError: showwarning('Error', 'SGF error')
        self.currentFileChanged()
        
        
    def newGame(self):
        """Add a new game tree to the current collection."""
        self.leaveNode()
        self.comments.delete('1.0', END)
        self.currentFileChanged()

        p = self.cursor.currentN.pathToNode()
        self.cursor.currentN = self.cursor.root
        self.cursor.add(';GM[1]FF[4]SZ[19]AP[Kombilo]')
        self.cursor.currentN.previous = None
        self.dataWindow.gamelist.insert(END, '[' + `self.cursor.root.numChildren-1` + ']')

        # if self.dataWindow.gamelist.list.curselection(): 
        #     index = int(self.dataWindow.gamelist.list.curselection()[0])
        #     self.dataWindow.gamelist.list.select_clear(index)
        # self.dataWindow.gamelist.list.select_set(END)
        # self.dataWindow.gamelist.list.see(END)
        # self.dataWindow.updateGameInfo(self.cursor)

        self.changeCurrentGame(None, self.cursor.root.numChildren-1, p)

    def delGame(self):
        """Add currently selected game from collection."""

        if not self.dataWindow.gamelist.list.curselection(): # should actually never happen!
            return

        self.comments.delete('1.0', END)
        self.currentFileChanged()
        index = int(self.dataWindow.gamelist.list.curselection()[0])

        n = self.cursor.root.next
        for i in range(index): n = n.down

        try:
            self.cursor.delVariation(n)
        except:
            pass
        
        if n.up:
            n.up.down = n.down
        else:
            self.cursor.root.next = n.down
            
        if n.down:
            n.down.up = n.up

        self.cursor.root.numChildren -= 1

        self.dataWindow.gamelist.delete(index)

        d = self.filelist[self.currentFileNum][4]

        if d.has_key(index): del d[index]
        new_d = {}
        for i in d.keys():
            if type(i) != type(0): continue                 # because of 'currentgame' 
            if i > index: new_d[i-1] = d[i]
            else: new_d[i] = d[i]

        self.filelist[self.currentFileNum][4] = new_d
        
        if index == self.cursor.root.numChildren and index != 0:
            index -= 1

        self.cursor.currentGame = -1 # for changeCurrentGame
        
        if self.cursor.root.numChildren == 0:    
            self.newGame()
            return

        self.changeCurrentGame(None, index, None, 1)

        for i in range(index, self.cursor.root.numChildren):
            
            s = '[' + `i` + ']'
            t = self.dataWindow.gamelist.list.get(i)
            l = split(t, ']')
            l[0] = s
            self.dataWindow.gamelist.delete(i)
            self.dataWindow.gamelist.insert(i, join(l, ''))
            
        self.dataWindow.gamelist.list.select_set(index)
        self.dataWindow.gamelist.list.see(index)
        self.dataWindow.updateGameInfo(self.cursor)


    def currentFileChanged(self, removeMark = 0):
        """This method is called, when the current file is changed (then it puts a * before the file name
        of the current SGF file), resp. when the file is saved (then the * is removed). """
        
        index = self.currentFileNum

        s = self.dataWindow.filelist.list.get(index)

        if removeMark:
            if s[0:2] != '* ': return
            self.dataWindow.filelist.delete(index)
            self.dataWindow.filelist.insert(index, s[2:])
            self.dataWindow.filelist.list.select_set(index)
        else:
            if s[0] == '*': return
            self.dataWindow.filelist.delete(index)
            self.dataWindow.filelist.insert(index, '* ' + s)
            self.dataWindow.filelist.list.select_set(index)


    def newFile(self):
        c = EnhancedCursor('(;GM[1]FF[4]SZ[19]AP[Kombilo])', self.comments, self.dataWindow.SGFtreeC, self)

        if self.dataWindow.filelist.list.curselection():
            index = int(self.dataWindow.filelist.list.curselection()[0])
            self.dataWindow.filelist.list.select_clear(index)

        self.filelist.insert(0, ['New', '', (), c, {}])
        self.currentFileNum += 1
        self.dataWindow.filelist.insert(0, 'New')
        self.changeCurrentFile(None, 0)


    def openFile(self, path = None, filename = None, encoding = ''):
        """ Read an SGF file given by filename (if None, ask
            for a filename). """

        self.board.clear()
        self.board.delLabels()
        self.board.delMarks()
        
        self.gameName.set('')
        
        self.board.state('normal', self.nextMove)
        self.master.update()
 
        if not path:
            path = '.'

        if not filename:
            path, filename = os.path.split(tkFileDialog.askopenfilename(filetypes=[('SGF files', '*.sgf'),
                                                                                   ('All files', '*')],
                                                                        initialdir=self.sgfpath))
        if filename:
            try:
                f = open(os.path.join(path,filename))
                s = f.read()
                f.close()
            except IOError:
                showwarning('Open file', 'Cannot open this file\n')
                return 0
            else:
                if not s: return 0
                self.sgfpath = path
                try:
                    c = EnhancedCursor(s, self.comments, self.dataWindow.SGFtreeC, self)

                except SGFError:
                    showwarning('Parsing Error', 'Error in SGF file!')
                    return 0

                if not c.encoding: c.encoding=encoding

                self.dataWindow.filelist.insert(0, filename)
                self.filelist.insert(0, [filename, os.path.join(path,filename), (), c, {}])
                self.currentFileNum+=1
                self.changeCurrentFile(None, 0)



    def delFile(self):
        if not self.dataWindow.filelist.list.curselection(): # should actually never happen!
            return

        index = int(self.dataWindow.filelist.list.curselection()[0])

        s = self.dataWindow.filelist.list.get(index)
        if self.options.confirmDelete.get() and s[0:2] == '* ':
            if not askokcancel('Confirm deletion', 'There are unsaved changes. Delete?'):
                return

        del self.filelist[index]

        self.dataWindow.filelist.list.select_clear(index)
        self.dataWindow.filelist.delete(index)

        if index == len(self.filelist): index -= 1

        if index == -1:                    # file list is empty now, so create a new game
            self.currentFileNum = -2
            self.newFile()
            return 1
        
        self.currentFileNum = -1
        self.changeCurrentFile(None, index)

        return 1
        
    def modeChange(self):
        self.board.currentColor = self.modeVar.get()[:5]


    def leaveNode(self):
        """This method should be called before leaving the current node in the
        SGF file. It will take care of saving the changes (currently that
        means: changes to the comments) to the SGF file."""

        if not self.cursor: return

        s = strip(self.comments.get('1.0', END))

        if self.cursor.encoding:
            try:
                s = s.encode(self.cursor.encoding, 'ignore')
            except:
                pass
        else:
            s = s.encode('utf-8', 'ignore')
        changed = 0

        try:
            d = self.cursor.currentNode()
        except SGFError:
            showwarning('Error', 'SGF Error')
            return
        
        if d.has_key('C'):
            if strip(d['C'][0]) != s:
                d['C'] = [s]
                changed = 1
        else:
            if s:
                d['C'] = [s]
                changed = 1

        if changed:
            self.cursor.updateCurrentNode()
            self.currentFileChanged()



    def changeCurrentGame(self, event, index = -1, p = None, no_leavenode = 0):

        if index == -1:
            index = self.dataWindow.gamelist.list.nearest(event.y)

        if index == -1:
            return

        if not event:
            if not self.dataWindow.gamelist.list.curselection():
                self.dataWindow.gamelist.list.select_set(index)
                self.dataWindow.gamelist.list.see(index)
            elif index != int(self.dataWindow.gamelist.list.curselection()[0]):
                self.dataWindow.gamelist.list.select_clear(self.dataWindow.gamelist.list.curselection())
                self.dataWindow.gamelist.list.select_set(index)
                self.dataWindow.gamelist.list.see(index)

        if 0 <= index <= self.cursor.root.numChildren and index != self.cursor.currentGame:
            filelistIndex = self.currentFileNum
            if self.cursor.currentGame != -1:           # cf. self.changeCurrentFile
                self.filelist[filelistIndex][4][self.cursor.currentGame] = p or self.cursor.currentN.pathToNode()
                                                                         # cf. self.newGame
            if not no_leavenode:
                self.leaveNode()
                self.comments.delete('1.0', END)
            self.setup(index)
            if self.filelist[filelistIndex][4].has_key(index):
                for i in self.filelist[filelistIndex][4][index]:
                    self.next(i,0)
                self.cursor.seeCurrent()

        self.dataWindow.updateGameInfo(self.cursor)



    def changeCurrentFile(self, event, index = -1):

        if index == -1:
            index = self.dataWindow.filelist.list.nearest(event.y)

        if index == -1:
            return

        if not event:
            if not self.dataWindow.filelist.list.curselection():
                self.dataWindow.filelist.list.select_set(index)
                self.dataWindow.filelist.list.see(index)
            elif index != int(self.dataWindow.filelist.list.curselection()[0]):
                self.filelist[int(self.dataWindow.filelist.list.curselection()[0])][4]['currentGame'] = self.cursor.currentGame
                self.dataWindow.filelist.list.select_clear(self.dataWindow.filelist.list.curselection())
                self.dataWindow.filelist.list.select_set(index)
                self.dataWindow.filelist.list.see(index)

        if 0 <= index < len(self.filelist):

            if self.currentFileNum != -1:      # cf. self.delFile
                self.filelist[self.currentFileNum][4]['currentGame'] = self.cursor.currentGame
                self.filelist[self.currentFileNum][4][self.cursor.currentGame] = self.cursor.currentN.pathToNode()

            self.currentFileNum = index
            self.currentFile = self.filelist[index][0]
            self.leaveNode()
            self.comments.delete('1.0', END)
            self.cursor = self.filelist[index][3]
            
            self.cursor.updateGamelist()
            self.cursor.currentGame = -1 # to get around 'index != self.cursor.currentGame' test in self.changeCurrentGame
            if self.filelist[index][4].has_key('currentGame'):
                self.changeCurrentGame(None, self.filelist[index][4]['currentGame'], None, 1)
            else:
                self.changeCurrentGame(None, 0, None, 1)


    def saveSGFfile(self):
        if not self.cursor: return
        
        if not self.dataWindow.filelist.list.curselection(): return
        index = int(self.dataWindow.filelist.list.curselection()[0])
        
        filename = self.filelist[index][1]
        if not filename:
            self.saveasSGFfile()
            return
        
        self.leaveNode()

        try:
            sgf_out = self.cursor.output()
            file = open(filename, 'w')
            file.write(sgf_out)
            file.close()
        except IOError: showwarning('Error', 'I/O Error')
        except SGFError: showwarning('Error', 'SGF error')

        else:
            self.currentFileChanged(1)

        
    def saveasSGFfile(self):
        if not self.cursor: return
        if not self.dataWindow.filelist.list.curselection(): return

        self.leaveNode()

        f = tkFileDialog.asksaveasfilename(filetypes=[('SGF files', '*.sgf'), ('All files', '*')],
                                           initialdir = self.sgfpath)

        if not f: return
        try:
            sgf_out = self.cursor.output() 
            file = open(f, 'w')
            file.write(sgf_out)
            file.close()
        except IOError:
            showwarning('I/O Error', 'Could not write to file ' + f)
        except SGFError:
            showwarning('Error', 'SGF error')

        else:
            self.sgfpath, self.currentFile = os.path.split(f)
            self.gameName.set(self.currentFile[:15])
        
            index = int(self.dataWindow.filelist.list.curselection()[0])
            self.filelist[index][0] = self.currentFile
            self.filelist[index][1] = f
            self.dataWindow.filelist.delete(index)
            self.dataWindow.filelist.insert(index, self.currentFile)
            self.dataWindow.filelist.list.select_set(index)

        self.master.focus()

       
    def quit(self):
        """ Exit the program. """

        for i in range(len(self.filelist)):
            s = self.dataWindow.filelist.list.get(i)
            if self.options.confirmDelete.get() and s[0:2] == '* ':
                if not askokcancel('Confirm deletion', 'There are unsaved changes. Delete?'):
                    return
                else: break

        self.saveOptions()
        self.dataWindow.window.destroy()
        self.master.destroy()


    def saveOptions(self):
        """ Save options to disk (to file 'v.opt'). """
        self.options.windowGeom = StringVar()
        self.options.windowGeom.set(self.master.geometry())

        self.options.dataWindowGeometry = StringVar()
        self.options.dataWindowGeometry.set(self.dataWindow.get_geometry())
        self.options.saveToDisk(os.path.join(self.optionspath,'v.opt'),
                                lambda: showwarning('Save options', 'IO Error'))


    def loadOptions(self):
        """ Load options from disk. """        
        self.options.windowGeom = StringVar()
        self.options.dataWindowGeometry = StringVar()
        self.options.loadFromDisk(os.path.join(self.optionspath,'v.opt'))
        if self.options.windowGeom.get():
            self.master.geometry(self.options.windowGeom.get())
        if self.options.dataWindowGeometry.get():
            self.dataWindow.set_geometry(self.options.dataWindowGeometry.get())
        else:
            self.dataWindow.set_geometry('1|%65|%0|%65|%1|%75|%1|%30|%1|%90|%1|%80|%415x380+10+10')

    def helpAbout(self):
        """ Display the 'About ...' window with some basic information. """
        
        t = 'v.py - written by Ulrich Goertz (u@g0ertz.de)\n\n'
        t = t + 'v.py is a program to display go game records in SGF format.\n'
        t = t + 'It comes together with the go database program Kombilo.\n'
        t = t + 'You can find more information on v.py and Kombilo and the newest '
        t = t + 'version at http://www.g0ertz.de/kombilo/\n\n'
        
        t = t + 'v.py is free software; for more information '
        t = t + 'see the file license.txt.\n\n'
        t = t + 'It is written in Python (see http://www.python.org/). '

        window = Toplevel()
        window.title('About v.py ...')

        text = Text(window, height=15, width=60, relief=FLAT, wrap=WORD)
        text.insert(1.0, t)
 
        text.config(state=DISABLED)
        text.pack()

        b = Button(window, text="OK", command = window.destroy)
        b.pack(side=RIGHT)
        
        window.update_idletasks()
        
        window.focus()
        window.grab_set()
        window.wait_window()


    def helpLicense(self):
        """ Display the GNU General Public License. """
        try:
            t = 'v.py\n (C) Ulrich Goertz (u@g0ertz.de), 2001-2002.\n' 
            t = t + '------------------------------------------------------------------------\n\n'
            file = open(os.path.join(self.basepath,'doc/license.txt'))
            t = t + file.read()
            file.close()
        except IOError:
            t = 'v.py was written by Ulrich Goertz (u@g0ertz.de).\n' 
            t = t + 'It is published under the GNU General Public License. ' 
            t = t + 'See the file doc/license.txt for more information. ' 
            t = t + 'The GPL is also available at http://www.gnu.org/copyleft/gpl.html\n'
            t = t + 'This program is distributed WITHOUT ANY WARRANTY!\n\n'
        self.textWindow(t,'v.py license')

    def textWindow(self, t, title='', grab=1):
        """ Open a window and display the text in the string t.
            The window has the title title, and grabs the focus if grab==1. """
        
        window = Toplevel()
        window.title(title)
        text = ScrolledText(window, height=25, width=80, relief=FLAT, wrap=WORD)
        text.insert(1.0, t)
 
        text.config(state=DISABLED)
        text.pack()

        b = Button(window, text="OK", command = window.destroy)
        b.pack(side=RIGHT)
        
        window.update_idletasks()
        if grab:
            window.focus()
            window.grab_set()
            window.wait_window()

    
    def gameinfoOK(self):
        keylist = ['PB', 'BR', 'PW', 'WR', 'EV', 'RE', 'DT', 'KM']
        for key in keylist:
            value = self.gameinfoVars[key].get()
            if type(value) == type(u''):
                try: value = value.encode(self.gameinfoDictencoding, 'ignore')
                except: pass
            self.gameinfoDict[key] = [value]

        value = strip(self.gameinfoGCText.get('1.0', END))
        if type(value) == type(u''):
            try: value = value.encode(self.gameinfoDictencoding, 'ignore')
            except: pass
        self.gameinfoDict['GC'] = [value]

        for key in keylist + ['GC']:
            if not strip(self.gameinfoDict[key][0]):
                del self.gameinfoDict[key]

        s = self.gameinfoOthersText.get('1.0', END)
        try: s = s.encode(self.gameinfoDictencoding, 'ignore')
        except: pass
        try:
            d = Cursor(('(;' + s + ')'),1).getRootNode(0)
            for k in d.keys():
                self.gameinfoDict[k] = d[k]

            for k in self.gameinfoDict.keys():
                if (not k in keylist + ['GC']) and (not k in d.keys()):
                    del self.gameinfoDict[k]

        except:
            showwarning('SGF Error', "Parse error in 'Other SGF tags'")
        else:
            if not self.returnChanges: self.currentFileChanged()            
            self.gameinfoWindow.destroy()

        
    def gameinfoCancel(self):
        self.gameinfoDict = None
        self.gameinfoWindow.destroy()


    def gameinfo(self, data=None, encoding=''):
        """ Open window with the game info of the current game."""

        if not data and not self.cursor:
            return
        
        if not data:
            try: self.gameinfoDict = self.cursor.getRootNode(self.cursor.currentGame)
            except:
                showwarning('Error', 'SGF Error')
                return
            encoding = self.cursor.encoding
            self.returnChanges = 0
        else:
            self.gameinfoDict = data
            self.returnChanges = 1
            
        window = Toplevel()
        window.transient(self.master)
        window.protocol('WM_DELETE_WINDOW', self.gameinfoCancel)
        self.gameinfoWindow = window
        window.title('Game Info')

        if encoding:
            self.gameinfoDictencoding = encoding
            for k in self.gameinfoDict.keys():
                try:
                    self.gameinfoDict[k] = [x.decode(encoding, 'ignore').encode('utf-8', 'ignore') \
                                            for x in self.gameinfoDict[k]]
                except:
                    pass
        else:
            self.gameinfoDictencoding = 'utf-8'
        
        keylist = ['GC', 'PB', 'BR', 'PW', 'WR', 'EV', 'RE', 'DT', 'KM']

        if self.gameinfoDict.has_key('GC'):
            self.gameinfoDict['GC'][0] = replace(self.gameinfoDict['GC'][0], '\r', '')               # ?!

        self.gameinfoVars = {}
        for key in keylist:
            self.gameinfoVars[key] = StringVar()
            if not (self.gameinfoDict.has_key(key) and self.gameinfoDict[key]):
                self.gameinfoVars[key].set('')
            else:
                self.gameinfoVars[key].set(self.gameinfoDict[key][0])
                
        self.gameinfoVars['others'] = StringVar()
        oth = ''
        for key in self.gameinfoDict.keys():
            if key not in keylist: oth += key + '[' + join([SGFescape(s) for s in self.gameinfoDict[key]], '][') + ']\n'
        self.gameinfoVars['others'].set(oth)

        f = Frame(window)
        f.pack(side=TOP, anchor=W)
        Label(f, text='White:', justify=LEFT).grid(row=0, column=0, sticky=W)
        Entry(f, width = 30, textvariable=self.gameinfoVars['PW']).grid(row=0, column=1)
        Entry(f, width = 5, textvariable=self.gameinfoVars['WR']).grid(row=0, column=2)

        Label(f, text='Black:', justify=LEFT).grid(row=1, column=0, sticky=W)
        Entry(f, width = 30, textvariable=self.gameinfoVars['PB']).grid(row=1, column=1)
        Entry(f, width = 5, textvariable=self.gameinfoVars['BR']).grid(row=1, column=2)

        i = 2
        for key, text in [('EV', 'Event'), ('RE', 'Result'), ('DT', 'Date'), 
                          ('KM', 'Komi')]:
            Label(f, text=text+':').grid(row=i, column=0, sticky=W)
            Entry(f, width = 35, textvariable=self.gameinfoVars[key]).grid(row=i, column=1, columnspan=2, sticky=W+E)
            i += 1
            
        Label(window, text='Game Comment: ').pack(anchor=W)
        self.gameinfoGCText = ScrolledText(window, height=5, width=40, relief=SUNKEN, wrap=WORD)
        self.gameinfoGCText.pack(expand=YES, fill=BOTH)
        self.gameinfoGCText.insert(END, self.gameinfoVars['GC'].get())
        
        Label(window, text='Other SGF tags: ').pack(anchor=W)
        self.gameinfoOthersText = ScrolledText(window, height=5, width=40, relief=FLAT, wrap=WORD)
        self.gameinfoOthersText.pack(expand=YES, fill=BOTH)
        self.gameinfoOthersText.insert(END, self.gameinfoVars['others'].get())
        
        Button(window, text='Cancel', command = self.gameinfoCancel).pack(side=RIGHT)
        Button(window, text="OK", command = self.gameinfoOK).pack(side=RIGHT)
        
        window.update_idletasks()
        window.focus()
        window.grab_set()
        window.wait_window()

        if self.returnChanges:
            return self.gameinfoDict
        elif not self.gameinfoDict is None:
            self.cursor.updateRootNode(self.gameinfoDict, self.cursor.currentGame)
            s = '[' + `self.cursor.currentGame` + '] '

            d = self.gameinfoDict
            s += self.cursor.transcode('PW',d)
            s += ' - '
            s += self.cursor.transcode('PB',d)
            if d.has_key('RE'): s += ' (' + self.cursor.transcode('RE',d) + ')'

            self.dataWindow.gamelist.delete(self.cursor.currentGame)
            self.dataWindow.gamelist.insert(self.cursor.currentGame, s)
            self.dataWindow.gamelist.list.select_set(self.cursor.currentGame)
            self.dataWindow.updateGameInfo(self.cursor)


    def toggleDatawindow(self):
        if self.options.datawindowVisible.get():
            self.dataWindow.window.deiconify()
        else:
            self.dataWindow.window.withdraw()


    def toggleCoordinates(self):
        if self.options.showCoordinates.get():
            self.board.coordinates = 1
            self.board.resize()
        else:
            self.board.coordinates = 0
            self.board.resize()

            
    def initMenus(self): 

        self.options.doNotChangeWindowTitle = IntVar()

        menu = Menu(self.master)
        self.master.config(menu=menu)

        self.filemenu = Menu(menu)
        menu.add_cascade(label='File', underline=0, menu=self.filemenu)
        self.filemenu.add_command(label='Open SGF', underline=0, command=self.openFile)
        self.filemenu.add_command(label='Save SGF', underline=0, command=self.saveSGFfile)
        self.filemenu.add_command(label='Save SGF as', underline=9, command=self.saveasSGFfile)
        
        self.filemenu.add_separator()
        self.filemenu.add_command(label='Exit', underline =1, command=self.quit)
        self.optionsmenu = Menu(menu)
        menu.add_cascade(label='Options', underline=0, menu=self.optionsmenu)

        self.options.fuzzy = self.board.fuzzy    # make 'save options' easy
        self.optionsmenu.add_checkbutton(label='Fuzzy stone placement', underline = 0, 
                                         variable=self.options.fuzzy,
                                         command=self.board.fuzzyStones)

        self.options.shadedStoneVar = self.board.shadedStoneVar
        self.optionsmenu.add_checkbutton(label='Shaded stone mouse pointer', 
                                         variable=self.options.shadedStoneVar)
        self.options.shadedStoneVar.set(1)

        self.options.use3Dstones = self.board.use3Dstones
        self.optionsmenu.add_checkbutton(label='Use 3D stones', variable=self.options.use3Dstones,
                                         command=self.board.resize)

        self.options.showNextMoveVar = IntVar()
        self.optionsmenu.add_checkbutton(label='Show next move', underline=5,
                                         variable = self.options.showNextMoveVar,
                                         command = self.showNextMove)

        self.options.showCurrMoveVar = IntVar()
        self.optionsmenu.add_checkbutton(label='Show last move', underline=5,
                                         variable = self.options.showCurrMoveVar,
                                         command = self.showNextMove)

        self.options.showCoordinates = IntVar()
        self.optionsmenu.add_checkbutton(label='Show coordinates',
                                         variable = self.options.showCoordinates,
                                         command = self.toggleCoordinates)

        self.options.confirmDelete = IntVar()
        self.optionsmenu.add_checkbutton(label='Ask before discarding unsaved changes',
                                         variable = self.options.confirmDelete)
                                         

        self.helpmenu = Menu(menu, name='help')
        menu.add_cascade(label='Help', underline=0, menu=self.helpmenu)

        self.helpmenu.add_command(label='About ...', underline=0, command=self.helpAbout)
        
        self.helpmenu.add_command(label='License', underline=0, command=self.helpLicense)

        self.mainMenu = menu


    def initButtons(self, navFrame, labelFrame):
        # The buttons

        self.nextButton = Button(navFrame, text='->', command=self.next)
        self.boardFrame.bind('<Right>', lambda e, s = self.nextButton: s.invoke())
        self.prevButton = Button(navFrame, text='<-', command=self.prev)                                
        self.boardFrame.bind('<Left>', lambda e, s = self.prevButton: s.invoke())
        self.next10Button = Button(navFrame, text='-> 10', command=self.next10)
        self.boardFrame.bind('<Down>', lambda e, s = self.next10Button: s.invoke())
        self.prev10Button = Button(navFrame, text='<- 10', command=self.prev10)
        self.boardFrame.bind('<Up>', lambda e, s = self.prev10Button: s.invoke())
        self.startButton = Button(navFrame, text='|<-', command=self.start)
        self.boardFrame.bind('<Home>', lambda e, s = self.startButton: s.invoke())
        self.endButton = Button(navFrame, text='->|', command=self.end)
        self.boardFrame.bind('<End>', lambda e, s = self.endButton: s.invoke())

        self.boardFrame.bind('<Prior>', self.upVariation)
        self.boardFrame.bind('<Next>', self.downVariation)
        
        self.passButton = Button(navFrame, text='Pass', command = self.passFct)
        
        self.gameinfoButton = Button(navFrame, text = 'Info', command = self.gameinfo,
                                     underline = 0)
        self.boardFrame.bind('<i>', lambda e, s = self.gameinfoButton: s.invoke())

        self.options.datawindowVisible = IntVar()
        self.options.datawindowVisible.set(1)
        self.datawindowButton = Checkbutton(navFrame, text='DATA', variable = self.options.datawindowVisible,
                                            command = self.toggleDatawindow, indicatoron=0)
        
        # try to load icons for navigation buttons

        if sys.path[0].endswith('library.zip'): SYSPATH = os.path.split(sys.path[0])[0]
        else: SYSPATH = sys.path[0]

        self.basepath = SYSPATH  
        self.sgfpath = os.curdir
        self.optionspath = self.basepath 
        self.currentFile = ''
        
        self.tkImages = []
        for button, filename in [(self.nextButton, 'fwd'), (self.prevButton, 'bwd'),
                                 (self.next10Button, 'ffwd'), (self.prev10Button, 'fbwd'),
                                 (self.startButton, 'start'), (self.endButton, 'end'),
                                 (self.passButton, 'pass'), (self.gameinfoButton, 'info') ]:
            try:
                im = PhotoImage(file=os.path.join(self.basepath, 'gifs', filename + '.gif'))
                button.config(image=im, width=22, height=22)
                self.tkImages.append(im)
            except:
                pass

        self.prevButton.grid(row=0, column=4)
        self.nextButton.grid(row=0, column=5)
        self.prev10Button.grid(row=0, column=6)
        self.next10Button.grid(row=0, column=7)
        self.startButton.grid(row=0, column=8)
        self.endButton.grid(row=0, column=9)
        self.passButton.grid(row=0, column=10)
        self.gameinfoButton.grid(row=0, column=11)
        self.datawindowButton.grid(row=0, column=12)
        
        self.modeVar = StringVar()
        self.modeVar.set('blackwhite')
        self.board.currentColor = 'black'

        if sys.platform[:3] == 'win':
            buttonsize = 20
        else:
            buttonsize = 24

        try:
            self.bwI = PhotoImage(file=os.path.join(self.basepath, 'gifs/bw.gif'))
            self.wbI = PhotoImage(file=os.path.join(self.basepath, 'gifs/wb.gif'))
            self.bI = PhotoImage(file=os.path.join(self.basepath, 'gifs/b.gif'))
            self.wI = PhotoImage(file=os.path.join(self.basepath, 'gifs/w.gif'))

            self.BWbutton = Radiobutton(navFrame, image=self.bwI, indicatoron = 0, width=buttonsize, height=buttonsize,
                                        variable = self.modeVar, value='blackwhite', command=self.modeChange)
            self.WBbutton = Radiobutton(navFrame, image=self.wbI, indicatoron = 0, width=buttonsize, height=buttonsize,
                                        variable = self.modeVar, value='whiteblack', command=self.modeChange)
            self.Bbutton = Radiobutton(navFrame, image=self.bI, indicatoron = 0, width=buttonsize, height=buttonsize,
                                       variable = self.modeVar, value='black', command=self.modeChange)
            self.Wbutton = Radiobutton(navFrame, image=self.wI, indicatoron = 0, width=buttonsize, height=buttonsize,
                                       variable = self.modeVar, value='white', command=self.modeChange)

            im = PhotoImage(file=os.path.join(self.basepath, 'gifs', 'data.gif'))
            self.datawindowButton.config(image=im, width=buttonsize, height=buttonsize)
            self.tkImages.append(im)
            
        except:
            self.BWbutton = Radiobutton(navFrame, text='BW', indicatoron = 0,
                                        variable = self.modeVar, value='blackwhite', command=self.modeChange)
            self.WBbutton = Radiobutton(navFrame, text='WB', indicatoron = 0,
                                        variable = self.modeVar, value='whiteblack', command=self.modeChange)
            self.Bbutton = Radiobutton(navFrame, text='B', indicatoron = 0,
                                       variable = self.modeVar, value='black', command=self.modeChange)
            self.Wbutton = Radiobutton(navFrame, text='W', indicatoron = 0,
                                       variable = self.modeVar, value='white', command=self.modeChange)
        
        self.BWbutton.grid(row=0, column=0)
        self.WBbutton.grid(row=0, column=1)
        self.Bbutton.grid(row=0, column=2)
        self.Wbutton.grid(row=0, column=3)

        self.boardFrame.bind('<q>', lambda e, self = self: self.quit())
        self.boardFrame.focus()
        
        self.moveno = StringVar()
        self.movenoLabel = Label(labelFrame, height=1, width=5, relief=SUNKEN, justify=RIGHT,
                                  textvariable=self.moveno)
        self.gameName = StringVar()
        self.gameNameLabel = Label(labelFrame, height=1, width=20, relief=SUNKEN, justify=LEFT,
                                   textvariable=self.gameName)

        self.capVar = StringVar()
        self.capLabel = Label(labelFrame, height=1, width=15, relief=SUNKEN, justify=LEFT,
                              textvariable=self.capVar)

        # pack everything

        self.gameNameLabel.pack(expand=NO, fill=X, side=LEFT, padx = 5)
        self.movenoLabel.pack(expand=NO, fill=X, side=LEFT, padx=5)
        self.capLabel.pack(expand=NO, fill=X, side=LEFT, padx=5)


    def balloonHelp(self):
        self.balloon = Pmw.Balloon(self.master)
        self.balloon.bind(self.prevButton, 'Back one move')
        self.balloon.bind(self.nextButton, 'Forward one move')
        self.balloon.bind(self.prev10Button, 'Back 10 moves')
        self.balloon.bind(self.next10Button, 'Forward 10 moves')
        self.balloon.bind(self.startButton, 'Start of game/Clear board')
        self.balloon.bind(self.endButton, 'End of game')
        self.balloon.bind(self.passButton, 'Pass')
        self.balloon.bind(self.gameinfoButton, 'Edit game info')
        self.balloon.bind(self.datawindowButton, 'Open/withdraw data window')

        self.balloon.bind(self.BWbutton, 'Play black/white stones')
        self.balloon.bind(self.WBbutton, 'Play white/black stones')
        self.balloon.bind(self.Bbutton, 'Place black stones')
        self.balloon.bind(self.Wbutton, 'Place white stones')

        self.balloon.bind(self.capLabel, 'Number of captured stones')
        self.balloon.bind(self.movenoLabel, 'Number of current move')
        self.balloon.bind(self.gameNameLabel, 'Current SGF file')

        self.balloon.bind(self.dataWindow.filelistB1, 'Create new SGF file')
        self.balloon.bind(self.dataWindow.filelistB2, 'Open SGF file')
        self.balloon.bind(self.dataWindow.filelistB3, 'Delete SGF file from list')
        self.balloon.bind(self.dataWindow.filelistB4, 'Split collection into single files')
        self.balloon.bind(self.dataWindow.gamelistB1, 'Create new game')
        self.balloon.bind(self.dataWindow.gamelistB2, 'Delete game')

        self.balloon.bind(self.dataWindow.removeStoneButton, 'Ctrl-click deletes stone')
        self.balloon.bind(self.dataWindow.triangleButton, 'Ctrl-click places/deletes triangle label')
        self.balloon.bind(self.dataWindow.squareButton, 'Ctrl-click places/deletes square label')
        self.balloon.bind(self.dataWindow.letterUButton, 'Ctrl-click places/deletes uppercase label')
        self.balloon.bind(self.dataWindow.letterLButton, 'Ctrl-click places/deletes lowercase label')
        self.balloon.bind(self.dataWindow.numberButton, 'Ctrl-click places/deletes number label')
        self.balloon.bind(self.dataWindow.delButton, 'Delete this and all following nodes')
        self.balloon.bind(self.dataWindow.mirrorButton1, 'Mirror SGF file vertically')
        self.balloon.bind(self.dataWindow.mirrorButton2, 'Mirror SGF diagonally')
        self.balloon.bind(self.dataWindow.mirrorButton3, 'Rotate SGF by 90 degrees')
        self.balloon.bind(self.dataWindow.guessModeButton, 'Enter/leave guess mode')
        




    def __init__(self, master):
        """ Initialize the GUI, some variables, etc. """

        self.options = BunchTkVar()

        # The main window

        self.master = master

        self.frame = Frame(self.master)
        self.frame.pack(expand=YES, fill=BOTH)

        navFrame = Frame(self.frame)
        self.boardFrame = Frame(self.frame)
        labelFrame = Frame(self.frame)
        
        navFrame.pack(side=TOP)
        self.boardFrame.pack(side=TOP, expand=YES, fill=BOTH)
        labelFrame.pack(side=TOP)

        # The board

        if sys.path[0].endswith('library.zip'): SYSPATH = os.path.split(sys.path[0])[0]
        else: SYSPATH = sys.path[0]

        gifpath = os.path.join(SYSPATH,'gifs')

        try:
            self.boardImg = PhotoImage(file=os.path.join(gifpath, 'board.gif'))
        except (TclError, IOError):
            self.boardImg = None
        try:
            self.blackStone = Image.open(os.path.join(gifpath, 'black.gif')).convert('RGBA')
            self.whiteStone = Image.open(os.path.join(gifpath, 'white.gif')).convert('RGBA')
        except (TclError, IOError, AttributeError):
            self.blackStone = None
            self.whiteStone = None
        
        self.board = Board(self.boardFrame, 19, (25,20), 1, None, 1, None,
                           self.boardImg, self.blackStone, self.whiteStone)

        self.initMenus()
        self.initButtons(navFrame, labelFrame)
        
        self.board.pack(expand=YES, fill=BOTH)
        self.board.pack_propagate(0)
        self.board.update_idletasks()

        self.dataWindow = DataWindow(self)

        self.comments = self.dataWindow.comments
        self.cursor = EnhancedCursor('(;GM[1]FF[4]SZ[19]AP[Kombilo])', self.comments, self.dataWindow.SGFtreeC, self)
        self.dataWindow.updateGameInfo(self.cursor)
        self.dataWindow.window.update()
        
        self.loadOptions()

        self.toggleDatawindow()
        self.master.deiconify()

        self.filelist = []
        self.currentFileNum = -2
        self.newFile()

        if self.options.showCoordinates.get():
            self.board.coordinates = 1
            self.board.resize()
        self.board.state('normal', self.nextMove)
        self.board.bind('<Control-1>', self.labelClick)
        self.board.bind('<Shift-3>', self.gotoMove)
        self.moveno.set(0)
        self.capB, self.capW = 0,0

        self.balloonHelp()


# ---------------------------------------------------------------------------------------

if __name__ == '__main__':

    root = Tk()
    root.withdraw()

    if sys.path[0].endswith('library.zip'): SYSPATH = os.path.split(sys.path[0])[0]
    else: SYSPATH = sys.path[0]
    
    try:
        root.option_readfile(os.path.join(SYSPATH, 'v.app'))
    except TclError:
        showwarning('Error', 'Error reading v.app')
        
    app = Viewer(root)

    root.protocol('WM_DELETE_WINDOW', app.quit)
    root.title('v.py')

    app.boardFrame.focus_force()
    root.tkraise()

    if len(sys.argv) > 1:              # load sgf file given as first argument
        app.openFile(os.path.split(sys.argv[1])[0], os.path.split(sys.argv[1])[1])
        app.changeCurrentFile(None, 1)
        app.delFile()

    if len(sys.argv) > 2:              # jump to move given as second argument
        try:
            for i in range(int(sys.argv[2])):
                app.next()
        except: pass
        
    root.mainloop()


