#! /usr/bin/env python
# File: kombilo.py

##   Copyright (C) 2001-10 Ulrich Goertz (u@g0ertz.de)

##   Kombilo 0.5m is a go database program.

##   This program is free software; you can redistribute it and/or modify
##   it under the terms of the GNU General Public License as published by
##   the Free Software Foundation; either version 2 of the License, or
##   (at your option) any later version.

##   This program is distributed in the hope that it will be useful,
##   but WITHOUT ANY WARRANTY; without even the implied warranty of
##   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##   GNU General Public License for more details.

##   You should have received a copy of the GNU General Public License
##   along with this program (see doc/license.txt); if not, write to the Free Software
##   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##   The GNU GPL is also currently available at
##   http://www.gnu.org/copyleft/gpl.html

from Tkinter import *
from tkMessageBox import *
from ScrolledText import ScrolledText
import tkFileDialog

from tkCommonDialog import Dialog
from tkSimpleDialog import askstring

try:
    import Pmw
except:
    root=Tk()
    t = Text(root, wrap=WORD, width=70, height=12)
    t.pack()

    t.insert(END, 'Kombilo 0.5m\n\n The Python Megawidgets (Pmw) library was not found. It is needed by Kombilo, ' + \
             'so you have to install it before you can use Kombilo. See the Kombilo tutorial at ' + \
             'http://www.u-go.net/kombilo/tutorial#installation or the Pmw web site at ' + \
             'http://pmw.sourceforge.net/ for more information.')
    root.mainloop()
    sys.exit()

try: import Image
except: pass

try:
    import cjkcodecs.aliases
    cjkcodecs_available = 1
except:
    cjkcodecs_available = 0
    
import time
import os
import sys
import cPickle
from copy import copy, deepcopy
from string import split, find, join, strip, replace, digits, maketrans, translate, lower
import glob
import re
from array import *
import webbrowser

from board1 import *
import v

try:
    from sgfpars import Node, Cursor, SGFError, SGFescape
except:
    from sgfparser import Node, Cursor, SGFError, SGFescape
    
try:
    import matchC
    CimportSucceeded = 1
except:
    CimportSucceeded = 0




def getDateOfFile(filename):

    try:
        t = time.localtime(os.stat(filename)[8])
        return time.strftime("%d %b %Y %H:%M", t)
    except:
        return ''

# ---------------------------------------------------------------------------------------

class chooseDirectory(Dialog):
    """ A wrapper tor the Tk chooseDirectory widget. """
    
    command = "tk_chooseDirectory"

    def _fixresult(self, widget, result):
        if result:
            self.options["initialdir"] = result
        self.directory = result
        return result

def askdirectory(**options):
    return apply(chooseDirectory, (), options).show()

#---------------------------------------------------------------------------------------

class BoardWC(Board):
    """ Board with support for wildcards and selection
        of search-relevant region. Furthermore, snapshot returns a dictionary which
        describes the current board position. It can then be restored with restore."""

    
    def __init__(self, master, boardSize, canvasSize, fuzzy, labelFontsize, fixedColor, smartFixedColor,
                 boardImg, blackImg, whiteImg, onlyOneMouseButton = 0):
        Board.__init__(self, master, boardSize, canvasSize, fuzzy, labelFontsize, 1, None,
                       boardImg, blackImg, whiteImg)

        self.wildcards = {}

        self.selection = ((1,1),(19,19))

        self.fixedColor = fixedColor
        self.smartFixedColor = smartFixedColor

        self.onlyOneMouseButton = onlyOneMouseButton
        if onlyOneMouseButton:
            self.bound3 = self.bind('<M2-Button-1>', self.selStart)
            self.bound3m = self.bind('<M2-B1-Motion>', self.selDrag)
        else:
            self.bound3 = self.bind('<Button-3>', self.selStart)
            self.bound3m = self.bind('<B3-Motion>', self.selDrag)

        self.bounds1 = self.bind('<Shift-1>', self.wildcard)

        self.invertSelection = IntVar()


    def rebindMouseButtons(self, onlyOneMouseButton):
        if self.onlyOneMouseButton:
            self.unbind('<M2-B1-Motion>', self.bound3m)
            self.unbind('<M2-Button-1>', self.bound3)
        else:
            self.unbind('<B3-Motion>', self.bound3m)
            self.unbind('<3>', self.bound3)
        self.onlyOneMouseButton = onlyOneMouseButton
        if onlyOneMouseButton:
            self.bound3 = self.bind('<M2-Button-1>', self.selStart)
            self.bound3m = self.bind('<M2-B1-Motion>', self.selDrag)
        else:
            self.bound3 = self.bind('<Button-3>', self.selStart)
            self.bound3m = self.bind('<B3-Motion>', self.selDrag)


    def resize(self, event = None):
        """ Resize the board. Take care of wildcards and selection here. """
        
        Board.resize(self, event)
        for x,y in self.wildcards.keys():
            x1, x2, y1, y2 = self.getPixelCoord((x,y),1)
            if self.canvasSize[1]<=7: margin = 5
            else: margin = 4
            self.wildcards[(x,y)] = self.create_oval(x1+margin, x2+margin, y1-margin, y2-margin, fill = 'green',
                                                     tags=('wildcard','non-bg'))

        self.delete('selection')
        if self.selection != ((1,1),(19,19)) and self.selection[1] != (0,0):
            p0 = self.getPixelCoord(self.selection[0],1)
            p1 = self.getPixelCoord((self.selection[1][0]+1, self.selection[1][1]+1), 1)
            min = self.getPixelCoord((1,1), 1)[0]+1
            max = self.getPixelCoord((self.boardSize+1,self.boardSize+1),1)[1]-1
            if self.canvasSize[1] <= 7:
                self.create_rectangle(p0[0], p0[1], p1[0], p1[1],
                                      tags=('selection', 'non-bg'))
            elif self.invertSelection.get():
                self.create_rectangle(p0[0], p0[1], p1[0], p1[1], fill='brown', stipple='gray50', outline='',
                                      tags='selection')
            else:
                if p0[1] > min:
                    self.create_rectangle(min, min, max, p0[1], fill='brown', stipple='gray50', outline='',
                                          tags='selection')
                if p0[0] > min and p0[1] < max:
                    self.create_rectangle(min, p0[1], p0[0], max, fill='brown', stipple='gray50', outline='',
                                          tags='selection')
                if p1[1] < max:
                    self.create_rectangle(p0[0], p1[1], p1[0], max, fill='brown', stipple='gray50', outline='',
                                          tags='selection')
                if p1[0] < max and p0[1] < max:
                    self.create_rectangle(p1[0], p0[1], max, max, fill='brown', stipple='gray50', outline='',
                                          tags='selection')
            self.tkraise('non-bg')
            
        self.update_idletasks()


    def wildcard(self, event):
        """ Place/delete a wildcard at position of click. """
        
        x, y = self.getBoardCoord((event.x, event.y), 1)
        if not x*y or self.status.has_key((x,y)): return

        if self.wildcards.has_key((x,y)):
            self.delete(self.wildcards[(x,y)])
            del self.wildcards[(x,y)]
        else:
            x1, x2, y1, y2 = self.getPixelCoord((x,y),1)
            if self.canvasSize[1]<=7: margin = 5
            else: margin = 4
            self.wildcards[(x,y)] = self.create_oval(x1+margin, x2+margin, y1-margin, y2-margin, fill = 'green',
                                                     tags=('wildcard','non-bg'))
            self.tkraise('label')
        self.changed.set(1)        


    def delWildcards(self):
        """ Delete all wildcards. """
        
        if self.wildcards: self.changed.set(1)
        self.delete('wildcard')
        self.wildcards = {}


    def placeLabel(self, pos, type, text=None, color=None):
        """ Place a label; take care of wildcards at same position. """
        
        if self.wildcards.has_key(pos): override = ('black', '')
        else: override = None

        Board.placeLabel(self, pos, type, text, color, override)

                              
    # ---- selection of search-relevant section -----------------------------------

    def selStart(self, event):
        """ React to right-click. """
        self.delete('selection')
        x, y = self.getBoardCoord((event.x, event.y), 1)
        x = max(x, 1)
        y = max(y, 1)
        self.selection = ((x,y), (0,0))
        if self.smartFixedColor.get():
            self.fixedColor.set(1)
        self.changed.set(1)


    def selDrag(self, event):
        """ React to right-mouse-key-drag. """
        pos = self.getBoardCoord((event.x, event.y), 1)
        if pos[0] >= self.selection[0][0] and pos[1] >= self.selection[0][1]:
            self.setSelection(self.selection[0], pos)
            

    def setSelection(self, pos0, pos1):
        self.selection = (pos0, pos1)
        self.delete('selection')
        p0 = self.getPixelCoord(pos0,1)
        p1 = self.getPixelCoord((pos1[0]+1, pos1[1]+1), 1)
        min = self.getPixelCoord((1,1), 1)[0]+1
        max = self.getPixelCoord((self.boardSize+1,self.boardSize+1),1)[1]-1
        if self.canvasSize[1] <= 7:
            self.create_rectangle(p0[0], p0[1], p1[0], p1[1],
                                  tags=('selection', 'non-bg'))
        elif self.invertSelection.get():
            self.create_rectangle(p0[0], p0[1], p1[0], p1[1], fill='brown', stipple='gray50', outline='',
                                  tags='selection')
        else:
            if p0[1] > min:
                self.create_rectangle(min, min, max, p0[1], fill='brown', stipple='gray50', outline='',
                                      tags='selection')
            if p0[0] > min and p0[1] < max:
                self.create_rectangle(min, p0[1], p0[0], max, fill='brown', stipple='gray50', outline='',
                                      tags='selection')
            if p1[1] < max:
                self.create_rectangle(p0[0], p1[1], p1[0], max, fill='brown', stipple='gray50', outline='',
                                      tags='selection')
            if p1[0] < max and p0[1] < max:
                self.create_rectangle(p1[0], p0[1], max, max, fill='brown', stipple='gray50', outline='',
                                      tags='selection')
            
        self.tkraise('non-bg')
        
        if self.smartFixedColor.get():
            if self.selection == ((1,1), (19,19)):
                self.fixedColor.set(1)
            else:
                self.fixedColor.set(0)
                
        
    def newPosition(self):
        """ Clear board, selection. """
        self.delete('selection')
        self.clear()
        self.delLabels()
        self.delMarks()
        self.delWildcards()
        self.selection = ((1,1),(19,19))

        if self.smartFixedColor.get():
            self.fixedColor.set(1)


    # ---- snapshot & restore (for 'back' button)

    def snapshot(self):
        """ Return a dictionary which contains the data of all the objects
            currently displayed on the board, which are not stored in the SGF file.
            This means, at the moment: wildcards, and selection. """
        
        data = {}
        data['wildcards'] = copy(self.wildcards)
        data['selection'] = self.selection
        return data

    
    def restore(self, d):
        """ Restore the data from a 'snapshot' dictionary. """
        
        for x,y in d['wildcards'].keys():
            x1, x2, y1, y2 = self.getPixelCoord((x,y),1)
            self.wildcards[(x,y)] = self.create_oval(x1+4, x2+4, y1-4, y2-4, fill = 'green',
                                                     tags=('wildcard','non-bg'))

        if d['selection'] != ((1,1),(19,19)) and d['selection'][1] != (0,0):
            self.setSelection(d['selection'][0], d['selection'][1])


# ---------------------------------------------------------------------------------------

class TextEditor:
    """ A very simple text editor, based on the Tkinter ScrolledText widget.
    You can perform very limited editing, and save the result to a file. """

    def __init__(self, t = '', defpath='', font = None):

        if font is None:
            font = (StringVar(), IntVar(), StringVar())
            font[0].set('Courier')
            font[1].set(10)
            font[2].set('')
            
        self.window = Toplevel()

        self.window.protocol('WM_DELETE_WINDOW', self.quit)

        self.text = ScrolledText(self.window, width=70, height=30,
                                 font=(font[0].get(), font[1].get(), font[2].get()))
        self.text.pack(side=BOTTOM, fill=BOTH, expand=YES)
        self.text.insert(END, t)

        self.buttonFrame = Frame(self.window)
        self.buttonFrame.pack(side=TOP, expand=NO, fill=X)

        Button(self.buttonFrame, text='Quit', command=self.quit).pack(side=RIGHT)
        Button(self.buttonFrame, text='Save as', command=self.saveas).pack(side=RIGHT)

        # self.window.tkraise()
        self.window.focus_force()

        if defpath:
            self.defpath = defpath
        else:
            self.defpath = os.curdir

            
    def saveas(self):
        f = tkFileDialog.asksaveasfilename(initialdir = self.defpath)
        if not f: return
        try:
            file = open(f, 'w')
            file.write(self.text.get('1.0', END).encode('utf-8', 'ignore'))
            file.close()
        except IOError:
            showwarning('IO Error', 'Cannot write to ' + f)

    def quit(self):
        self.window.destroy()


class ESR_TextEditor(TextEditor):
    """The text editor which is used by the exportSearchResults function.
    It adds a button to include the complete game list to the TextEditor."""


    def __init__(self, master, style, t='', defpath='', font=None):
        TextEditor.__init__(self, t, defpath, font)
        self.mster = master
        self.style = style
        
        Button(self.buttonFrame, text='Include game list', command=self.includeGameList).pack(side=LEFT)


    def includeGameList(self):
        if self.style: # wiki
            self.text.insert(END, '\n\n!Game list\n\n' + join(self.mster.gamelist.list.get(0, END), ' %%%\n'))
        else:
            self.text.insert(END, '\n\nGame list\n\n' + join(self.mster.gamelist.list.get(0, END), '\n'))
        

# -------------------------------------------------------------------------------------

class DataWindow(v.DataWindow):

    def __init__(self, master):

        v.DataWindow.__init__(self, master)

        self.prevSF = Pmw.ScrolledFrame(self.prevSearchF, usehullsize=1, hull_width=300, hull_height=135,
                                        hscrollmode='static', vscrollmode='none', vertflex='elastic')
        self.prevSF.pack(expand=YES, fill=X)
        self.prevSV = IntVar()
        self.prevSV.set(1)

        b1 = Checkbutton(self.toolbarF, text = 'History', variable = self.prevSV,
                         command = self.togglePrevSearches, indicatoron=0)
        b1.grid(row=0, column=6)

        self.win.setnaturalsize()

        
    def initPanes(self):
        """ Create the panes in the data window. """

        self.toolbarF = self.win.add(name='toolb', min=28, max=28)
        self.filelistF = self.win.add(name='filel', min=0.01, max=1.0, size=65)
        self.gamelistF = self.win.add(name='gamel', min=0.01, max=1.0, size=65)
        self.gameinfoF = self.win.add(name='gamei', min=0.01, max=1.0, size=75)
        self.editToolsF = self.win.add(name='editt', min=1, max=30, size=30) 
        self.gametreeF = self.win.add(name='gamet', min=0.01, max=1.0, size=90) 
        self.commentsF = self.win.add(name='comm', min=0.01, max=1.0, size=80) 
        self.prevSearchF = self.win.add('prse', min=1, max=1.0, size=135)


    def get_geometry(self):
        """ Return a list of current sizes of the panes. """
        
        l = [ self.filelistV.get(), self.win._size['filel'],
              self.gamelistV.get(), self.win._size['gamel'],
              self.gameinfoV.get(), self.win._size['gamei'],
              self.editToolsV.get(), self.win._size['editt'],
              self.gametreeV.get(), self.win._size['gamet'],
              self.commentsV.get(), self.win._size['comm'],
              self.prevSV.get(), self.win._size['prse']]
        l1 = [ `x` for x in l ]
        l1.append(self.window.geometry())

        return join(l1, '|%')
        
    def set_geometry(self, s):
        """ Reset the sizes of the panes to the given ones. """
        
        l = split(s, '|%')
        
        l1 = [ self.filelistV, 
               self.gamelistV, 
               self.gameinfoV, 
               self.editToolsV,
               self.gametreeV, 
               self.commentsV,
               self.prevSV ]
               
        for i in range(len(l)/2):
            l1[i].set(int(l[2*i]))
            if int(l[2*i]):
                self.win.configurepane(i+1, min=10, max=1.0, size = int(l[2*i+1]))
            else:
                self.win.configurepane(i+1, min=0.0, max=0.0, size = 0.0)
        self.win.updatelayout()
        self.window.geometry(l[-1])


    def gamelistRelease(self, event):
        index1, index2 = v.DataWindow.gamelistRelease(self, event)
        if index1:
            self.mster.prevSearches.exchangeGames(self.mster.cursor, index1, index2)
        

    def togglePrevSearches(self):
        if self.prevSV.get():
            s = self.window.geometry()
            x1, x2, x3 = split(s, '+')
            x, y = split(x1, 'x')
            y = `int(y)+135`
            self.window.geometry('%sx%s+%s+%s' % (x, y, x2, x3))
            self.win.configurepane(7, min=10, max=1.0, size=135)
        else:
            s = self.window.geometry()
            x1, x2, x3 = split(s, '+')
            x, y = split(x1, 'x')
            y = `int(y)-self.win._size['prse']`
            self.window.geometry('%sx%s+%s+%s' % (x, y, x2, x3))
            self.win.configurepane(7, min=0.0, max=0.0, size=0.0)
        self.win.updatelayout()
        

# -------------------------------------------------------------------------------------

class GameList(v.ScrolledList):
    """ This is a scrolled list which shows the game list. All the underlying data
        is contained in self.DBlist, which is a list of dictionaries containing the
        information for the single databases. self.DBlist[i] will contain the keys
        'name': name of the database, i.e. path to the *.db files
        'sgfpath': path to the SGF files
        'data': list of all games in the database.
                This is a list of tuples of the form
                (filename, namelistIndex, PB, PW, result, signature)
        'current': a list of the indices of the games in data which are in the current
                   game list
        'results': for each game in current, the list of matches found in the previous
                   search (If there are lots of matches, this list will consume a lot
                   of memory. Maybe at some time I will get around to fix this ...) 
        'numNewGames': number of games which were "append"ed, but not yet inserted into
                       the list (by addAppendedGames) 
        """
    
    def __init__(self, parent, master, noGamesLabel, winPercLabel, gameinfo):
        v.ScrolledList.__init__(self, parent)
        self.list.config(width=52, height=6) 

        self.onSelectionChange = self.printGameInfo
        self.list.bind('<Button-1>', self.onSelectionChange)

        self.bind('<Up>', self.up)
        self.bind('<Down>', self.down)
        self.bind('<Prior>', self.pgup)
        self.bind('<Next>', self.pgdown)
        self.bind('<Return>', self.handleDoubleClick)
        self.list.bind('<Double-1>', self.handleDoubleClick)
        self.list.bind('<Shift-1>', self.handleShiftClick)
        self.list.bind('<Button-3>', self.rightMouseButton)

        self.mster = master
        
        self.DBlist = []      # list of dicts

        self.noGamesLabel = noGamesLabel
        self.winPercLabel = winPercLabel
        self.gameinfo = gameinfo
        
        self.Bwins = 0
        self.Wwins = 0
        self.Owins = 0 # others: Jigo, Void, Left unfinished, ? (Unknown)

        self.appendClist = []

        self.references = {}

        self.sort = None
        self.gameIndex = []

        self.showFilename = 1
        self.showDate = 0


    def rightMouseButton(self, event):
        
        index = self.list.nearest(event.y)
        if index == -1: return

        DBindex, index = self.getIndex(index)
        if DBindex == -1: return

        # print self.DBlist[DBindex]['data'][self.DBlist[DBindex]['current'][index]][5]
        # return

        f1 = strip(os.path.join(self.DBlist[DBindex]['sgfpath'],
                                self.DBlist[DBindex]['data'][self.DBlist[DBindex]['current'][index]][0]))
        
        if find(f1, '[') != -1:
            f1, f2 = split(f1, '[')
            gameNumber = int(strip(f2)[:-1])
        else:
            gameNumber = 0

        filename = getFilename(f1)

        try:
            file = open(filename)
            sgf = file.read()
            file.close()
            c = Cursor(sgf, 1)
            rootNode = c.getRootNode(gameNumber)
        except IOError:
            showwarning('Error', 'I/O Error')
            return
        except SGFError:
            showwarning('Error', 'SGF error')
            return
        
        backup = copy(rootNode)

        if c.encoding:
            enc = c.encoding
        elif self.DBlist[DBindex]['encoding']:
            enc = self.DBlist[DBindex]['encoding']
        else:
            enc = ''
        newRootNode = self.mster.gameinfo(rootNode, enc)
        if (not newRootNode is None) and backup != newRootNode:
            c.updateRootNode(newRootNode, gameNumber)
            try:
                s = c.output()
                file = open(filename, 'w')
                file.write(s)
                file.close()
            except IOError:
                showwarning('I/O Error', 'Could not write to file ' + filename)


    def handleDoubleClick(self, event):
        """ This is called upon double-clicks."""
        
        index = self.list.curselection()
        if index:
            label = self.list.get(index)
            self.mster.openViewer(index, label)


    def handleShiftClick(self, event):
        index = self.list.nearest(event.y)
        index1 = self.list.curselection()
        if index1: self.list.select_clear(index1[0])
        self.list.select_set(index)
        self.onSelectionChange(event)
        try:
            label = self.list.get(index)
        except:
            return
        self.mster.altOpenViewer([index], label)


    def sortCrit(self, index, c):
        dbIndex, j = self.getIndex(index)
        return self.DBlist[dbIndex]['data'][self.DBlist[dbIndex]['current'][j]][c]


    def sortlist(self, perDB, criterion):
        """ Sort the game list wrt the given criterion, and update it. """

        self.sort = []

        if criterion == 'PW': c = 3
        elif criterion == 'PB': c = 2
        elif criterion == 'DT': c = 6
        else: c = 0          # criterion == 'filename'

        self.appendClist = []

        if perDB and c == 0:
            for i in range(len(self.DBlist)):
                db = self.DBlist[i]
                if db['disabled']: continue
                if db.has_key('sort'): del db['sort']
                self.appendClist.extend([(i, db['current'][j], j) for j in xrange(len(db['current']))])
        elif perDB:
            for i in range(len(self.DBlist)):
                db = self.DBlist[i]
                if db['disabled']: continue
                if db.has_key('sort'): del db['sort']
                
                s = []
                s.extend([(db['data'][j][c], j) for j in range(len(db['data']))])

                s.sort()

                self.sort.extend([(i, j) for dummy, j in s])
        else:
            
            for i in range(len(self.DBlist)):
                db = self.DBlist[i]
                if db['disabled']: continue
                self.sort.extend([(db['data'][j][c], i, j) for j in range(len(db['data']))])

            self.sort.sort()
            
            self.sort = [(i, j) for dummy, i, j in self.sort]

        if not (perDB and c==0):

            for i in range(len(self.DBlist)):
                db = self.DBlist[i]
                db['sort'] = array('I', [0] * len(db['data']))

                for ii in range(len(self.sort)):
                    if self.sort[ii][0] == i: db['sort'][self.sort[ii][1]] = ii
                
                self.appendClist.extend([(i, db['current'][j], j) for j in xrange(len(db['current']))])
            
        self.list.delete(0, END)
        self.Bwins, self.Wwins, self.Owins = 0,0,0
        self.update()


    def getIndex(self, i):
        """ Returns dbIndex, j, such that self.DBlist[dbIndex]['current'][j] corresponds to the
        i-th line of the ScrolledList. """

        if i < len(self.gameIndex):
            return self.gameIndex[i]
        else:
            return -1, -1


    def clear(self):
        """ Clear the list. """

        self.update()

        for db in self.DBlist:
            if not db['disabled']:
                db['current'] = array('I')
                db['results'] = []
            
        self.Bwins = 0
        self.Wwins = 0
        self.Owins = 0 
        self.update()

        self.delete(0,END)
        self.list.update_idletasks()


    def append(self, i, d):
        """ Append an entry to the gamelist. """
        self.DBlist[i]['data'].append(d)

        if not self.DBlist[i]['disabled']: 
            self.DBlist[i]['numNewGames'] += 1

    def addAppendedGames(self):
        """ Display the games which were "append"ed in the list. """
        
        for i in range(len(self.DBlist)):
            db = self.DBlist[i]
            if db['numNewGames']:                              # display append'ed entries
                db['current'].extend(array('I', range(len(db['data'])-db['numNewGames'], len(db['data']))))
                self.gameIndex.extend([(i, j) for j in range(db['numNewGames'])])
            db['results'] = db['results'] + [''] * db['numNewGames']
            db['numNewGames'] = 0

        noOfG = self.noOfGames()
        self.noGamesLabel.config(text = `noOfG` + ' games')

        if noOfG:
            Bperc = self.Bwins * 100.0 / noOfG
            Wperc = self.Wwins * 100.0 / noOfG
            self.winPercLabel.config(text='B: %1.1f%%, W: %1.1f%%' % (Bperc, Wperc))
        else: self.winPercLabel.config(text='')

        self.delete(0,END)
        self.sortlist(self.mster.options.sortPerDatabase.get(), self.mster.options.sortCriterion.get())
        

    def appendC(self, dbIndex, i, res = ''):
        """ Append an entry to self.current """

        db = self.DBlist[dbIndex]
        
        db['current'].append(i)
        db['results'].append(res)
        self.appendClist.append((dbIndex, i, len(db['current'])-1))


    def reset(self):
        """ Reset the list, s.t. it includes all the games from self.data. """
        
        for i in range(len(self.DBlist)):
            db = self.DBlist[i]
            if not db['disabled']:
                db['current'] = array('I', range(len(db['data'])))
                db['results'] = [''] * len(db['data'])
                self.appendClist += [ (i, j, j) for j in range(len(db['data']))]
            else:
                db['current'] = array('I')
                db['results'] = []
        self.delete(0, END)
        self.Bwins, self.Wwins, self.Owins = 0, 0, 0
        self.update()
        self.clearGameInfo()

        
    def update(self):
        """ Put the changes in data, current in effect in self.list. """
        
        if self.appendClist:         # display appendC'ed entries
            
            strl = []
            self.gameIndex = []

            if self.sort:
                self.appendClist = [ (self.DBlist[x[0]]['sort'][x[1]], x) for x in self.appendClist ]
                self.appendClist.sort()
                self.appendClist = map(lambda x: x[1], self.appendClist) 

            if self.mster.options.sortReverse.get():
                self.appendClist.reverse()

            for db, index, index1 in self.appendClist:
                d = self.DBlist[db]['data'][index]
                res = self.DBlist[db]['results'][index1]
                if d[4] == 'B': self.Bwins += 1
                elif d[4] == 'W': self.Wwins += 1
                else: self.Owins += 1

                li = [d[7]]
                if self.showFilename:
                    endFilename = find(d[0], '[')
                    if endFilename == -1: endFilename = len(d[0])
                
                    if d[0][endFilename-1] == '.':
                        filename = d[0][:endFilename-1] + d[0][endFilename:]
                    elif d[0][endFilename-2:endFilename] == '.m':
                        filename = d[0][:endFilename-2] + d[0][endFilename:]
                    else: filename = d[0]

                    li.append(filename + ': ')

                li.append(d[3] + ' - ' + d[2] + ' (' + d[4] + '), ')
                if self.showDate: li.append(d[6]+', ')
                li.append(res)
                strl.append(join(li, ''))
                self.gameIndex.append((db, index1))

            ti = time.time()
            self.insert(END, strl)
            
            self.appendClist = []

        noOfG = self.noOfGames()
        self.noGamesLabel.config(text = `noOfG` + ' games')

        if noOfG:
            Bperc = self.Bwins * 100.0 / noOfG
            Wperc = self.Wwins * 100.0 / noOfG
            self.winPercLabel.config(text='B: %1.1f%%, W: %1.1f%%' % (Bperc, Wperc))
        else: self.winPercLabel.config(text='')
        

    def noOfGames(self):
        return reduce(lambda x,y:x+y, [ len(db['current']) for db in self.DBlist if not db['disabled'] ], 0)


    def printGameInfo(self, event, index = -1):
        """ Print game info of selected game. """

        if index == -1:
            index = self.list.nearest(event.y)
            self.focus()
        if index == -1:
            return

        DBindex, index = self.getIndex(index)
        if DBindex == -1: return
            
        f1 = strip(os.path.join(self.DBlist[DBindex]['sgfpath'],
                                self.DBlist[DBindex]['data'][self.DBlist[DBindex]['current'][index]][0]))

        if find(f1, '[') != -1:
            f1, f2 = split(f1, '[')
            gameNumber = int(strip(f2)[:-1])
        else:
            gameNumber = 0

        filename = getFilename(f1)

        try:
            f = open(filename)
            sgf = f.read()
            c = Cursor(sgf, 1)
            encoding = c.encoding or self.DBlist[DBindex]['encoding']

            f.close()
            node = c.getRootNode(gameNumber)
        except:
            return

        t = ''
        
        if node.has_key('PW'): t = t + node['PW'][0]
        else: t = t + ' ?'
        if node.has_key('WR'): t = t + ', ' + node['WR'][0]

        t = t + ' - '

        if node.has_key('PB'): t = t + node['PB'][0]
        else: t = t + ' ?'
        if node.has_key('BR'): t = t + ', ' + node['BR'][0]

        if node.has_key('RE'): t = t + ', ' + node['RE'][0]
        if node.has_key('KM'): t = t + ' (Komi ' + node['KM'][0] + ')'
        if node.has_key('HA'): t = t + ' (Hcp ' + node['HA'][0] + ')'

        t = t + '\n'

        if node.has_key('EV'): t = t + node['EV'][0] + ', '
        if node.has_key('RO'): t = t + node['RO'][0] + ', '
        if node.has_key('DT'): t = t + node['DT'][0] + '\n'

        if node.has_key('GC'):
            gc = node['GC'][0]
            gc = replace(gc, '\n\r', ' ')
            gc = replace(gc, '\r\n', ' ')
            gc = replace(gc, '\r', ' ')
            gc = replace(gc, '\n', ' ')
            
            t = t + gc

        if encoding:
            try:
                t = t.decode(encoding, 'ignore').encode('utf-8', 'ignore')
            except:
                pass
            

        if self.DBlist[DBindex]['data'][self.DBlist[DBindex]['current'][index]][7]:
            try:
                t2 = 'Commentary in ' +\
                     self.references[self.DBlist[DBindex]['data'][self.DBlist[DBindex]['current'][index]][5]]
            except: t2 = ''
        else: t2 = ''
        
        self.gameinfo.config(state=NORMAL)
        self.gameinfo.delete('1.0', END)
        self.gameinfo.insert('1.0', t)
        if t2:
            if t[-1]!='\n': self.gameinfo.insert(END, '\n')
            self.gameinfo.insert(END, t2, 'blue')
        self.gameinfo.config(state=DISABLED)


    def clearGameInfo(self):
        self.gameinfo.config(state=NORMAL)
        self.gameinfo.delete('1.0', END)
        self.gameinfo.config(state=DISABLED)


    def readDB(self, datap):
        """ Read the database files containing the information for a pattern search. """

        try:
            posTable = array('I')
            mainlistArr = array('B')
            finalpos = array('B')
            hashT = array('i')
            hash = array('B')

            filelist = [(mainlistArr, 'lists'), (posTable, 'posTable'),
                        (finalpos, 'finalpos'), 
                        (hashT, 'hashT'), (hash, 'hash')]

            for var, filename in filelist:
                file = open(os.path.join(datap[0], filename+datap[1]+'.db'), 'rb')
                while 1:
                    try: var.fromfile(file, 1000000)
                    except EOFError: break
                file.close()

            return tuple([x[0] for x in filelist])
                        
        except IOError:
            showwarning('Error', 'Database files not found: ' + datap[0] + ', ' + datap[1])
            return None, None, None, None, None


    def checkDuplicate(self, sig, pos, strict):                                   
        """ Check if a game with Dyer signature sig and final position pos
            occurs among the games in self.data. """
        
        duplist = []
        for db in self.DBlist:
            if db['disabled']: continue
            for g in db['data']:
                if sig == g[5]:
                    if not strict:
                        duplist.append(os.path.join(db['sgfpath'], g[0]))
                    else:
                        mainlistArr, posTable, finalpos, hashT, hash = self.readDB(db['name'])
                        if not finalpos: continue
                        if pos == finalpos[150*g[1]:150*g[1]+100]:
                            duplist.append(os.path.join(db['sgfpath'], g[0]))
        return duplist


    def listOfCurrentSGFFiles(self):
        l = []
        for db in self.DBlist:
            if db['disabled']: continue
            for i in db['current']:
                l.append(os.path.join(db['sgfpath'], getFilename(db['data'][i][0])))
        return l  



# ---------------------------------------------------------------------------------------

class PrevSearchesStack:

    """ This class provides a stack which contains the data of the previous searches,
    s.t. one can return to the previous search with the back button. """


    def __init__(self, maxLength, boardChanged, prevSF, master):
        self.data = []
        self.mster = master
        
        self.maxLength = maxLength
        self.boardChanged = boardChanged

        self.prevSF = prevSF
        self.labelSize = IntVar()
        self.labelSize.set(4)
        self.smartFC = IntVar()
        self.FC = IntVar()

        self.selected = -1
        self.prev_select = -1

        self.popupMenu = None

        
    def append(self, s, data = None):
        
        if not data: return

        if self.mster.options.maxLengthSearchesStack.get() and\
           len(self.data) >= self.mster.options.maxLengthSearchesStack.get():
            for i in range(len(self.data)):
                if not self.data[i][2]:
                    self.delete(i)
                    break
            
        b = BoardWC(self.prevSF.interior(), 19, (9,5), 0, self.labelSize, self.FC, self.smartFC,
                    self.mster.boardImg, None, None)
        b.resizable = 0
        b.pack(side=LEFT, expand=YES, fill=Y)
        b.update_idletasks()
        b.bound1 = b.bind('<1>', lambda event, self=self, l=len(self.data): self.click(l))

        if b.onlyOneMouseButton:
            b.unbind('<M2-B1-Motion>', b.bound3m)
            b.unbind('<M2-Button-1>', b.bound3)
        else:
            b.unbind('<B3-Motion>', b.bound3m)
            b.unbind('<3>', b.bound3)
        b.unbind('<Configure>', b.boundConf)
        b.bound3 = b.bind('<3>', lambda event, self=self, l = len(self.data): self.postMenu(event, l))
        b.unbind('<Shift-1>', b.bounds1)
        
        for key in data[0].keys():
            p = b.getPixelCoord(key)
            b.status[key] = data[0][key]
            b.stones[key] = b.create_oval(p, fill=data[0][key], tags='non-bg')
        for key in data[1].keys():
            x1, x2, y1, y2 = b.getPixelCoord(key,1)
            b.wildcards[key] = b.create_oval(x1+2, x2+2, y1-2, y2-2,
                                             fill = 'green', outline='', tags=('wildcard','non-bg'))

        if data[2] != ((1,1), (19,19)) and data[2][1] != (0,0):
            b.setSelection(data[2][0], data[2][1])

        b.tkraise('non-bg')
        b.resizable = 0

        self.prevSF.reposition()
        self.data.append([s, b, 0])
        if data: self.select(len(self.data)-1)

        
    def postMenu(self, event, board):
        self.popupMenu = Menu(self.mster.dataWindow.window)
        self.popupMenu.config(tearoff=0)
        self.popupMenu.add_command(label = 'Delete', command = lambda self=self,
                                   board=board: self.unpostAndDelete(board))

        if self.data[board][2]:
            self.popupMenu.add_command(label = 'Release', command = lambda self=self,
                                       board=board: self.unpostAndRelease(board))
        else:
            self.popupMenu.add_command(label = 'Hold', command = lambda self=self,
                                      board=board: self.unpostAndHold(board))
            
        self.popupMenu.tk_popup(event.x_root, event.y_root)


    def unpost(self):
        if self.popupMenu:
            self.popupMenu.unpost()
            self.popupMenu = None
        
        
    def unpostAndDelete(self, board):
        self.unpost()
        self.delete(board)


    def unpostAndHold(self, board):
        self.unpost()
        self.data[board][2] = 1
        c = self.data[board][1].getPixelCoord((21,21), 1)[0]
        self.data[board][1].create_rectangle(6,6,c-4,c-4, fill='', outline='blue', width=2, tags='hold')
        

    def unpostAndRelease(self, board):
        self.unpost()
        self.data[board][2] = 0
        self.data[board][1].delete('hold')


    def delete(self, board):
        s, b, dummy = self.data[board]
        del self.data[board]

        if not b: return
        b.delete(ALL)
        b.unbind('<1>', b.bound1)
        b.unbind('<3>', b.bound3)
        b.pack_forget()
        b.destroy()
        
        if board == self.selected: self.selected = -1
        if board == self.prev_select: self.prev_select = -1
        if board < self.selected: self.selected -= 1
        if board < self.prev_select: self.prev_select -= 1
        
        for i in range(len(self.data)):
            s, b, dummy = self.data[i]
            if b:
                b.bound1 = b.bind('<1>', lambda event, self=self, l=i: self.click(l))
                b.bound3 = b.bind('<3>', lambda event, self=self, l=i: self.postMenu(event, l))

        self.prevSF.reposition()


    def deleteFile(self, cursor):
        d = []
        for i in range(len(self.data)):
            if self.data[i][0][-1][0] == cursor:
                d.append(i)
        d.reverse()
        for i in d:
            self.delete(i)


    def deleteGame(self, cursor, game):
        d = []
        for i in range(len(self.data)):
            if self.data[i][0][-1][0] == cursor:
                if self.data[i][0][-1][1] == game :
                    d.append(i)
                elif self.data[i][0][-1][1] > game:
                    self.data[i][0][-1][1] -= 1
        d.reverse()
        for i in d:
            self.delete(i)


    def exchangeGames(self, cursor, index1, index2):
        if index1 < index2:
            for i in range(len(self.data)):
                if self.data[i][0][-1][0] == cursor:
                    if self.data[i][0][-1][1] == index1:
                        self.data[i][0][-1][1] = index2
                    elif index1 < self.data[i][0][-1][1] <= index2:
                        self.data[i][0][-1][1] -= 1
        elif index1 > index2:
            for i in range(len(self.data)):
                if self.data[i][0][-1][0] == cursor:
                    if self.data[i][0][-1][1] == index1:
                        self.data[i][0][-1][1] = index2
                    elif index2 <= self.data[i][0][-1][1] < index1:
                        self.data[i][0][-1][1] += 1


    def deleteNode(self, cursor, game, pathToNode):
        d = []
        for i in range(len(self.data)):
            if self.data[i][0][-1][0] == cursor and self.data[i][0][-1][1] == game:
                j = 0
                p = self.data[i][0][-1][2]
                
                while j < len(p) and j < len(pathToNode) and p[j] == pathToNode[j]:
                    j += 1

                if j == len(pathToNode):
                    d.append(i)

                if j < len(pathToNode) and j < len(p) and p[j] > pathToNode[j]:
                    self.data[i][0][-1][2][j] -= 1
        d.reverse()
        for i in d:
            self.delete(i)
        
        
    def see(self, board):
        if board == END: self.prevSF.xview('moveto', 1.0)
        else:
            self.prevSF.xview('moveto', 1.0 / (len(self.data)+1) * board)


    def select(self, board):
        
        if board == -1 or board >= len(self.data): return

        if not self.data[board][1]: return
        else: b = self.data[board][1]
            
        self.select_clear()
        
        c = b.getPixelCoord((21,21), 1)[0]
        b.create_rectangle(2,2, c-2,c-2, width=3, outline = 'red', tags = 'sel')
        
        self.selected = board
        self.see(board)


    def click(self, board):
        self.select(board)
        self.mster.back(self.data[self.selected][0], self.selected)

        
    def select_clear(self):
        if 0 <= self.selected < len(self.data):
            self.data[self.selected][1].delete('sel')
            self.prev_select = self.selected 
            self.selected = -1


    def pop(self):

        if not self.data or self.selected == 0:
            self.select_clear()
            return None, None

        if self.selected == -1:
            if self.prev_select == -1: return None, None
            else: self.selected = self.prev_select + 1
        else:
            se = self.selected
            self.select_clear()
            self.selected = se
            
        self.selected -= 1

        if self.data[self.selected][1]:
            return (self.data[self.selected][0], self.selected)
        else:
            return (self.data[self.selected][0], -1)


    def clear(self):
        while self.data:
            self.delete(0)
        self.selected = -1
        self.prev_select = -1

# ---------------------------------------------------------------------------------------

class Message(ScrolledText):
    """ A ScrolledText widget which is usually DISABLED (i.e. the user cannot
        enter any text), and which automatically scrolls down upon insertion. """


    def __init__(self, window):
        ScrolledText.__init__(self, window, height=8, width=45, relief=SUNKEN, wrap=WORD)
        self.config(state=DISABLED)


    def insert(self, pos, text):
        self.config(state=NORMAL)
        ScrolledText.insert(self, pos, text)
        self.see(END)
        self.config(state=DISABLED)


    def delete(self, pos1, pos2):
        self.config(state=NORMAL)
        ScrolledText.delete(self, pos1, pos2)
        self.config(state=DISABLED)

# ---------------------------------------------------------------------------------------

class ProgressBar(Canvas):
    """ A simple progress bar. """
    
    def __init__(self, parent, width=390, font = None, colors = None):
        if not font:
            self.font = (StringVar(), IntVar(), StringVar())
            self.font[0].set('Helvetica')
            self.font[1].set(8)
            self.font[2].set('')
        else: self.font = font
        if not colors:
            self.colors = (StringVar(), StringVar())
            self.colors[0].set('black')
            self.colors[1].set('white')
        else: self.colors = colors
        self.width = width
        Canvas.__init__(self, parent, height=15, width=width, highlightthickness=0,
                        borderwidth=2, relief=RIDGE)

        self.current, self.previous = -1, -1
        
    def redraw(self, percent):
        self.delete(ALL)
        x = int(self.width*percent+3)
        self.create_rectangle(0,0,x,18, fill=self.colors[0].get(), outline='')
        self.update()
        

    def clear(self):
        self.delete(ALL)
        
    def write(self, text):
        self.create_text(self.width-10,9, text=text, fill=self.colors[1].get(),
                         font=(self.font[0].get(), self.font[1].get(), self.font[2].get()), anchor=E)
        
# ---------------------------------------------------------------------------------------

class customMenus:

    def __init__(self, master):
        self.mainMenu = master.mainMenu
        self.master = master
        self.htmlpath = os.curdir
        self.windowOpen = 0
        self.path = os.curdir

    def compare(self, entry1, entry2):
        if entry1['name'] < entry2['name']: return -1
        elif entry1['name'] == entry2['name']: return 0
        elif entry1['name'] > entry2['name']: return 1 

    def buildMenus(self, reload=1, alternativePath = ''):
        """
        format of menuList entries:
        { 'name' : name (which will be displayed in the menu),
          'entries': [list of entries]
          'subm' : [list of submenus]  }

        format of entry:
        { 'name' : name which will be displayed in the menu,
          'file' : name of a html file which will be displayed upon clicking
                   this entry (or empty),
          'gisearch' : parameters of a game info search to be done when
                       this is chosen,
          'psearch' : pattern and options for a pattern search }

        format of submenu: {} as in menuList
        """

        if reload:

            fallback = "(lp1\n(dp2\nS'subm'\np3\n(lsS'name'\np4\nS'Fuseki'\nsS'entries'\np5\n(lp6\n(dp7\nS'gisearch'\np8\n(tp9\nsS'reset'\np10\nI1\nsS'psearch'\np11\n(((I10\nI1\ntp12\n(I19\nI19\ntp13\nt(dp14\n(I7\nI15\ntS'.'\ns(I7\nI14\ntS'.'\ns(I7\nI13\ntS'.'\ns(I7\nI12\ntS'.'\ns(I7\nI11\ntS'.'\ns(I7\nI10\ntS'.'\ns(I7\nI9\ntS'.'\ns(I10\nI9\ntS'.'\ns(I2\nI18\ntS'.'\ns(I7\nI18\ntS'.'\ns(I7\nI17\ntS'.'\ns(I7\nI16\ntS'.'\ns(I0\nI18\ntS'.'\ns(I8\nI11\ntS'.'\ns(I0\nI16\ntS'.'\ns(I8\nI9\ntS'.'\ns(I8\nI14\ntS'.'\ns(I8\nI15\ntS'.'\ns(I8\nI12\ntS'.'\ns(I8\nI13\ntS'.'\ns(I8\nI18\ntS'.'\ns(I0\nI11\ntS'.'\ns(I8\nI16\ntS'.'\ns(I0\nI9\ntS'.'\ns(I0\nI14\ntS'.'\ns(I0\nI15\ntS'.'\ns(I0\nI12\ntS'.'\ns(I0\nI13\ntS'.'\ns(I2\nI9\ntS'.'\ns(I2\nI10\ntS'.'\ns(I16\nI10\ntS'.'\ns(I16\nI11\ntS'.'\ns(I16\nI9\ntS'.'\ns(I16\nI14\ntS'.'\ns(I16\nI15\ntS'X'\ns(I16\nI12\ntS'.'\ns(I16\nI13\ntS'.'\ns(I1\nI17\ntS'.'\ns(I1\nI16\ntS'.'\ns(I9\nI11\ntS'.'\ns(I1\nI18\ntS'.'\ns(I9\nI13\ntS'.'\ns(I9\nI12\ntS'.'\ns(I9\nI15\ntS'.'\ns(I9\nI14\ntS'.'\ns(I1\nI9\ntS'.'\ns(I9\nI9\ntS'.'\ns(I1\nI11\ntS'.'\ns(I1\nI10\ntS'.'\ns(I1\nI13\ntS'.'\ns(I1\nI12\ntS'.'\ns(I1\nI15\ntS'.'\ns(I1\nI14\ntS'.'\ns(I16\nI18\ntS'.'\ns(I16\nI16\ntS'.'\ns(I16\nI17\ntS'.'\ns(I9\nI10\ntS'.'\ns(I17\nI9\ntS'.'\ns(I9\nI17\ntS'.'\ns(I17\nI11\ntS'.'\ns(I17\nI10\ntS'.'\ns(I17\nI13\ntS'.'\ns(I17\nI12\ntS'.'\ns(I17\nI15\ntS'.'\ns(I9\nI16\ntS'.'\ns(I10\nI12\ntS'.'\ns(I10\nI13\ntS'.'\ns(I10\nI14\ntS'.'\ns(I10\nI15\ntS'.'\ns(I2\nI16\ntS'.'\ns(I2\nI17\ntS'.'\ns(I10\nI10\ntS'.'\ns(I10\nI11\ntS'.'\ns(I2\nI12\ntS'.'\ns(I2\nI13\ntS'.'\ns(I2\nI14\ntS'.'\ns(I2\nI15\ntS'.'\ns(I10\nI16\ntS'X'\ns(I10\nI17\ntS'.'\ns(I10\nI18\ntS'.'\ns(I2\nI11\ntS'.'\ns(I17\nI17\ntS'.'\ns(I17\nI16\ntS'.'\ns(I17\nI18\ntS'.'\ns(I18\nI14\ntS'.'\ns(I12\nI15\ntS'.'\ns(I11\nI11\ntS'.'\ns(I11\nI10\ntS'.'\ns(I11\nI9\ntS'.'\ns(I18\nI15\ntS'.'\ns(I11\nI15\ntS'.'\ns(I11\nI14\ntS'.'\ns(I11\nI13\ntS'.'\ns(I11\nI12\ntS'.'\ns(I8\nI10\ntS'.'\ns(I11\nI18\ntS'.'\ns(I11\nI17\ntS'.'\ns(I11\nI16\ntS'.'\ns(I18\nI16\ntS'.'\ns(I18\nI17\ntS'.'\ns(I18\nI18\ntS'.'\ns(I0\nI17\ntS'.'\ns(I3\nI11\ntS'.'\ns(I3\nI10\ntS'.'\ns(I3\nI9\ntS'.'\ns(I3\nI15\ntS'X'\ns(I3\nI14\ntS'.'\ns(I3\nI13\ntS'.'\ns(I3\nI12\ntS'.'\ns(I3\nI18\ntS'.'\ns(I3\nI17\ntS'.'\ns(I3\nI16\ntS'.'\ns(I12\nI14\ntS'.'\ns(I0\nI10\ntS'.'\ns(I12\nI12\ntS'.'\ns(I12\nI13\ntS'.'\ns(I12\nI10\ntS'.'\ns(I12\nI11\ntS'.'\ns(I12\nI9\ntS'.'\ns(I8\nI17\ntS'.'\ns(I12\nI18\ntS'.'\ns(I12\nI16\ntS'.'\ns(I12\nI17\ntS'.'\ns(I18\nI12\ntS'.'\ns(I18\nI13\ntS'.'\ns(I4\nI14\ntS'.'\ns(I4\nI15\ntS'.'\ns(I4\nI12\ntS'.'\ns(I4\nI13\ntS'.'\ns(I4\nI10\ntS'.'\ns(I4\nI11\ntS'.'\ns(I4\nI9\ntS'.'\ns(I4\nI18\ntS'.'\ns(I4\nI16\ntS'.'\ns(I4\nI17\ntS'.'\ns(I18\nI9\ntS'.'\ns(I13\nI13\ntS'.'\ns(I13\nI12\ntS'.'\ns(I13\nI15\ntS'.'\ns(I13\nI14\ntS'.'\ns(I13\nI9\ntS'.'\ns(I18\nI10\ntS'.'\ns(I13\nI11\ntS'.'\ns(I13\nI10\ntS'.'\ns(I18\nI11\ntS'.'\ns(I13\nI17\ntS'.'\ns(I13\nI16\ntS'.'\ns(I13\nI18\ntS'.'\ns(I5\nI13\ntS'.'\ns(I5\nI12\ntS'.'\ns(I5\nI15\ntS'.'\ns(I5\nI14\ntS'.'\ns(I5\nI9\ntS'.'\ns(I5\nI11\ntS'.'\ns(I5\nI10\ntS'.'\ns(I5\nI17\ntS'.'\ns(I5\nI16\ntS'.'\ns(I5\nI18\ntS'.'\ns(I14\nI16\ntS'.'\ns(I14\nI17\ntS'.'\ns(I14\nI18\ntS'.'\ns(I14\nI9\ntS'.'\ns(I14\nI10\ntS'.'\ns(I14\nI11\ntS'.'\ns(I14\nI12\ntS'.'\ns(I14\nI13\ntS'.'\ns(I14\nI14\ntS'.'\ns(I14\nI15\ntS'.'\ns(I6\nI9\ntS'.'\ns(I6\nI10\ntS'.'\ns(I6\nI11\ntS'.'\ns(I6\nI12\ntS'.'\ns(I6\nI13\ntS'.'\ns(I6\nI14\ntS'.'\ns(I6\nI15\ntS'.'\ns(I6\nI16\ntS'.'\ns(I6\nI17\ntS'.'\ns(I6\nI18\ntS'.'\ns(I15\nI18\ntS'.'\ns(I15\nI17\ntS'.'\ns(I15\nI16\ntS'.'\ns(I17\nI14\ntS'.'\ns(I9\nI18\ntS'.'\ns(I15\nI15\ntS'.'\ns(I15\nI14\ntS'.'\ns(I15\nI13\ntS'.'\ns(I15\nI12\ntS'.'\ns(I15\nI11\ntS'.'\ns(I15\nI10\ntS'.'\ns(I15\nI9\ntS'.'\nsI0\nI0\nI6\nI0\ntp15\nsg4\nS'Low chinese'\nsS'file'\np16\nS''\nsa(dp17\ng8\n(tp18\nsg10\nI1\nsg11\n(((I1\nI1\nt(I19\nI19\nttp18\n(dp19\n(I10\nI15\ntS'.'\ns(I10\nI8\ntS'.'\ns(I16\nI12\ntS'.'\ns(I17\nI4\ntS'.'\ns(I17\nI7\ntS'.'\ns(I10\nI11\ntS'.'\ns(I8\nI2\ntS'.'\ns(I8\nI3\ntS'.'\ns(I8\nI0\ntS'.'\ns(I8\nI1\ntS'.'\ns(I8\nI6\ntS'.'\ns(I8\nI7\ntS'.'\ns(I8\nI4\ntS'.'\ns(I8\nI5\ntS'.'\ns(I8\nI10\ntS'.'\ns(I8\nI11\ntS'.'\ns(I8\nI8\ntS'.'\ns(I8\nI9\ntS'.'\ns(I8\nI14\ntS'.'\ns(I8\nI15\ntS'.'\ns(I8\nI12\ntS'.'\ns(I8\nI13\ntS'.'\ns(I8\nI18\ntS'.'\ns(I8\nI16\ntS'.'\ns(I8\nI17\ntS'.'\ns(I1\nI17\ntS'.'\ns(I1\nI16\ntS'.'\ns(I1\nI18\ntS'.'\ns(I1\nI9\ntS'.'\ns(I1\nI8\ntS'.'\ns(I1\nI11\ntS'.'\ns(I1\nI10\ntS'.'\ns(I1\nI13\ntS'.'\ns(I1\nI12\ntS'.'\ns(I1\nI15\ntS'.'\ns(I1\nI14\ntS'.'\ns(I1\nI1\ntS'.'\ns(I1\nI0\ntS'.'\ns(I1\nI3\ntS'.'\ns(I1\nI2\ntS'.'\ns(I1\nI5\ntS'.'\ns(I1\nI4\ntS'.'\ns(I1\nI7\ntS'.'\ns(I1\nI6\ntS'.'\ns(I18\nI7\ntS'.'\ns(I10\nI4\ntS'.'\ns(I17\nI8\ntS'.'\ns(I10\nI6\ntS'.'\ns(I10\nI7\ntS'.'\ns(I10\nI0\ntS'.'\ns(I10\nI1\ntS'.'\ns(I10\nI2\ntS'.'\ns(I17\nI14\ntS'.'\ns(I10\nI12\ntS'.'\ns(I9\nI9\ntS'.'\ns(I17\nI3\ntS'.'\ns(I17\nI2\ntS'.'\ns(I17\nI5\ntS'.'\ns(I10\nI9\ntS'.'\ns(I10\nI10\ntS'.'\ns(I17\nI6\ntS'.'\ns(I10\nI16\ntS'.'\ns(I10\nI17\ntS'.'\ns(I10\nI18\ntS'.'\ns(I17\nI17\ntS'.'\ns(I17\nI16\ntS'.'\ns(I17\nI18\ntS'.'\ns(I17\nI9\ntS'.'\ns(I3\nI3\ntS'*'\ns(I3\nI2\ntS'*'\ns(I3\nI1\ntS'.'\ns(I3\nI0\ntS'.'\ns(I3\nI7\ntS'.'\ns(I3\nI6\ntS'.'\ns(I3\nI5\ntS'.'\ns(I3\nI4\ntS'*'\ns(I3\nI11\ntS'.'\ns(I3\nI10\ntS'.'\ns(I3\nI9\ntS'.'\ns(I3\nI8\ntS'.'\ns(I3\nI15\ntS'X'\ns(I3\nI14\ntS'.'\ns(I3\nI13\ntS'.'\ns(I3\nI12\ntS'.'\ns(I3\nI18\ntS'.'\ns(I3\nI17\ntS'.'\ns(I3\nI16\ntS'.'\ns(I12\nI14\ntS'.'\ns(I12\nI15\ntS'.'\ns(I12\nI12\ntS'.'\ns(I12\nI13\ntS'.'\ns(I12\nI10\ntS'.'\ns(I12\nI11\ntS'.'\ns(I12\nI8\ntS'.'\ns(I12\nI9\ntS'.'\ns(I12\nI6\ntS'.'\ns(I12\nI7\ntS'.'\ns(I12\nI4\ntS'.'\ns(I12\nI5\ntS'.'\ns(I12\nI2\ntS'.'\ns(I12\nI3\ntS'.'\ns(I12\nI0\ntS'.'\ns(I12\nI1\ntS'.'\ns(I12\nI18\ntS'.'\ns(I12\nI16\ntS'.'\ns(I12\nI17\ntS'.'\ns(I18\nI0\ntS'.'\ns(I18\nI13\ntS'.'\ns(I18\nI14\ntS'.'\ns(I18\nI15\ntS'.'\ns(I18\nI8\ntS'.'\ns(I18\nI9\ntS'.'\ns(I18\nI10\ntS'.'\ns(I18\nI11\ntS'.'\ns(I16\nI1\ntS'.'\ns(I18\nI4\ntS'.'\ns(I18\nI5\ntS'.'\ns(I18\nI6\ntS'.'\ns(I5\nI5\ntS'.'\ns(I5\nI4\ntS'.'\ns(I5\nI7\ntS'.'\ns(I5\nI6\ntS'.'\ns(I5\nI1\ntS'.'\ns(I5\nI0\ntS'.'\ns(I5\nI3\ntS'.'\ns(I5\nI2\ntS'.'\ns(I5\nI13\ntS'.'\ns(I5\nI12\ntS'.'\ns(I5\nI15\ntS'.'\ns(I5\nI14\ntS'.'\ns(I5\nI9\ntS'.'\ns(I5\nI8\ntS'.'\ns(I5\nI11\ntS'.'\ns(I5\nI10\ntS'.'\ns(I5\nI17\ntS'.'\ns(I5\nI16\ntS'.'\ns(I5\nI18\ntS'.'\ns(I18\nI3\ntS'.'\ns(I14\nI16\ntS'.'\ns(I14\nI17\ntS'.'\ns(I14\nI18\ntS'.'\ns(I14\nI0\ntS'.'\ns(I14\nI1\ntS'.'\ns(I14\nI2\ntS'.'\ns(I14\nI3\ntS'.'\ns(I14\nI4\ntS'.'\ns(I14\nI5\ntS'.'\ns(I14\nI6\ntS'.'\ns(I14\nI7\ntS'.'\ns(I14\nI8\ntS'.'\ns(I14\nI9\ntS'.'\ns(I14\nI10\ntS'.'\ns(I14\nI11\ntS'.'\ns(I14\nI12\ntS'.'\ns(I14\nI13\ntS'.'\ns(I14\nI14\ntS'.'\ns(I14\nI15\ntS'.'\ns(I7\nI15\ntS'.'\ns(I7\nI14\ntS'.'\ns(I7\nI13\ntS'.'\ns(I7\nI12\ntS'.'\ns(I7\nI11\ntS'.'\ns(I7\nI10\ntS'.'\ns(I7\nI9\ntS'.'\ns(I7\nI8\ntS'.'\ns(I7\nI7\ntS'.'\ns(I7\nI6\ntS'.'\ns(I7\nI5\ntS'.'\ns(I7\nI4\ntS'.'\ns(I7\nI3\ntS'.'\ns(I7\nI2\ntS'.'\ns(I7\nI1\ntS'.'\ns(I7\nI0\ntS'.'\ns(I7\nI18\ntS'.'\ns(I7\nI17\ntS'.'\ns(I7\nI16\ntS'.'\ns(I0\nI18\ntS'.'\ns(I0\nI16\ntS'.'\ns(I0\nI17\ntS'.'\ns(I0\nI10\ntS'.'\ns(I0\nI11\ntS'.'\ns(I0\nI8\ntS'.'\ns(I0\nI9\ntS'.'\ns(I0\nI14\ntS'.'\ns(I0\nI15\ntS'.'\ns(I0\nI12\ntS'.'\ns(I0\nI13\ntS'.'\ns(I0\nI2\ntS'.'\ns(I0\nI3\ntS'.'\ns(I0\nI0\ntS'.'\ns(I0\nI1\ntS'.'\ns(I0\nI6\ntS'.'\ns(I0\nI7\ntS'.'\ns(I0\nI4\ntS'.'\ns(I0\nI5\ntS'.'\ns(I9\nI1\ntS'.'\ns(I9\nI0\ntS'.'\ns(I9\nI3\ntS'.'\ns(I9\nI2\ntS'.'\ns(I9\nI5\ntS'.'\ns(I9\nI4\ntS'.'\ns(I16\nI10\ntS'X'\ns(I16\nI11\ntS'.'\ns(I16\nI8\ntS'.'\ns(I16\nI9\ntS'.'\ns(I16\nI14\ntS'.'\ns(I16\nI15\ntS'.'\ns(I9\nI7\ntS'.'\ns(I16\nI13\ntS'.'\ns(I16\nI2\ntS'.'\ns(I9\nI8\ntS'.'\ns(I9\nI11\ntS'.'\ns(I9\nI6\ntS'.'\ns(I9\nI13\ntS'.'\ns(I9\nI12\ntS'.'\ns(I9\nI15\ntS'.'\ns(I9\nI14\ntS'.'\ns(I9\nI17\ntS'.'\ns(I9\nI16\ntS'.'\ns(I9\nI18\ntS'.'\ns(I16\nI3\ntS'.'\ns(I16\nI18\ntS'.'\ns(I16\nI16\ntS'.'\ns(I16\nI17\ntS'.'\ns(I16\nI0\ntS'.'\ns(I9\nI10\ntS'.'\ns(I16\nI6\ntS'.'\ns(I16\nI7\ntS'.'\ns(I16\nI4\ntS'.'\ns(I16\nI5\ntS'X'\ns(I6\nI15\ntS'.'\ns(I2\nI16\ntS'.'\ns(I2\nI17\ntS'.'\ns(I2\nI18\ntS'.'\ns(I2\nI12\ntS'.'\ns(I2\nI13\ntS'.'\ns(I2\nI14\ntS'.'\ns(I2\nI15\ntS'.'\ns(I2\nI8\ntS'.'\ns(I2\nI9\ntS'.'\ns(I2\nI10\ntS'.'\ns(I2\nI11\ntS'.'\ns(I2\nI4\ntS'*'\ns(I2\nI5\ntS'.'\ns(I2\nI6\ntS'.'\ns(I2\nI7\ntS'.'\ns(I2\nI0\ntS'.'\ns(I2\nI1\ntS'.'\ns(I2\nI2\ntS'*'\ns(I2\nI3\ntS'*'\ns(I6\nI5\ntS'.'\ns(I11\nI11\ntS'.'\ns(I11\nI10\ntS'.'\ns(I11\nI9\ntS'.'\ns(I11\nI8\ntS'.'\ns(I11\nI15\ntS'.'\ns(I11\nI14\ntS'.'\ns(I11\nI13\ntS'.'\ns(I11\nI12\ntS'.'\ns(I11\nI3\ntS'.'\ns(I11\nI2\ntS'.'\ns(I11\nI1\ntS'.'\ns(I11\nI0\ntS'.'\ns(I11\nI7\ntS'.'\ns(I11\nI6\ntS'.'\ns(I11\nI5\ntS'.'\ns(I11\nI4\ntS'.'\ns(I11\nI18\ntS'.'\ns(I11\nI17\ntS'.'\ns(I11\nI16\ntS'.'\ns(I18\nI16\ntS'.'\ns(I18\nI17\ntS'.'\ns(I18\nI18\ntS'.'\ns(I6\nI17\ntS'.'\ns(I18\nI2\ntS'.'\ns(I18\nI1\ntS'.'\ns(I4\nI6\ntS'.'\ns(I4\nI7\ntS'.'\ns(I4\nI4\ntS'*'\ns(I4\nI5\ntS'.'\ns(I4\nI2\ntS'*'\ns(I4\nI3\ntS'*'\ns(I4\nI0\ntS'.'\ns(I4\nI1\ntS'.'\ns(I4\nI14\ntS'.'\ns(I4\nI15\ntS'.'\ns(I4\nI12\ntS'.'\ns(I4\nI13\ntS'.'\ns(I4\nI10\ntS'.'\ns(I4\nI11\ntS'.'\ns(I4\nI8\ntS'.'\ns(I4\nI9\ntS'.'\ns(I10\nI13\ntS'.'\ns(I4\nI18\ntS'.'\ns(I4\nI16\ntS'.'\ns(I4\nI17\ntS'.'\ns(I15\nI18\ntS'.'\ns(I13\nI13\ntS'.'\ns(I13\nI12\ntS'.'\ns(I13\nI15\ntS'.'\ns(I13\nI14\ntS'.'\ns(I13\nI9\ntS'.'\ns(I13\nI8\ntS'.'\ns(I13\nI11\ntS'.'\ns(I13\nI10\ntS'.'\ns(I13\nI5\ntS'.'\ns(I13\nI4\ntS'.'\ns(I13\nI7\ntS'.'\ns(I13\nI6\ntS'.'\ns(I13\nI1\ntS'.'\ns(I13\nI0\ntS'.'\ns(I13\nI3\ntS'*'\ns(I13\nI2\ntS'*'\ns(I13\nI17\ntS'.'\ns(I13\nI16\ntS'.'\ns(I13\nI18\ntS'.'\ns(I18\nI12\ntS'.'\ns(I15\nI6\ntS'.'\ns(I6\nI8\ntS'.'\ns(I6\nI9\ntS'.'\ns(I6\nI10\ntS'.'\ns(I6\nI11\ntS'.'\ns(I6\nI12\ntS'.'\ns(I6\nI13\ntS'.'\ns(I6\nI14\ntS'.'\ns(I10\nI5\ntS'.'\ns(I6\nI0\ntS'.'\ns(I6\nI1\ntS'.'\ns(I6\nI2\ntS'.'\ns(I6\nI3\ntS'.'\ns(I6\nI4\ntS'.'\ns(I17\nI11\ntS'.'\ns(I6\nI6\ntS'.'\ns(I6\nI7\ntS'.'\ns(I17\nI10\ntS'.'\ns(I15\nI0\ntS'.'\ns(I6\nI16\ntS'.'\ns(I17\nI13\ntS'.'\ns(I6\nI18\ntS'.'\ns(I17\nI12\ntS'.'\ns(I17\nI15\ntS'.'\ns(I15\nI17\ntS'.'\ns(I15\nI16\ntS'X'\ns(I10\nI3\ntS'.'\ns(I15\nI10\ntS'.'\ns(I15\nI7\ntS'.'\ns(I17\nI1\ntS'.'\ns(I15\nI5\ntS'.'\ns(I15\nI4\ntS'.'\ns(I15\nI3\ntS'O'\ns(I15\nI2\ntS'.'\ns(I15\nI1\ntS'.'\ns(I17\nI0\ntS'.'\ns(I15\nI15\ntS'.'\ns(I15\nI14\ntS'.'\ns(I15\nI13\ntS'.'\ns(I15\nI12\ntS'.'\ns(I15\nI11\ntS'.'\ns(I10\nI14\ntS'.'\ns(I15\nI9\ntS'.'\ns(I15\nI8\ntS'.'\nsI0\nI1\nI250\nI0\ntp20\nsg4\nS'Mini chinese'\nsg16\nS''\nsa(dp21\ng8\n(tp22\nsg10\nI1\nsg11\n((g12\ng13\nt(dp22\n(I7\nI15\ntS'.'\ns(I7\nI14\ntS'.'\ns(I7\nI13\ntS'.'\ns(I7\nI12\ntS'.'\ns(I7\nI11\ntS'.'\ns(I7\nI10\ntS'.'\ns(I7\nI9\ntS'.'\ns(I10\nI9\ntS'.'\ns(I2\nI18\ntS'.'\ns(I7\nI18\ntS'.'\ns(I7\nI17\ntS'.'\ns(I7\nI16\ntS'.'\ns(I0\nI18\ntS'.'\ns(I8\nI11\ntS'.'\ns(I0\nI16\ntS'.'\ns(I8\nI9\ntS'.'\ns(I8\nI14\ntS'.'\ns(I8\nI15\ntS'.'\ns(I8\nI12\ntS'.'\ns(I8\nI13\ntS'.'\ns(I8\nI18\ntS'.'\ns(I0\nI11\ntS'.'\ns(I8\nI16\ntS'.'\ns(I0\nI9\ntS'.'\ns(I0\nI14\ntS'.'\ns(I0\nI15\ntS'.'\ns(I0\nI12\ntS'.'\ns(I0\nI13\ntS'.'\ns(I2\nI9\ntS'.'\ns(I2\nI10\ntS'.'\ns(I16\nI10\ntS'.'\ns(I16\nI11\ntS'.'\ns(I16\nI9\ntS'.'\ns(I16\nI14\ntS'.'\ns(I16\nI15\ntS'.'\ns(I16\nI12\ntS'.'\ns(I16\nI13\ntS'.'\ns(I1\nI17\ntS'.'\ns(I1\nI16\ntS'.'\ns(I9\nI11\ntS'.'\ns(I1\nI18\ntS'.'\ns(I9\nI13\ntS'.'\ns(I9\nI12\ntS'.'\ns(I9\nI15\ntS'X'\ns(I9\nI14\ntS'.'\ns(I1\nI9\ntS'.'\ns(I9\nI9\ntS'.'\ns(I1\nI11\ntS'.'\ns(I1\nI10\ntS'.'\ns(I1\nI13\ntS'.'\ns(I1\nI12\ntS'.'\ns(I1\nI15\ntS'.'\ns(I1\nI14\ntS'.'\ns(I16\nI18\ntS'.'\ns(I16\nI16\ntS'.'\ns(I16\nI17\ntS'.'\ns(I9\nI10\ntS'.'\ns(I17\nI9\ntS'.'\ns(I9\nI17\ntS'.'\ns(I17\nI11\ntS'.'\ns(I17\nI10\ntS'.'\ns(I17\nI13\ntS'.'\ns(I17\nI12\ntS'.'\ns(I17\nI15\ntS'.'\ns(I9\nI16\ntS'.'\ns(I10\nI12\ntS'.'\ns(I10\nI13\ntS'.'\ns(I10\nI14\ntS'.'\ns(I10\nI15\ntS'.'\ns(I2\nI16\ntS'.'\ns(I2\nI17\ntS'.'\ns(I10\nI10\ntS'.'\ns(I10\nI11\ntS'.'\ns(I2\nI12\ntS'.'\ns(I2\nI13\ntS'.'\ns(I2\nI14\ntS'.'\ns(I2\nI15\ntS'.'\ns(I10\nI16\ntS'.'\ns(I10\nI17\ntS'.'\ns(I10\nI18\ntS'.'\ns(I2\nI11\ntS'.'\ns(I17\nI17\ntS'.'\ns(I17\nI16\ntS'.'\ns(I17\nI18\ntS'.'\ns(I18\nI14\ntS'.'\ns(I12\nI15\ntS'.'\ns(I11\nI11\ntS'.'\ns(I11\nI10\ntS'.'\ns(I11\nI9\ntS'.'\ns(I18\nI15\ntS'.'\ns(I11\nI15\ntS'.'\ns(I11\nI14\ntS'.'\ns(I11\nI13\ntS'.'\ns(I11\nI12\ntS'.'\ns(I8\nI10\ntS'.'\ns(I11\nI18\ntS'.'\ns(I11\nI17\ntS'.'\ns(I11\nI16\ntS'.'\ns(I18\nI16\ntS'.'\ns(I18\nI17\ntS'.'\ns(I18\nI18\ntS'.'\ns(I0\nI17\ntS'.'\ns(I3\nI11\ntS'.'\ns(I3\nI10\ntS'.'\ns(I3\nI9\ntS'.'\ns(I3\nI15\ntS'X'\ns(I3\nI14\ntS'.'\ns(I3\nI13\ntS'.'\ns(I3\nI12\ntS'.'\ns(I3\nI18\ntS'.'\ns(I3\nI17\ntS'.'\ns(I3\nI16\ntS'.'\ns(I12\nI14\ntS'.'\ns(I0\nI10\ntS'.'\ns(I12\nI12\ntS'.'\ns(I12\nI13\ntS'.'\ns(I12\nI10\ntS'.'\ns(I12\nI11\ntS'.'\ns(I12\nI9\ntS'.'\ns(I8\nI17\ntS'.'\ns(I12\nI18\ntS'.'\ns(I12\nI16\ntS'.'\ns(I12\nI17\ntS'.'\ns(I18\nI12\ntS'.'\ns(I18\nI13\ntS'.'\ns(I4\nI14\ntS'.'\ns(I4\nI15\ntS'.'\ns(I4\nI12\ntS'.'\ns(I4\nI13\ntS'.'\ns(I4\nI10\ntS'.'\ns(I4\nI11\ntS'.'\ns(I4\nI9\ntS'.'\ns(I4\nI18\ntS'.'\ns(I4\nI16\ntS'.'\ns(I4\nI17\ntS'.'\ns(I18\nI9\ntS'.'\ns(I13\nI13\ntS'.'\ns(I13\nI12\ntS'.'\ns(I13\nI15\ntS'.'\ns(I13\nI14\ntS'.'\ns(I13\nI9\ntS'.'\ns(I18\nI10\ntS'.'\ns(I13\nI11\ntS'.'\ns(I13\nI10\ntS'.'\ns(I18\nI11\ntS'.'\ns(I13\nI17\ntS'.'\ns(I13\nI16\ntS'.'\ns(I13\nI18\ntS'.'\ns(I5\nI13\ntS'.'\ns(I5\nI12\ntS'.'\ns(I5\nI15\ntS'.'\ns(I5\nI14\ntS'.'\ns(I5\nI9\ntS'.'\ns(I5\nI11\ntS'.'\ns(I5\nI10\ntS'.'\ns(I5\nI17\ntS'.'\ns(I5\nI16\ntS'.'\ns(I5\nI18\ntS'.'\ns(I14\nI16\ntS'.'\ns(I14\nI17\ntS'.'\ns(I14\nI18\ntS'.'\ns(I14\nI9\ntS'.'\ns(I14\nI10\ntS'.'\ns(I14\nI11\ntS'.'\ns(I14\nI12\ntS'.'\ns(I14\nI13\ntS'.'\ns(I14\nI14\ntS'.'\ns(I14\nI15\ntS'.'\ns(I6\nI9\ntS'.'\ns(I6\nI10\ntS'.'\ns(I6\nI11\ntS'.'\ns(I6\nI12\ntS'.'\ns(I6\nI13\ntS'.'\ns(I6\nI14\ntS'.'\ns(I6\nI15\ntS'.'\ns(I6\nI16\ntS'.'\ns(I6\nI17\ntS'.'\ns(I6\nI18\ntS'.'\ns(I15\nI18\ntS'.'\ns(I15\nI17\ntS'.'\ns(I15\nI16\ntS'.'\ns(I17\nI14\ntS'.'\ns(I9\nI18\ntS'.'\ns(I15\nI15\ntS'X'\ns(I15\nI14\ntS'.'\ns(I15\nI13\ntS'.'\ns(I15\nI12\ntS'.'\ns(I15\nI11\ntS'.'\ns(I15\nI10\ntS'.'\ns(I15\nI9\ntS'.'\nsI0\nI0\nI5\nI0\ntp23\nsg4\nS'San ren sei'\nsg16\nS''\nsasa(dp24\ng3\n(lsg4\nS'Players'\nsg5\n(lp25\n(dp26\ng8\n(S''\nS''\nS'Go Seigen'\nS''\nS''\nS''\nS''\nI0\ntp27\nsg10\nI1\nsg11\n(tp28\nsg4\nS'Go Seigen'\nsg16\nS''\nsa(dp28\ng8\n(S''\nS''\nS'Shusaku'\nS''\nS'1830'\nS'1880'\nS''\nI0\ntp29\nsg10\nI1\nsg11\n(tp30\nsg4\nS'Shusaku'\nsg16\nS''\nsasa."
            
            try:
                file = open(os.path.join(self.path, 'menus.def'))      
                self.customMenuList = cPickle.load(file)
                file.close()
            except IOError:
                if alternativePath:
                    try:
                        file = open(os.path.join(alternativePath, 'menus.def'))      
                        self.customMenuList = cPickle.load(file)
                        file.close()
                    except IOError:
                        self.customMenuList = cPickle.loads(fallback)
                else:
                    self.customMenuList = cPickle.loads(fallback)
                    
        self.noOfMenus = len(self.customMenuList)
        
        # build menus

        self.customMenuCommands = []
        self.mmIndex = 3
        self.customMenuList.sort(self.compare)

        for item in self.customMenuList:
            m = Menu(self.mainMenu)
            self.mainMenu.insert_cascade(self.mmIndex, menu=m, label = item['name'])
            self.mmIndex += 1
   
            self.addMenu(item['subm'], m)
            item['entries'].sort(self.compare)
            
            for e in item['entries']:
                m.add_command(label=e['name'],
                              command=lambda self=self, i=len(self.customMenuCommands):
                              self.doCommand(i))
                self.customMenuCommands.append((e['file'], e['gisearch'], e['psearch'], e['reset']))


    def addMenu(self, menuList, menu):
        menuList.sort(self.compare)
        
        for item in menuList:
            m = Menu(menu)
            menu.add_cascade(label = item['name'], menu=m)
            self.addMenu(item['subm'], m)
            for e in item['entries']:
                m.add_command(label=e['name'],
                              command=lambda self=self, i=len(self.customMenuCommands):
                              self.doCommand(i))
                self.customMenuCommands.append((e['file'], e['gisearch'], e['psearch'], e['reset']))


    def removeMenus(self):
        for i in range(self.noOfMenus):
            self.mainMenu.delete(3) # The first custom menu is the 3rd overall menu


    def doCommand(self, i):
        if self.customMenuCommands[i][0]:
            try:
                webbrowser.open(self.customMenuCommands[i][0], new=1)
            except:
                showwarning('Error', 'Failed to open the web browser.')
        if self.customMenuCommands[i][3]:   # reset game list
            self.master.reset()
        if self.customMenuCommands[i][2]:   # prepare pattern search
            self.displayPattern(self.customMenuCommands[i][2])
        if self.customMenuCommands[i][1]:   # do game info search
            self.displayGI(self.customMenuCommands[i][1])
            self.master.doGISearch()
        if self.customMenuCommands[i][2]:   # do pattern search                                                             
            self.master.search()


    def displayPattern(self, data):
        if not data: return
        if self.master.dataWindow.filelist.list.curselection():
            index = int(self.master.dataWindow.filelist.list.curselection()[0])
            if self.master.filelist[index][1]:
                self.master.newFile()
        else: self.master.newFile()

        self.master.start()
        
        sel = data[0]
        if sel == ((1,1), (19,19)):
            self.master.board.delete('selection')
            self.master.board.selection = sel
        else:
            self.master.board.setSelection(sel[0], sel[1])
        
        self.master.board.changed.set(1)        

        self.master.fixedAnchorVar.set(data[2])
        self.master.fixedColorVar.set(data[3])
        self.master.moveLimit.set(data[4])
        self.master.nextMoveVar.set(data[5])
        
        d = data[1]
        for ii in range(sel[0][1], sel[1][1]+1):
            for jj in range(sel[0][0], sel[1][0]+1):
                if d[(ii-1,jj-1)] == 'X':
                    nM = 'AB'
                elif d[(ii-1,jj-1)] == 'O':
                    nM = 'AW'
                elif d[(ii-1,jj-1)] == '*':
                    x1, x2, y1, y2 = self.master.board.getPixelCoord((jj,ii),1)
                    self.master.board.wildcards[(jj,ii)] = self.master.board.create_oval(x1+4, x2+4, y1-4, y2-4,
                                                                                         fill = 'green',
                                                                                         tags=('wildcard','non-bg'))
                    continue
                else: continue

                pos = chr(jj+ord('a')-1) + chr(ii+ord('a')-1)

                try:
                    if self.master.cursor.currentNode().has_key('AB') or self.master.cursor.currentNode().has_key('AW') \
                       or self.master.cursor.currentNode().has_key('AE'):
                        if not self.master.cursor.currentNode().has_key(nM):
                            self.master.cursor.currentNode()[nM] = []
                    
                        l = self.master.cursor.currentNode()[nM]
                        if not pos in l:
                            l.append(pos)
                            self.master.cursor.updateCurrentNode()

                        if nM == 'AB': color = 'black'
                        else: color = 'white'
                    
                        if not self.master.board.play((jj,ii), color):
                            self.master.board.undostack.append(((20,20), '' ,[]))
                        self.master.board.currentColor = self.master.modeVar.get()

                        if nM == 'AB':
                            self.master.capB = self.master.capB + \
                                               len(self.master.board.undostack[len(self.master.board.undostack)-1][2])
                        if nM == 'AW':
                            self.master.capW = self.master.capW + \
                                               len(self.master.board.undostack[len(self.master.board.undostack)-1][2])
                        self.master.capVar.set('Cap - B: ' + str(self.master.capB) + ', W: ' + str(self.master.capW))

                    else:
                        s = ';' + nM + '[' + pos + ']'
                        self.master.cursor.add(s)
                        c = self.master.cursor.currentNode()
            
                        self.master.board.delMarks()
                        self.master.board.delLabels()
            
                        self.master.moveno.set(str(int(self.master.moveno.get())+1))

                        self.master.displayNode(c)

                        if nM == 'AB':
                            self.master.capB = self.master.capB + \
                                               len(self.master.board.undostack[len(self.master.board.undostack)-1][2])
                        if nM == 'AW':
                            self.master.capW = self.master.capW + \
                                               len(self.master.board.undostack[len(self.master.board.undostack)-1][2])
                        self.master.capVar.set('Cap - B: ' + str(self.master.capB) + ', W: ' + str(self.master.capW))
                except SGFError: showwarning('Error', 'SGF Error')
                    
        self.master.board.currentColor = self.master.modeVar.get()[:5]


    def c_buildList(self, menuList, level):
        for i in range(len(menuList)):
            item = menuList[i]
            self.list.insert(END, '-' * level + ' ' + item['name'])
            self.correspEntry.append([1, item, '-'*level+' ', menuList, i])                  # 1 = submenu
            self.c_buildList(item['subm'], level+1)
            for j in range(len(item['entries'])):
                entry = item['entries'][j]
                self.list.insert(END, '-' * level + '* ' + entry['name'])
                self.correspEntry.append([2, entry, '-'*level + '* ', item['entries'], j])             # 2 = entry
                


    def count(self, i):
        res = len(i['entries']) + 1
        for j in i['subm']:
            res += self.count(j)
        return res 


    def displayGI(self, data):
        if not data: return
        varList = [self.master.pwVar, self.master.pbVar, self.master.pVar,
                   self.master.evVar, self.master.frVar, self.master.toVar,
                   self.master.awVar, self.master.referencedVar]

        for i in range(len(varList)):
            varList[i].set(data[i])

                
    def c_addEntry(self):
        if not self.current: return
        index = int(self.current[0])
        if self.correspEntry[index][0] != 1: return
        
        # add entry to currently selected submenu, named 'New'

        entry = {'name':'New', 'file':'', 'gisearch':(), 'psearch':(), 'reset':1}
        self.correspEntry[index][1]['entries'].append(entry)

        insertIndex = index + self.count(self.correspEntry[index][1]) - 1
        
        self.correspEntry[insertIndex:insertIndex] = [[2, entry, self.correspEntry[index][2][:-1] + '* ',
                                               self.correspEntry[index][1]['entries'],
                                               len(self.correspEntry[index][1]['entries'])-1]]
        
        self.list.insert(insertIndex, self.correspEntry[index][2][:-1] + '* ' + 'New')
        self.list.list.select_clear(self.list.list.curselection())
        self.list.list.select_set(insertIndex)


    def c_addSubmenu(self):
        if not self.current: return
        index = int(self.current[0])
        if self.correspEntry[index][0] == 2: return
        
        # add submenu to currently selected submenu, named 'New'

        submenu = {'name':'New', 'entries':[], 'subm':[]}

        if self.correspEntry[index][0] == 0:
            self.customMenuList.append(submenu)
            insertIndex = self.list.list.index(END)
            prefix = '- '
            self.correspEntry.append([1, submenu, prefix, self.customMenuList, len(self.customMenuList)-1])
        else:    
            self.correspEntry[index][1]['subm'].append(submenu)

            insertIndex = index + self.count(self.correspEntry[index][1])  \
                          - len(self.correspEntry[index][1]['entries']) - 1

            prefix = self.correspEntry[index][2][:-1] + '- '
        
            self.correspEntry[insertIndex:insertIndex] = [[1, submenu, prefix,
                                                           self.correspEntry[index][1]['subm'],
                                                           len(self.correspEntry[index][1]['subm'])-1]]
        
        self.list.insert(insertIndex, prefix + 'New')
        self.list.list.select_clear(self.list.list.curselection())
        self.list.list.select_set(insertIndex)


    def c_delete(self):

        if not self.current: return
        index = int(self.current[0])
        if self.correspEntry[index][0] == 0: return

        if self.correspEntry[index][0] == 1:
            n = self.count(self.correspEntry[index][1])
        elif self.correspEntry[index][0] == 2:
            n = 1

        del self.correspEntry[index][3][self.correspEntry[index][4]]
        del self.correspEntry[index:index+n]
        
        self.correspEntry = [[0, '', '', None, None]]
        self.list.delete(1,END)
        self.c_buildList(self.customMenuList, 1)
            
        if self.list.list.index(END) > index:
            self.list.list.select_set(index)
        else:
            self.list.list.select_set(END)
        self.current = None



    def c_addPattern(self):
        if not self.current: return
        index = int(self.current[0])
        if self.correspEntry[index][0] != 2: return

        d = {}

        if self.master.board.selection[0][0] > self.master.board.selection[1][0] or \
           self.master.board.selection[0][1] > self.master.board.selection[1][1]:
            self.master.board.selection = ((1,1), (19, 19))

        sel = self.master.board.selection # copy this because the selection on the board may
                                          # be changed by the user although the search is not yet
                                          # finished

        for i in range(sel[0][1], sel[1][1]+1):
            for j in range(sel[0][0], sel[1][0]+1):
                if not self.master.board.status.has_key((j,i)): d[(i-1,j-1)] = '.'
                elif self.master.board.status[(j,i)] == 'black': d[(i-1,j-1)] = 'X'
                else:                                    d[(i-1,j-1)] = 'O'
                if self.master.board.wildcards.has_key((j,i)): d[(i-1,j-1)] = '*'

        self.correspEntry[index][1]['psearch'] = (sel, d,
                                                  self.master.fixedAnchorVar.get(),
                                                  self.master.fixedColorVar.get(),
                                                  self.master.moveLimit.get(),
                                                  self.master.nextMoveVar.get())


        showinfo('Add pattern', 'Successfully stored pattern.')



    def c_addGI(self):
        if not self.current: return
        index = int(self.current[0])
        if self.correspEntry[index][0] != 2: return

        self.correspEntry[index][1]['gisearch'] = (self.master.pwVar.get(), self.master.pbVar.get(), self.master.pVar.get(),
                                                   self.master.evVar.get(), self.master.frVar.get(), self.master.toVar.get(),
                                                   self.master.awVar.get(), self.master.referencedVar.get())

        showinfo('Add game info', 'Successfully stored game info.')


    def c_OK(self):
        if self.current:
            index = int(self.current[0])
            if self.correspEntry[index][0] == 1:
                entry = self.correspEntry[index][1]
                entry['name'] = self.nameCurrent.get()

            if self.correspEntry[index][0] == 2:
                entry = self.correspEntry[index][1]
                entry['name'] = self.nameCurrent.get()
                entry['file'] = self.htmlCurrent.get()
                entry['reset'] = self.resetVar.get()
        self.removeMenus()
        self.buildMenus(0)

        # save menus ...

        try:
            file = open(os.path.join(self.path, 'menus.def'), 'wb') 
            cPickle.dump(self.customMenuList, file)
            file.close()
        except IOError:
            showwarning('I/O Error', 'Could not save custom menu file.')
        
        self.window.destroy()

    def c_cancel(self):
        self.removeMenus()
        self.buildMenus()      
        self.window.destroy()

    def c_setName(self, event=None):
        index = int(self.current[0])
        self.list.delete(index)
        self.correspEntry[index][1]['name'] = self.nameCurrent.get()
        self.list.insert(index, self.correspEntry[index][2] + self.correspEntry[index][1]['name'])
        # self.htmlE.focus()
        # self.list.list.select_set(index)


    def c_browse(self, event=None):
        if not self.current: return
        index = int(self.current[0])
        if self.correspEntry[index][0] != 2: return

        filename = tkFileDialog.askopenfilename(filetypes=[('HTML files', ('*.html', '*.htm')), ('All files', '*')],
                                                initialdir = self.htmlpath)

        self.window.tkraise()

        if filename:
            filename = os.path.normpath(filename)
            self.htmlpath = os.path.split(filename)[0]
            self.correspEntry[index][1]['file'] = filename
            self.htmlCurrent.set(filename)

        
    def pollList(self):
        now = self.list.list.curselection()
        if not now and self.current:
            self.list.list.select_set(self.current)
        elif now != self.current:
            if self.current:
                index = int(self.current[0])
                if self.correspEntry[index][0] == 1:
                    entry = self.correspEntry[index][1]
                    self.c_setName(None)
                    self.list.list.select_set(now)
                    entry['name'] = self.nameCurrent.get()
                if self.correspEntry[index][0] == 2:
                    entry = self.correspEntry[index][1]
                    self.c_setName(None)
                    self.list.list.select_set(now)
                    entry['file'] = self.htmlCurrent.get()
                    entry['reset'] = self.resetVar.get()
                    
            self.nameCurrent.set('')
            self.htmlCurrent.set('')
            self.resetVar.set(0)
            for widget in [self.addEntryB, self.addSubmenuB, self.deleteB,
                           self.nameE, self.htmlE, self.browseB,
                           self.patternB, self.giB, self.resetB]: widget.config(state=DISABLED)
            self.nameE.config(takefocus=0)
            self.htmlE.config(takefocus=0)
        
            if now:
                self.list.list.see(now)
                index = int(now[0])
                if self.correspEntry[index][0] == 0:
                    for widget in [self.addSubmenuB]:
                        widget.config(state=NORMAL)
                elif self.correspEntry[index][0] == 1:
                    for widget in [self.addEntryB, self.addSubmenuB, self.deleteB, self.nameE]:
                        widget.config(state=NORMAL)
                    self.nameE.config(takefocus=1)
                    self.nameCurrent.set(self.correspEntry[index][1]['name'])
                elif self.correspEntry[index][0] == 2:
                    for widget in [self.deleteB, self.nameE, self.htmlE, self.browseB,
                                   self.patternB, self.giB, self.resetB]:
                        widget.config(state=NORMAL)

                    self.nameE.config(takefocus=1)
                    self.htmlE.config(takefocus=1)

                    self.nameCurrent.set(self.correspEntry[index][1]['name'])
                    self.htmlCurrent.set(self.correspEntry[index][1]['file'])
                    self.resetVar.set(self.correspEntry[index][1]['reset'])

                    self.displayPattern(self.correspEntry[index][1]['psearch'])
                    self.displayGI(self.correspEntry[index][1]['gisearch'])
                        
                    
            self.current = now

        self.window.after(250, self.pollList)

        
    def change(self):
        if self.windowOpen: return
        else: self.windowOpen = 1
        
        self.window = Toplevel()
        self.window.title('Change custom menus')
        self.window.protocol('WM_DELETE_WINDOW', self.c_cancel)
        
        # scrolled list with all submenus, entries

        self.list = v.ScrolledList(self.window, width = 35, font='Courier')
        self.list.pack()
        self.current = None
        
        self.correspEntry = [[0, '', '', None, None]]
        self.list.insert(END, '*')
        self.c_buildList(self.customMenuList, 1)

        # buttons ... for currently selected entry

        f = Frame(self.window)
        self.addEntryB = Button(f, text='Add entry', command = self.c_addEntry)
        self.addSubmenuB = Button(f, text='Add submenu', command = self.c_addSubmenu)
        self.deleteB = Button(f, text='Delete', command = self.c_delete)
        self.addEntryB.pack(side=LEFT, anchor=W)
        self.addSubmenuB.pack(side=LEFT, anchor=W)
        self.deleteB.pack(side=LEFT, anchor=W)
        f.pack(anchor=W)

        Frame(self.window, background='black', height=1, width=100).pack(expand=YES, fill=X, pady=10)

        Label(self.window, text='Name:').pack(anchor=W)

        self.nameCurrent = StringVar()
        self.nameE = Entry(self.window, width=40, textvariable=self.nameCurrent, takefocus=1)
        self.nameE.bind('<Return>', self.c_setName)
        self.nameE.bind('<Tab>', self.c_setName)
        self.nameE.pack(anchor=W)

        Frame(self.window, height=15, width=200).pack()

        Label(self.window, text='HTML file:').pack(anchor=W)

        f = Frame(self.window)
        f.pack(anchor=W)
        self.htmlCurrent = StringVar()
        self.htmlE = Entry(f, width=35, textvariable=self.htmlCurrent, takefocus=1)
        self.htmlE.pack(side=LEFT)

        self.browseB = Button(f, text='Browse ...', command = self.c_browse, height=1)
        self.browseB.pack(side=RIGHT)

        Frame(self.window, height=15, width=200).pack()
        
        f = Frame(self.window)
        self.patternB = Button(f, text='Add pattern info', command = self.c_addPattern)
        self.patternB.pack(side=LEFT)
        self.giB = Button(f, text='Add game info', command = self.c_addGI)
        self.giB.pack(side=LEFT)

        self.resetVar = IntVar()
        self.resetB = Checkbutton(f, text='Reset game list', highlightthickness=0,
                                  variable = self.resetVar)
        self.resetB.pack(side=LEFT)
        f.pack(anchor=W)
        
        # OK, cancel buttons

        Frame(self.window, background='black', height=1, width=100).pack(expand=YES, fill=X, pady=10)

        f = Frame(self.window)
        f.pack(side=BOTTOM, anchor=E)
        Button(f, text='Cancel', command = self.c_cancel).pack(side=RIGHT, anchor=E)
        Button(f, text='OK', command = self.c_OK).pack(side=RIGHT, anchor=E)
        
        self.list.list.select_set(0)
        self.pollList()

        self.window.update_idletasks()  
        self.window.focus()
        # self.window.grab_set()
        self.window.wait_window()

        del self.window
        self.windowOpen = 0

# --------------------------------------------------------------------------------------

class Pattern:
    """ A pattern on the go board, which is basically a matrix with coefficients in
        { 'X', 'O', '.', '*' } (black, white, empty wildcard), plus an 'anchor
        rectangle', given by the top left and bottom right points.

        patternList gives back a list of all rotated/mirrored/cloro-reversed patterns
        of the current pattern. """

    def patternList(self, fixedColor = 0, nextMove = 0):                      
        
        l = []
        lCS = []
        symmetries = []
        special = -1
        
        for ii in range(len(Pattern.flips)):
            f = Pattern.flips[ii]
            
            newA = []

            a = [(self.anchors[0][0],self.anchors[0][1]),
                 (self.anchors[1][0],self.anchors[0][1]),
                 (self.anchors[1][0],self.anchors[1][1]),
                 (self.anchors[0][0],self.anchors[1][1])]
            
            for anchor in a:
                extr = [ f(anchor[0], anchor[1]),
                         f(anchor[0]+self.sizeX-1, anchor[1]),
                         f(anchor[0], anchor[1]+self.sizeY-1),
                         f(anchor[0]+self.sizeX-1, anchor[1]+self.sizeY-1) ]

                extr.sort()
                newA.append(extr[0])
                
            newA.sort()
            
            newAnchors = [newA[0], newA[3]]

            newPd = {}
            
            for i in range(self.sizeX):
                for j in range(self.sizeY):
                    newPd[f(i,j)] = self.data[i][j]

            pNeu = Pattern(newAnchors, newPd)
            pNeu.flip = ii

            if not pNeu in l: l.append(pNeu)
            if pNeu == self: symmetries.append((ii,0))
  
            if nextMove or not fixedColor:
                newPd1 = {}
                for i in range(self.sizeX):
                    for j in range(self.sizeY):
                        if self.data[i][j] == 'X': newPd1[f(i,j)] = 'O'
                        elif self.data[i][j] == 'O': newPd1[f(i,j)] = 'X'
                        else: newPd1[f(i,j)] = self.data[i][j]

                pNeu1 = Pattern(newAnchors, newPd1)
                pNeu1.flip = ii
                pNeu1.colorSwitch = 1

                if not fixedColor and not pNeu1 in lCS: lCS.append(pNeu1)
                if pNeu1 == self:
                    if not fixedColor: symmetries.append((ii,1))
                    if nextMove: special = ii
                    
        for p in lCS:
            if not p in l: l.append(p)
        
        # compute symmetries
        
        symm = {}
        for i in range(self.sizeX):
            for j in range(self.sizeY):
                symm[(i,j)] = ((i,j),0)    

        for s, c in symmetries:
            symm1 = {}
            for i in range(self.sizeX):
                for j in range(self.sizeY):
                    if (i,j) != Pattern.flips[s](i,j,self.sizeX-1,self.sizeY-1) \
                       and not symm1.has_key(Pattern.flips[s](i,j,self.sizeX-1,self.sizeY-1)):
                        symm1[(i,j)] = (Pattern.flips[s](i,j,self.sizeX-1,self.sizeY-1), c)
 
            for k in symm.keys():
                if symm1.has_key(symm[k][0]):
                    if (symm1[symm[k][0]][1] or symm[k][1]) and \
                       not (symm1[symm[k][0]][1] and symm[k][1]):
                        cs = 1
                    else: cs = 0
                    symm[k] = symm1[symm[k][0]][0], cs


        if special == -1:
            symm[-1] = -1
        else:
            symm[-1] = PatternInvFlip(special) # if there is a flip f such that color_swap(f(pattern)) == pattern,
                                               # then this is stored as symm[-1]
                                               # this is needed to get the continuation symmetries right when the
                                               # color of the next move is fixed
        return l, symm

    
    def compute_hashkey(self, type):
        self.hashkey = []
        noOfStones = 0

        for a0,b0,a1,b1 in type:
            self.hashkey.append(0)
            counter = len(self.hashkey)-1

            valid = 1
            
            for i in range(a0,a1+1):
                for j in range(b0,b1+1):
                    ii, jj = Pattern.flips[self.flip](i, j)
                    if self.data[ii-self.anchors[0][0]][jj-self.anchors[0][1]] == '*': valid = 0
                    elif self.data[ii-self.anchors[0][0]][jj-self.anchors[0][1]] == 'X':
                        self.hashkey[counter] += hashTable.hash_value[(ii,jj)]
                        noOfStones += 1
                    elif self.data[ii-self.anchors[0][0]][jj-self.anchors[0][1]] == 'O':
                        self.hashkey[counter] -= hashTable.hash_value[(ii,jj)]
                        noOfStones += 1

            if (type[0] != (0,0,19,19) and noOfStones > 25) or noOfStones > 30: valid = 0

            if not valid:
                del self.hashkey[counter]

        if not self.hashkey: return 0
        else: return 1
    
    def __cmp__(self, p):                    # never returns -1, so this works only for
                                             # testing for equality
        if self.sizeX != p.sizeX or self.sizeY != p.sizeY: return 1
        if self.anchors != p.anchors: return 1
        for i in range(self.sizeX):
            if self.data[i] != p.data[i]: return 1
        return 0

    def __repr__(self):
        return join(self.data, '')

    def __init__(self, anchors, dict):
        self.flip = 0
        self.colorSwitch = 0
        
        self.anchors = anchors
        self.anchors.sort()               # important for __cmp__
        
        l = dict.keys()[:]
        l.sort()
        
        start = l[0]
        end = l[len(l)-1]

        self.sizeX = end[0] - start[0] + 1
        self.sizeY = end[1] - start[1] + 1

        self.data = []
        for i in range(self.sizeX):
            help = []
            for j in range(self.sizeY):
                help.append(dict[(i+start[0],j+start[1])])
            self.data.append(join(help,''))

        self.bits = [] # list of 4 bit-patterns

        for i in range(2):
            for j in range(2):
                xBlocks = (self.sizeY+i+1)/2
                yBlocks = (self.sizeX+j+1)/2
                nextBlock = array('B', [yBlocks]) 
                for k1 in range(yBlocks):
                    nlist = []
                    
                    for k2 in range(xBlocks):
                        n = 0
                        for x in range(2):
                            for y in range(2):
                                indexX = k1 * 2 + y - j
                                indexY = k2 * 2 + x - i
                                if 0 <= indexX < self.sizeX and \
                                   0 <= indexY < self.sizeY:
                                    if self.data[indexX][indexY]=='X':
                                        n |= 1 << (2*(2*x+y))
                                    elif self.data[indexX][indexY]=='O':
                                        n |= 1 << (2*(2*x+y)+1)
                        nlist.append(n)
                        
                    start, end = 0, len(nlist)
                    while start < end and not nlist[start]: start += 1
                    while end > start and not nlist[end-1]: end -= 1

                    nextBlock.append(start)
                    nextBlock.append(end-start)
                    for current in range(start, end): nextBlock.append(nlist[current])

                self.bits.append(nextBlock)


    flips = []
    flips.append(lambda x,y, XX=18, YY=18: (x,y))
    flips.append(lambda x,y, XX=18, YY=18: (XX-x,y))
    flips.append(lambda x,y, XX=18, YY=18: (x,YY-y))
    flips.append(lambda x,y, XX=18, YY=18: (XX-x,YY-y))
    flips.append(lambda x,y, XX=18, YY=18: (y,x))
    flips.append(lambda x,y, XX=18, YY=18: (YY-y,x))
    flips.append(lambda x,y, XX=18, YY=18: (y,XX-x))
    flips.append(lambda x,y, XX=18, YY=18: (YY-y,XX-x))


def PatternInvFlip(i):
    return [0,1,2,3,4,6,5,7][i]


class hashTable:

    def __init__(self, noOfGames):
         self.data = array('i', [0]*(2*130*noOfGames))
         self.currentCtr = -1
         self.dataIndex = 0

    def newCtr(self, ctr):
        if ctr == self.currentCtr:
            while self.dataIndex > 0 and self.data[self.dataIndex-1] == ctr:
                self.dataIndex -= 2
        self.currentCtr = ctr
        self.currentCtrL = []
         
    def append(self, i):
        
        ctr = self.currentCtr
            
        if i not in self.currentCtrL:
            if self.dataIndex >= len(self.data)-1:
                self.data.extend(array('i', [0]*1000))
            self.data[self.dataIndex] = i
            self.dataIndex += 1
            self.data[self.dataIndex] = ctr
            self.dataIndex += 1
            self.currentCtrL.append(i)


    def swap(self, l, r):
        help1 = self.data[2*l]
        help2 = self.data[2*l+1]
        self.data[2*l] = self.data[2*r]
        self.data[2*l+1] = self.data[2*r+1]
        self.data[2*r] = help1
        self.data[2*r+1] = help2
            
    def sort(self):

        if CimportSucceeded:
            matchC.sortArray(self.data, 0, self.dataIndex/2-1)
        else:
            l = 0
            r = self.dataIndex/2 - 1
            stack = [(l, r)]

            while stack:
                if r > l:
                    if r-l == 1:
                        if self.data[2*l] > self.data[2*r]:
                            self.swap(l,r)
                        l, r = stack.pop()
                    else:
                        li = [(self.data[2*l], l), (self.data[2*r], r), (self.data[2*((r+l)/2)], (r+l)/2)]
                        li.sort()

                        self.swap(li[1][1], r)
                        help1 = self.data[2*r]

                        left = l
                        right = r-1

                        while 1: 
                            while self.data[2*left] < help1: left += 1
                            while self.data[2*right] > help1: right -= 1

                            if right <= left: break
                            
                            self.swap(left, right)

                            left += 1
                            right -= 1
                            
                        self.swap(left, r)

                        if left-l > r-left:
                            stack.append((l, left-1))
                            l = left+1
                        else:
                            stack.append((left+1, r))
                            r = left-1                    
                else:
                    l, r = stack.pop()

    def sortedOutput(self, path):
        self.sort()

        if CimportSucceeded:
            hashT, hash = matchC.buildOutput(self.data, self.dataIndex)
            file = open(os.path.join(path[0], 'hashT' + path[1] + '.db'), 'wb')
            file.write(hashT)
            file.close()
            file = open(os.path.join(path[0], 'hash' + path[1] + '.db'), 'wb')
            file.write(hash)
            file.close()
        else:
            dataX = array('i')
            data1X = array('B')

            currentCode = self.data[0] - 1  # != self.data[0] !
            currentValues = []
        
            for i in xrange(self.dataIndex/2):
                if self.data[2*i] != currentCode:
                    currentValues.sort()
                    for v in currentValues:
                        data1X.append(v/256)
                        data1X.append(v%256)
                    currentValues = []
                    currentCode = self.data[2*i]
                    dataX.append(currentCode)
                    dataX.append(len(data1X))

                currentValues.append(self.data[2*i+1])

            currentValues.sort()
            for v in currentValues:
                data1X.append(v/256)
                data1X.append(v%256)
            
            file = open(os.path.join(path[0], 'hashT' + path[1] + '.db'), 'wb')
            dataX.tofile(file)
            file.close()
            file = open(os.path.join(path[0], 'hash' + path[1] + '.db'), 'wb')
            data1X.tofile(file)
            file.close()
            

    hash_value = {
        ( 0 ,  0 ):  9149491 , 
        ( 0 ,  1 ):  4143844 , 
        ( 0 ,  2 ):  8452911 , 
        ( 0 ,  3 ):  3100691 , 
        ( 0 ,  4 ):  3196613 , 
        ( 0 ,  5 ):  7210416 , 
        ( 0 ,  6 ):  6086509 , 
        ( 0 ,  7 ):  7957641 , 
        ( 0 ,  8 ):  131916 , 
        ( 0 ,  9 ):  6918136 , 
        ( 0 ,  10 ):  3016410 , 
        ( 0 ,  11 ):  3288711 , 
        ( 0 ,  12 ):  8539380 , 
        ( 0 ,  13 ):  758094 , 
        ( 0 ,  14 ):  7179238 , 
        ( 0 ,  15 ):  7718041 , 
        ( 0 ,  16 ):  5204153 , 
        ( 0 ,  17 ):  6711869 , 
        ( 0 ,  18 ):  8411491 , 
        ( 1 ,  0 ):  2434388 , 
        ( 1 ,  1 ):  1162683 , 
        ( 1 ,  2 ):  927102 , 
        ( 1 ,  3 ):  197419 , 
        ( 1 ,  4 ):  9435893 , 
        ( 1 ,  5 ):  2149216 , 
        ( 1 ,  6 ):  2967038 , 
        ( 1 ,  7 ):  5557593 , 
        ( 1 ,  8 ):  4044723 , 
        ( 1 ,  9 ):  4077413 , 
        ( 1 ,  10 ):  8810194 , 
        ( 1 ,  11 ):  7673830 , 
        ( 1 ,  12 ):  1290402 , 
        ( 1 ,  13 ):  704914 , 
        ( 1 ,  14 ):  2782252 , 
        ( 1 ,  15 ):  6972890 , 
        ( 1 ,  16 ):  2649162 , 
        ( 1 ,  17 ):  8444272 , 
        ( 1 ,  18 ):  5863711 , 
        ( 2 ,  0 ):  8701652 , 
        ( 2 ,  1 ):  9073146 , 
        ( 2 ,  2 ):  796709 , 
        ( 2 ,  3 ):  3071327 , 
        ( 2 ,  4 ):  81078 , 
        ( 2 ,  5 ):  413463 , 
        ( 2 ,  6 ):  9899215 , 
        ( 2 ,  7 ):  3900485 , 
        ( 2 ,  8 ):  4258038 , 
        ( 2 ,  9 ):  9853936 , 
        ( 2 ,  10 ):  5479882 , 
        ( 2 ,  11 ):  4613169 , 
        ( 2 ,  12 ):  7404562 , 
        ( 2 ,  13 ):  5251570 , 
        ( 2 ,  14 ):  1932628 , 
        ( 2 ,  15 ):  9394001 , 
        ( 2 ,  16 ):  6022071 , 
        ( 2 ,  17 ):  216105 , 
        ( 2 ,  18 ):  4278005 , 
        ( 3 ,  0 ):  7136853 , 
        ( 3 ,  1 ):  2500566 , 
        ( 3 ,  2 ):  4107671 , 
        ( 3 ,  3 ):  4081838 , 
        ( 3 ,  4 ):  1785699 , 
        ( 3 ,  5 ):  1424534 , 
        ( 3 ,  6 ):  3024969 , 
        ( 3 ,  7 ):  5734282 , 
        ( 3 ,  8 ):  4175470 , 
        ( 3 ,  9 ):  9404994 , 
        ( 3 ,  10 ):  8430123 , 
        ( 3 ,  11 ):  1255451 , 
        ( 3 ,  12 ):  5986866 , 
        ( 3 ,  13 ):  7055240 , 
        ( 3 ,  14 ):  6371620 , 
        ( 3 ,  15 ):  9283568 , 
        ( 3 ,  16 ):  7313682 , 
        ( 3 ,  17 ):  7717289 , 
        ( 3 ,  18 ):  2299204 , 
        ( 4 ,  0 ):  6500542 , 
        ( 4 ,  1 ):  1976702 , 
        ( 4 ,  2 ):  7553422 , 
        ( 4 ,  3 ):  5187929 , 
        ( 4 ,  4 ):  3201058 , 
        ( 4 ,  5 ):  2222174 , 
        ( 4 ,  6 ):  3817451 , 
        ( 4 ,  7 ):  7344416 , 
        ( 4 ,  8 ):  4960530 , 
        ( 4 ,  9 ):  6786163 , 
        ( 4 ,  10 ):  8006227 , 
        ( 4 ,  11 ):  442411 , 
        ( 4 ,  12 ):  4688575 , 
        ( 4 ,  13 ):  5012537 , 
        ( 4 ,  14 ):  3651243 , 
        ( 4 ,  15 ):  7062230 , 
        ( 4 ,  16 ):  6716660 , 
        ( 4 ,  17 ):  4891547 , 
        ( 4 ,  18 ):  5702153 , 
        ( 5 ,  0 ):  8827952 , 
        ( 5 ,  1 ):  7097129 , 
        ( 5 ,  2 ):  9410525 , 
        ( 5 ,  3 ):  1319909 , 
        ( 5 ,  4 ):  6697319 , 
        ( 5 ,  5 ):  6008737 , 
        ( 5 ,  6 ):  9209372 , 
        ( 5 ,  7 ):  1708104 , 
        ( 5 ,  8 ):  8930801 , 
        ( 5 ,  9 ):  3357765 , 
        ( 5 ,  10 ):  4158639 , 
        ( 5 ,  11 ):  8563534 , 
        ( 5 ,  12 ):  1000997 , 
        ( 5 ,  13 ):  4582316 , 
        ( 5 ,  14 ):  6276028 , 
        ( 5 ,  15 ):  9658685 , 
        ( 5 ,  16 ):  4453094 , 
        ( 5 ,  17 ):  9081939 , 
        ( 5 ,  18 ):  6565077 , 
        ( 6 ,  0 ):  3029400 , 
        ( 6 ,  1 ):  3036923 , 
        ( 6 ,  2 ):  2527385 , 
        ( 6 ,  3 ):  3663094 , 
        ( 6 ,  4 ):  1261450 , 
        ( 6 ,  5 ):  9321901 , 
        ( 6 ,  6 ):  4118265 , 
        ( 6 ,  7 ):  7506223 , 
        ( 6 ,  8 ):  8638174 , 
        ( 6 ,  9 ):  1830356 , 
        ( 6 ,  10 ):  9269438 , 
        ( 6 ,  11 ):  379190 , 
        ( 6 ,  12 ):  5190162 , 
        ( 6 ,  13 ):  4415341 , 
        ( 6 ,  14 ):  455343 , 
        ( 6 ,  15 ):  2706427 , 
        ( 6 ,  16 ):  374332 , 
        ( 6 ,  17 ):  4806453 , 
        ( 6 ,  18 ):  3940651 , 
        ( 7 ,  0 ):  7845245 , 
        ( 7 ,  1 ):  2132422 , 
        ( 7 ,  2 ):  313121 , 
        ( 7 ,  3 ):  1581806 , 
        ( 7 ,  4 ):  4042198 , 
        ( 7 ,  5 ):  4890562 , 
        ( 7 ,  6 ):  4352418 , 
        ( 7 ,  7 ):  5580773 , 
        ( 7 ,  8 ):  9938852 , 
        ( 7 ,  9 ):  6930891 , 
        ( 7 ,  10 ):  8376253 , 
        ( 7 ,  11 ):  5737628 , 
        ( 7 ,  12 ):  8471297 , 
        ( 7 ,  13 ):  1142249 , 
        ( 7 ,  14 ):  848865 , 
        ( 7 ,  15 ):  883348 , 
        ( 7 ,  16 ):  9812485 , 
        ( 7 ,  17 ):  5717034 , 
        ( 7 ,  18 ):  7432763 , 
        ( 8 ,  0 ):  2354745 , 
        ( 8 ,  1 ):  4069809 , 
        ( 8 ,  2 ):  9707233 , 
        ( 8 ,  3 ):  9588944 , 
        ( 8 ,  4 ):  3016914 , 
        ( 8 ,  5 ):  554095 , 
        ( 8 ,  6 ):  1638369 , 
        ( 8 ,  7 ):  9162618 , 
        ( 8 ,  8 ):  611584 , 
        ( 8 ,  9 ):  5407576 , 
        ( 8 ,  10 ):  3543495 , 
        ( 8 ,  11 ):  640558 , 
        ( 8 ,  12 ):  5662155 , 
        ( 8 ,  13 ):  3501725 , 
        ( 8 ,  14 ):  7560991 , 
        ( 8 ,  15 ):  3133362 , 
        ( 8 ,  16 ):  9084840 , 
        ( 8 ,  17 ):  6812692 , 
        ( 8 ,  18 ):  1594571 , 
        ( 9 ,  0 ):  9306451 , 
        ( 9 ,  1 ):  238482 , 
        ( 9 ,  2 ):  609444 , 
        ( 9 ,  3 ):  1700604 , 
        ( 9 ,  4 ):  3082475 , 
        ( 9 ,  5 ):  5804654 , 
        ( 9 ,  6 ):  6738079 , 
        ( 9 ,  7 ):  2956313 , 
        ( 9 ,  8 ):  3546729 , 
        ( 9 ,  9 ):  3491719 , 
        ( 9 ,  10 ):  1214856 , 
        ( 9 ,  11 ):  888594 , 
        ( 9 ,  12 ):  4911844 , 
        ( 9 ,  13 ):  315409 , 
        ( 9 ,  14 ):  983871 , 
        ( 9 ,  15 ):  5306260 , 
        ( 9 ,  16 ):  7005661 , 
        ( 9 ,  17 ):  4089903 , 
        ( 9 ,  18 ):  2599372 , 
        ( 10 ,  0 ):  7376553 , 
        ( 10 ,  1 ):  5249330 , 
        ( 10 ,  2 ):  1810656 , 
        ( 10 ,  3 ):  765893 , 
        ( 10 ,  4 ):  2667106 , 
        ( 10 ,  5 ):  7557868 , 
        ( 10 ,  6 ):  9214606 , 
        ( 10 ,  7 ):  5347135 , 
        ( 10 ,  8 ):  6166404 , 
        ( 10 ,  9 ):  1400859 , 
        ( 10 ,  10 ):  1919455 , 
        ( 10 ,  11 ):  8828173 , 
        ( 10 ,  12 ):  6251839 , 
        ( 10 ,  13 ):  4004673 , 
        ( 10 ,  14 ):  2006377 , 
        ( 10 ,  15 ):  9013385 , 
        ( 10 ,  16 ):  6847406 , 
        ( 10 ,  17 ):  7019454 , 
        ( 10 ,  18 ):  1408057 , 
        ( 11 ,  0 ):  6063976 , 
        ( 11 ,  1 ):  4380230 , 
        ( 11 ,  2 ):  8034820 , 
        ( 11 ,  3 ):  1813992 , 
        ( 11 ,  4 ):  9773930 , 
        ( 11 ,  5 ):  2520157 , 
        ( 11 ,  6 ):  1124498 , 
        ( 11 ,  7 ):  9369992 , 
        ( 11 ,  8 ):  1110455 , 
        ( 11 ,  9 ):  1395733 , 
        ( 11 ,  10 ):  2582409 , 
        ( 11 ,  11 ):  1697715 , 
        ( 11 ,  12 ):  2368383 , 
        ( 11 ,  13 ):  745577 , 
        ( 11 ,  14 ):  8857326 , 
        ( 11 ,  15 ):  2179278 , 
        ( 11 ,  16 ):  4267432 , 
        ( 11 ,  17 ):  6892826 , 
        ( 11 ,  18 ):  6099809 , 
        ( 12 ,  0 ):  9973299 , 
        ( 12 ,  1 ):  5025252 , 
        ( 12 ,  2 ):  3632262 , 
        ( 12 ,  3 ):  3081860 , 
        ( 12 ,  4 ):  2809085 , 
        ( 12 ,  5 ):  1683429 , 
        ( 12 ,  6 ):  8369333 , 
        ( 12 ,  7 ):  7991628 , 
        ( 12 ,  8 ):  8365980 , 
        ( 12 ,  9 ):  465773 , 
        ( 12 ,  10 ):  1820067 , 
        ( 12 ,  11 ):  1457115 , 
        ( 12 ,  12 ):  2753928 , 
        ( 12 ,  13 ):  282741 , 
        ( 12 ,  14 ):  8546560 , 
        ( 12 ,  15 ):  6233640 , 
        ( 12 ,  16 ):  3079605 , 
        ( 12 ,  17 ):  4697340 , 
        ( 12 ,  18 ):  4846782 , 
        ( 13 ,  0 ):  5022954 , 
        ( 13 ,  1 ):  9875417 , 
        ( 13 ,  2 ):  9168302 , 
        ( 13 ,  3 ):  6618946 , 
        ( 13 ,  4 ):  6928618 , 
        ( 13 ,  5 ):  8503198 , 
        ( 13 ,  6 ):  5052403 , 
        ( 13 ,  7 ):  2078686 , 
        ( 13 ,  8 ):  2995926 , 
        ( 13 ,  9 ):  2194938 , 
        ( 13 ,  10 ):  820379 , 
        ( 13 ,  11 ):  7612996 , 
        ( 13 ,  12 ):  1375726 , 
        ( 13 ,  13 ):  8614873 , 
        ( 13 ,  14 ):  5945335 , 
        ( 13 ,  15 ):  4672767 , 
        ( 13 ,  16 ):  3036307 , 
        ( 13 ,  17 ):  6840821 , 
        ( 13 ,  18 ):  3921566 , 
        ( 14 ,  0 ):  9316769 , 
        ( 14 ,  1 ):  5099779 , 
        ( 14 ,  2 ):  1957262 , 
        ( 14 ,  3 ):  82550 , 
        ( 14 ,  4 ):  8061649 , 
        ( 14 ,  5 ):  5685904 , 
        ( 14 ,  6 ):  7357125 , 
        ( 14 ,  7 ):  3158126 , 
        ( 14 ,  8 ):  9681206 , 
        ( 14 ,  9 ):  9939320 , 
        ( 14 ,  10 ):  1416538 , 
        ( 14 ,  11 ):  9680390 , 
        ( 14 ,  12 ):  4083051 , 
        ( 14 ,  13 ):  1970772 , 
        ( 14 ,  14 ):  8454640 , 
        ( 14 ,  15 ):  8443104 , 
        ( 14 ,  16 ):  4061492 , 
        ( 14 ,  17 ):  6778512 , 
        ( 14 ,  18 ):  5077238 , 
        ( 15 ,  0 ):  5263977 , 
        ( 15 ,  1 ):  6837884 , 
        ( 15 ,  2 ):  4767983 , 
        ( 15 ,  3 ):  1402914 , 
        ( 15 ,  4 ):  1063494 , 
        ( 15 ,  5 ):  8805596 , 
        ( 15 ,  6 ):  2899665 , 
        ( 15 ,  7 ):  5717105 , 
        ( 15 ,  8 ):  5032108 , 
        ( 15 ,  9 ):  7482452 , 
        ( 15 ,  10 ):  95784 , 
        ( 15 ,  11 ):  6793965 , 
        ( 15 ,  12 ):  7138721 , 
        ( 15 ,  13 ):  7934664 , 
        ( 15 ,  14 ):  6505269 , 
        ( 15 ,  15 ):  6678380 , 
        ( 15 ,  16 ):  2901608 , 
        ( 15 ,  17 ):  5665651 , 
        ( 15 ,  18 ):  3599459 , 
        ( 16 ,  0 ):  5070383 , 
        ( 16 ,  1 ):  1714950 , 
        ( 16 ,  2 ):  8047561 , 
        ( 16 ,  3 ):  8205607 , 
        ( 16 ,  4 ):  67238 , 
        ( 16 ,  5 ):  2156761 , 
        ( 16 ,  6 ):  5827686 , 
        ( 16 ,  7 ):  1391658 , 
        ( 16 ,  8 ):  3591582 , 
        ( 16 ,  9 ):  7018330 , 
        ( 16 ,  10 ):  9232127 , 
        ( 16 ,  11 ):  8849431 , 
        ( 16 ,  12 ):  820043 , 
        ( 16 ,  13 ):  7847325 , 
        ( 16 ,  14 ):  3583231 , 
        ( 16 ,  15 ):  2966011 , 
        ( 16 ,  16 ):  5159425 , 
        ( 16 ,  17 ):  4855713 , 
        ( 16 ,  18 ):  1079091 , 
        ( 17 ,  0 ):  2870996 , 
        ( 17 ,  1 ):  5867619 , 
        ( 17 ,  2 ):  9327733 , 
        ( 17 ,  3 ):  4398712 , 
        ( 17 ,  4 ):  9371591 , 
        ( 17 ,  5 ):  2182067 , 
        ( 17 ,  6 ):  2409473 , 
        ( 17 ,  7 ):  709735 , 
        ( 17 ,  8 ):  4862464 , 
        ( 17 ,  9 ):  8833884 , 
        ( 17 ,  10 ):  1260720 , 
        ( 17 ,  11 ):  37637 , 
        ( 17 ,  12 ):  2367746 , 
        ( 17 ,  13 ):  8087960 , 
        ( 17 ,  14 ):  4198525 , 
        ( 17 ,  15 ):  82482 , 
        ( 17 ,  16 ):  4389571 , 
        ( 17 ,  17 ):  787391 , 
        ( 17 ,  18 ):  8032185 , 
        ( 18 ,  0 ):  2095647 , 
        ( 18 ,  1 ):  5448853 , 
        ( 18 ,  2 ):  850526 , 
        ( 18 ,  3 ):  2278124 , 
        ( 18 ,  4 ):  6280764 , 
        ( 18 ,  5 ):  4617893 , 
        ( 18 ,  6 ):  2879123 , 
        ( 18 ,  7 ):  806861 , 
        ( 18 ,  8 ):  4007840 , 
        ( 18 ,  9 ):  9545335 , 
        ( 18 ,  10 ):  9072999 , 
        ( 18 ,  11 ):  1446181 , 
        ( 18 ,  12 ):  7946775 , 
        ( 18 ,  13 ):  6118313 , 
        ( 18 ,  14 ):  5073149 , 
        ( 18 ,  15 ):  8222792 , 
        ( 18 ,  16 ):  4448444 , 
        ( 18 ,  17 ):  4579004 , 
        ( 18 ,  18 ):  5179082       }
        
# ---------------------------------------------------------------------------------------

def getFilename(s):
    if s[-1] == '.': return s[:-1]         # no extension
    elif s[-2:] == '.m': return s + 'gt'   # extension '.mgt'
    else: return s + '.sgf'                # extension '.sgf'


def symmetrizeSig(s):
    """ Given a signature s, compute the 'rotated'/'mirrored' signatures,
    and return the first, w.r.t. lexicographic order, of all these.
    Games which differ only by a symmetry of the board will thus have the
    same symmetrized signature. """

    l = []
    oa = ord('a')
    for i in range(6): l.append((ord(s[2*i])-oa, ord(s[2*i+1])-oa))

    m = []
    for f in Pattern.flips:
        k1 = [ f(x[0], x[1]) for x in l] 
        k2 = []
        for i in range(6):
            if s[2*i] == '?': k2.append('?')
            else: k2.append(chr(k1[i][0]+oa))
            if s[2*i+1] == '?': k2.append('?')
            else: k2.append(chr(k1[i][1]+oa))
        m.append(join(k2, ''))

    m.sort()
    return m[0]

# ---------------------------------------------------------------------------------------

class App(v.Viewer):
    """ The main class of Kombilo. """

    def process(self, dbpath, datap, messages):
        """ Process the database dbpath, i.e. translate the SGF files into the
        .db files that the pattern search function uses.
        """

        currentTime = time.time()
        self.progBar.clear()
        
        messages.insert(END, 'Processing ' + dbpath + ' ...\n')
        
        mainlistArr = array('B')
        posTable = array('I')
        finalpos = array('B')
        namelist = []
        siglist = []
        
        oa = ord('a')

        if self.filenamesVar.get() == '*.sgf':
            filelist = glob.glob(os.path.join(dbpath,'*.sgf'))
        elif self.filenamesVar.get() == '*.sgf, *.mgt':
            filelist = glob.glob(os.path.join(dbpath,'*.sgf')) + glob.glob(os.path.join(dbpath, '*.mgt'))
        else:
            filelist = glob.glob(os.path.join(dbpath,'*'))

        encoding = self.encodingVar.get()
        
        encoding1 = 0
        if self.encoding1Var.get() == 'Do not change SGF': encoding1 = 0
        elif self.encoding1Var.get() == 'Add CA tag': encoding1 = 1
        elif self.encoding1Var.get() == 'Transcode SGF to utf-8': encoding1 = 2
        
            
        filelist.sort()
        hash_global = hashTable(len(filelist))

        total = len(filelist)
        gameCtr = 0 # counts games which are added to the new database
        fileCtr = 0 # counts files which are processed
        
        for f in filelist:

            if fileCtr % 10 == 0:
                self.progBar.redraw(fileCtr*1.0/total)
                self.searchWindow.update()
            fileCtr += 1

            if self.stopAddDBVar.get() == 1:
                showinfo('Stop processing', 'Processing will be stopped after this database has been finished.\n')
                self.stopAddDBVar.set(2)

            if not os.path.isfile(f): continue

            if find(f, '|%') != -1 or find(f, '[') != -1:
                messages.insert(END, 'Invalid characters in filename ' + f + ', skipped.\n')
                continue

            try:
                file = open(f)
                sOrig = file.read()
                file.close()
            except IOError:
                messages.insert(END, 'I/O Error: Cannot read file ' + f + ', skipped.\n')
                continue

            try:
                cursor = Cursor(sOrig, self.options.sloppySGF.get())
            except:
                messages.insert(END, 'Error in SGF file '+f+': Skipped.\n')
                continue
            
            if encoding1 == 1 and not cursor.encoding:
                d = cursor.getRootNode(0)
                d['CA'] = [encoding]
                cursor.encoding = encoding
                cursor.updateRootNode(d,0)
                try:
                    sgf_out = cursor.output()
                    file = open(f, 'w')
                    file.write(sgf_out)
                    file.close()
                except:
                    messages.insert(END, 'Unable to write CA-tagged file ' + f + '\n')
            currentEncoding = cursor.encoding or encoding
            if currentEncoding != 'utf-8':
                try:
                    s = sOrig.decode(currentEncoding, 'ignore').encode('utf-8', 'ignore')
                    cursor = Cursor(s, self.options.sloppySGF.get())
                    
                    if encoding1 == 2 and not cursor.encoding:
                        d = cursor.getRootNode(0)
                        d['CA'] = ['utf-8']
                        cursor.encoding = 'utf-8'
                        cursor.updateRootNode(d,0)
                        file = open(f, 'w')
                        file.write(cursor.output())
                        file.close()
                except IOError:
                    messages.insert(END, 'Unable to write transcoded file ' + f + '\n')
                except:
                    messages.insert(END, 'Unable to transcode file ' + f + '\n')
                    
            if cursor.root.numChildren > 1:
                collection = 1
                messages.insert(END,'Warning: '+f+' is a collection; for performance reasons it would be better to' + \
                                ' split it into files with single game records.\n')
            elif cursor.root.numChildren == 1:
                collection = 0
            else:
                messages.insert(END, 'File '+f+' does not contain a game record. Skipped.\n')
                continue

            for indexColl in range(cursor.root.numChildren):
                if collection: iColl = '[' + `indexColl` + ']'
                else: iColl = ''

                try:
                    cursor.game(indexColl)
                except SGFError:
                    messages.insert(END, 'SGF Error in game ' + iColl + ' of file ' + f + ', skipped.\n')
                    continue
                
                try:
                    if cursor.currentNode().has_key('SZ') and strip(cursor.currentNode()['SZ'][0]) != '19':
                        messages.insert(END, 'SGF file '+ f + iColl +' does not contain a 19x19 game. Skipped.\n')
                        continue
            
                    if cursor.currentNode().has_key('PB'):
                        playerB = replace(cursor.currentNode()['PB'][0], '|%', ' ')
                        playerB = replace(playerB, '\r', ' ')
                        playerB = replace(playerB, '\n', ' ')
                    else: playerB = '?'
                    if cursor.currentNode().has_key('PW'):
                        playerW = replace(cursor.currentNode()['PW'][0], '|%', ' ')
                        playerW = replace(playerW, '\r', ' ')
                        playerW = replace(playerW, '\n', ' ')
                    else: playerW = '?'

                    if cursor.currentNode().has_key('RE') and strip(cursor.currentNode()['RE'][0]):
                        result = strip(cursor.currentNode()['RE'][0])[0] # We assume that the first letter in
                                                                         # the RE tag is the winner's color
                    else: result = '-'

                    if cursor.currentNode().has_key('DT') and cursor.currentNode()['DT'][0]:
                        date_str = cursor.currentNode()['DT'][0]
                        
                        m = re.search('(^|[^\d])(\d\d\d\d-\d\d-\d\d)($|[^\d])', date_str)
                    
                        try:
                            date = m.group(2)
                        except (AttributeError, ValueError):
                            m = re.search('(^|[^\d])(\d\d\d\d-\d\d)($|[^\d])', date_str)
                            try:
                                date = m.group(2)
                            except (AttributeError, ValueError):
                                m = re.search('(^|[^\d])(\d\d\d\d)($|[^\d])', date_str)
                                try:
                                    date = m.group(2)
                                except (AttributeError, ValueError):
                                    date = ''
                    else:
                        date = ''
                except:
                    messages.insert(END, 'SGF Error in root node of game ' + iColl + ', file ' + f + ', skipped.\n')
                    continue

                try:
                    fp = array('B', [255] * 100)   # '20x20x2 bit-array', in the end 00 means: B, W at some time,
                                                   #                                 01 = never B, but W ...
                    fpC = array('B', [0] * 50)     # '20x20 bit array',   0 = never captured, 1 = captured
                    movelist = array('B')
                    caplist = array('B')
                    startGame = 0
                    counter = 0

                    b = v.abstractBoard()
                    b.noOfStones = 0
                    b.noOfStonesUL = 0
                    b.noOfStonesUR = 0
                    b.noOfStonesLL = 0
                    b.noOfStonesLR = 0

                    h_key = 0
                    h_keyUL = 0
                    h_keyUR = 0
                    h_keyLL = 0
                    h_keyLR = 0

                    hash_global.newCtr(gameCtr)

                    # AB's

                    while not (cursor.currentNode().has_key('AB') or cursor.currentNode().has_key('AW') or \
                               cursor.currentNode().has_key('B') or cursor.currentNode().has_key('W')):
                        if not cursor.atEnd: cursor.next()
                        else: break

                    if cursor.atEnd and not (cursor.currentNode().has_key('AB') or \
                                             cursor.currentNode().has_key('AW') or \
                                             cursor.currentNode().has_key('B') or \
                                             cursor.currentNode().has_key('W')):
                        continue

                    if cursor.currentNode().has_key('AB'):
                        movelist.append(25)
                        movelist.append(3) # AB's
                        for p in cursor.currentNode()['AB']:
                            if len(p) != 2: continue
                            x, y = ord(p[1])-oa, ord(p[0])-oa
                            b.play((x+1, y+1), 'black')
                    
                            b.noOfStones += 1
                            h_key += hashTable.hash_value[(x,y)]
                            if 0 <= x <= 6 and 0 <= y <= 6:
                                b.noOfStonesUL += 1
                                h_keyUL += hashTable.hash_value[(x,y)]
                            elif 0 <= x <= 6 and 12 <= y <= 18:
                                b.noOfStonesLL += 1
                                h_keyLL += hashTable.hash_value[(x,y)]
                            elif 12 <= x <= 18 and 0 <= y <= 6:
                                b.noOfStonesUR += 1
                                h_keyUR += hashTable.hash_value[(x,y)]
                            elif 12 <= x <= 18 and 12 <= y <= 18:
                                b.noOfStonesLR += 1
                                h_keyLR += hashTable.hash_value[(x,y)]
                    
                            fp[y/2 + 10*(x/2)] &= ~(1 << (2*(x%2 + 2*(y%2))))
                            movelist.append(x)
                            movelist.append(y)
                
                    # AW's
                    
                    if cursor.currentNode().has_key('AW'):
                        movelist.append(25)
                        movelist.append(4) # AW's
                        for p in cursor.currentNode()['AW']:
                            if len(p) != 2: continue
                            x, y = ord(p[1])-oa, ord(p[0])-oa
                            b.play((x+1, y+1), 'white')
                    
                            b.noOfStones += 1
                            h_key -= hashTable.hash_value[(x,y)]
                            if 0 <= x <= 6 and 0 <= y <= 6:
                                b.noOfStonesUL += 1
                                h_keyUL -= hashTable.hash_value[(x,y)]
                            elif 0 <= x <= 6 and 12 <= y <= 18:
                                b.noOfStonesLL += 1
                                h_keyLL -= hashTable.hash_value[(x,y)]
                            elif 12 <= x <= 18 and 0 <= y <= 6:
                                b.noOfStonesUR += 1
                                h_keyUR -= hashTable.hash_value[(x,y)]
                            elif 12 <= x <= 18 and 12 <= y <= 18:
                                b.noOfStonesLR += 1
                                h_keyLR -= hashTable.hash_value[(x,y)]
                    
                            fp[y/2 + 10*(x/2)] &= ~(1 << (2*(x%2 + 2*(y%2))+1))
                            movelist.append(x)
                            movelist.append(y)

                    if cursor.currentNode().has_key('AB') or cursor.currentNode().has_key('AW') and not cursor.atEnd:
                        cursor.next()

                    if b.noOfStones <= 30: hash_global.append(h_key)
                    if b.noOfStonesUL <= 25: hash_global.append(h_keyUL)
                    if b.noOfStonesLL <= 25: hash_global.append(h_keyLL)
                    if b.noOfStonesUR <= 25: hash_global.append(h_keyUR)
                    if b.noOfStonesLR <= 25: hash_global.append(h_keyLR)

                    # B's and W's

                    if cursor.currentNode().has_key('B'):
                        color = 'black'
                        co = 'B'
                        movelist.append(25)
                        movelist.append(1)
                    elif cursor.currentNode().has_key('W'):
                        color = 'white'
                        co = 'W'
                        movelist.append(25)
                        movelist.append(2)
                    else:
                        messages.insert(END, 'No moves in file ' + f + iColl + ', skipped.\n')
                        continue

                    startGame = len(movelist)/2 - 1 # movelist[2*(startGame + i)], movelist[2*(startgame+i)+1]
                                                    # will contain the i-th
                                                    # proper move, i = 1,2,... (used to compute the
                                                    # Dyer signature

                    while 1:
                        p = cursor.currentNode()[co][0]
                        if len(p) != 2:
                            if len(p) != 0: messages.insert(END, 'SGF Error in file ' + f + '\n')
                            break

                        if co == 'B': co = 'W'
                        else:         co = 'B'
                
                        x, y = ord(p[1])-oa, ord(p[0])-oa
                    
                        if 0 <= x <= 18 and 0 <= y <= 18:
                            if not b.play((x+1, y+1), color):
                                messages.insert(END, 'Illegal move in file ' + f \
                                                + ' (' + `counter+1` + ')\n')
                                break

                            if color == 'black': epsilon = 1
                            else: epsilon = -1
                        
                            b.noOfStones += 1
                            h_key += epsilon*hashTable.hash_value[(x,y)]
                            if 0 <= x <= 6 and 0 <= y <= 6:
                                b.noOfStonesUL += 1
                                h_keyUL += epsilon*hashTable.hash_value[(x,y)]
                            elif 0 <= x <= 6 and 12 <= y <= 18:
                                b.noOfStonesLL += 1
                                h_keyLL += epsilon*hashTable.hash_value[(x,y)]
                            elif 12 <= x <= 18 and 0 <= y <= 6:
                                b.noOfStonesUR += 1
                                h_keyUR += epsilon*hashTable.hash_value[(x,y)]
                            elif 12 <= x <= 18 and 12 <= y <= 18:
                                b.noOfStonesLR += 1
                                h_keyLR += epsilon*hashTable.hash_value[(x,y)]
                    
                            movelist.append(x)
                            movelist.append(y)
                            if color == 'black': fp[y/2 + 10*(x/2)] &= ~(1 << (2*(x%2 + 2*(y%2))))
                            else:                fp[y/2 + 10*(x/2)] &= ~(1 << (2*(x%2 + 2*(y%2))+1))
                        else:
                            if (x,y) != (19,19):
                                messages.insert(END, 'Invalid move in file '+f+
                                                ' (' + `counter+1` + ')\n')
                            break
                
                        if b.undostack:
                            captures = b.undostack[len(b.undostack)-1][2]
                            if captures:
                                caplist.append(ord('y')) # separating symbol
                                caplist.append(counter/256)
                                caplist.append(counter%256)

                                if color == 'black': epsilon = 1
                                else: epsilon = -1
                        
                                for c in captures:
                                    x, y = c[0]-1, c[1]-1
                                    caplist.append(x+1)
                                    caplist.append(y+1)
                                    fpC[y/4 + 5*(x/2)] |= 1 << (x%2 + 2*(y%4))
                                    
                                    b.noOfStones -= 1
                                    h_key += epsilon*hashTable.hash_value[(x,y)]
                                    if 0 <= x <= 6 and 0 <= y <= 6:
                                        b.noOfStonesUL -= 1
                                        h_keyUL += epsilon*hashTable.hash_value[(x,y)]
                                    elif 0 <= x <= 6 and 12 <= y <= 18:
                                        b.noOfStonesLL -= 1
                                        h_keyLL += epsilon*hashTable.hash_value[(x,y)]
                                    elif 12 <= x <= 18 and 0 <= y <= 6:
                                        b.noOfStonesUR -= 1
                                        h_keyUR += epsilon*hashTable.hash_value[(x,y)]
                                    elif 12 <= x <= 18 and 12 <= y <= 18:
                                        b.noOfStonesLR -= 1
                                        h_keyLR += epsilon*hashTable.hash_value[(x,y)]
                        
                        if color == 'black': color = 'white'
                        else:                color = 'black'
                
                        # Hash values:
                
                        if b.noOfStones <= 30: hash_global.append(h_key)
                        if b.noOfStonesUL <= 25: hash_global.append(h_keyUL)
                        if b.noOfStonesLL <= 25: hash_global.append(h_keyLL)
                        if b.noOfStonesUR <= 25: hash_global.append(h_keyUR)
                        if b.noOfStonesLR <= 25: hash_global.append(h_keyLR)
 
                        if cursor.atEnd: break
                        else:
                            cursor.next()
                            while not cursor.currentNode().has_key(co) and not cursor.atEnd:
                                if not cursor.currentNode().has_key('B') and not cursor.currentNode().has_key('W') \
                                   and not cursor.currentNode().has_key('AB') and not cursor.currentNode().has_key('AW') \
                                   and not cursor.currentNode().has_key('AE'):
                                    cursor.next()
                                else:
                                    break
                                
                            if not cursor.currentNode().has_key(co):
                                if not cursor.atEnd:
                                    messages.insert(END, 'File ' + f + iColl + ': no B/W tag in node ' + `counter+3` + '.\n')
                                break

                        counter = counter + 1

                except SGFError:
                    messages.insert(END, 'File ' + f +iColl + ': SGF error in node ' + `counter+3` + '.\n')

                except:
                    messages.insert(END, 'File ' + f + iColl + ': Error in node ' + `counter+3` + '.\n')


                try:

                    # compute Dyer signature (moves 20, 40, 60, 31, 51, 71 in sgf coordinates -
                    # this will mostly characterise the game uniquely and allow to search
                    # for duplicates in an easy way)

                    signature = ''
                    for index in [20, 40, 60, 31, 51, 71]:
                        if len(movelist)/2 > index + startGame:
                            signature = signature + chr(movelist[2*(index+startGame)+1] + oa) + \
                                        chr(movelist[2*(index+startGame)] + oa)
                        else: signature = signature + '??'
           
                    signature = symmetrizeSig(signature)

                    # Now check for duplicates; first in the currently processed database,
                    # then in the others in self.gamelist.DBlist. If games with the same signature
                    # are found, the final position is compared. If those agree, too, is is concluded
                    # that the games are identical.

                    if self.options.duplicateCheck.get() and \
                       (counter > 20 or (counter > 5 and self.options.strictDuplicateCheck.get())):

                        strictDupCheck = self.options.strictDuplicateCheck.get()

                        duplist = []    
                        for sli in range(len(siglist)):
                            if siglist[sli] == signature and \
                               (not strictDupCheck or finalpos[sli*150:sli*150+100] == fp):
                                duplist.append(getFilename(split(namelist[sli], '|%')[0]))

                        duplist = duplist + self.gamelist.checkDuplicate(signature, fp,
                                                                         self.options.strictDuplicateCheck.get())

                        if strictDupCheck:
                            dupText = ' is the same as '
                        else:
                            dupText = ' is probably the same as '

                        if duplist:
                            ff = join(duplist, ', ')
                            if self.options.duplicateCheck.get() == 1:
                                messages.insert(END, 'Duplicate: ' + os.path.split(f)[1] +iColl+ dupText + ff + '.\n')
                            elif self.options.duplicateCheck.get() == 3:
                                messages.insert(END, 'Duplicate: '+os.path.split(f)[1]+iColl+\
                                                dupText + ff + ', omitted.\n')
                                continue
                            elif self.options.duplicateCheck.get() == 2:
                                if askokcancel('Duplicate', 'File ' +os.path.split(f)[1]+iColl\
                                               + ' is the same as ' + ff + '. Include it?'):
                                    messages.insert(END, 'Duplicate: ' + os.path.split(f)[1]+iColl\
                                                    + dupText + ff + '\n')
                                else:
                                    messages.insert(END, 'Duplicate: ' + os.path.split(f)[1]\
                                                    + dupText + ff + ', omitted.\n')
                                    continue
                    
                    siglist.append(signature)

                except:
                    messages.insert(END, 'File ' + f +iColl + ': Error computing the Dyer signature.\n')
                    signature = '????????????'


                # ----------------------------------
         
                finalpos.extend(fp)
                finalpos.extend(fpC)

                posTable.append(len(mainlistArr))
            
                mainlistArr.append(ord('z'))
                mainlistArr.append(oa)
                mainlistArr.extend(movelist)
                mainlistArr.append(ord('z'))
                mainlistArr.append(ord('b'))
                mainlistArr.extend(caplist)
                mainlistArr.append(ord('z'))
                mainlistArr.append(ord('c'))

                if os.path.split(f)[1][-4:] == '.sgf': filenameDB = os.path.split(f)[1][:-4]
                elif os.path.split(f)[1][-4:] == '.mgt': filenameDB = os.path.split(f)[1][:-2] 
                else: filenameDB = os.path.split(f)[1] + '.'
                
                namelist.append(filenameDB + iColl + '|%' + playerB + '|%' + playerW + '|%' \
                                + result + '|%' + signature + '|%' + date)

                gameCtr += 1

        hash_global.newCtr(gameCtr)  # in order to correct the hash table if the last game was
                                     # erroneous or a duplicate

        if len(mainlistArr):
            try:
                hash_global.sortedOutput(datap)
            except IOError:
                messages.insert(END, 'I/O Error: Could not write hash tables (Directory write-protected?).\n')
                return
        
            try:
                filelist = [(posTable, 'posTable'), (mainlistArr, 'lists'),
                            (finalpos, 'finalpos')]

                for var, filename in filelist:
                    file = open(os.path.join(datap[0], filename + datap[1] + '.db'), 'wb')
                    var.tofile(file)
                    file.close()
                
                file = open(os.path.join(datap[0], 'namelist' + datap[1] + '.db'), 'w')
                for s in namelist:
                    file.write(s+'\n')
                file.close()
            except IOError:
                messages.insert(END, 'I/O Error: Could not write .db files.\n')

        self.progBar.redraw(1)
        self.progBar.write('%1.1f seconds' % (time.time() - currentTime))

        messages.insert(END, '... finished.\n')
        

    def checkCandidate(self, dbIndex, index, patternList, matchList, fixedColor):
        """ This is part of the search engine. See search for a description. """

        db = self.gamelist.DBlist[dbIndex]
        
        result = []
        noSwitched = 0

        current = self.posTable[index]

        current = current + 2
        movelist = []
        while self.mainlistArr[current] != ord('z'):
            movelist.append((self.mainlistArr[current], self.mainlistArr[current+1]))
            current = current + 2
        
        current = current + 2
        
        caplist = []
        while self.mainlistArr[current] != ord('z'):
            current = current + 1
            ctr = self.mainlistArr[current]*256 + self.mainlistArr[current+1]
            current = current + 2
            cap = []
            while not self.mainlistArr[current] in [ord('y'), ord('z')]:
                cap.append((self.mainlistArr[current], self.mainlistArr[current+1]))
                current = current + 2
            caplist.append((ctr, cap))
        
        current = current + 2

        dicts = {}
        dictsNO = {}
        Xinterv = {}
        Yinterv = {}
        for m in matchList:
            d = {}
            dNO = 0
            for i in range(patternList[m[0]].sizeX):
                for j in range(patternList[m[0]].sizeY):
                    if patternList[m[0]].data[i][j] != '.':
                        d[(i+m[1][0],j+m[1][1])] = patternList[m[0]].data[i][j]
                        if patternList[m[0]].data[i][j] != '*': dNO = dNO + 1
            dicts[m] = d
            dictsNO[m] = dNO # no of places that are currently not correct
                        
            Xinterv[m] = (m[1][0], m[1][0] + patternList[m[0]].sizeX)
            Yinterv[m] = (m[1][1], m[1][1] + patternList[m[0]].sizeY)

        movelistIndex = 0

        # AB's

        if movelist[movelistIndex] == (25,3):
            movelistIndex = movelistIndex + 1
            while movelist[movelistIndex][0] != 25:
                x, y = movelist[movelistIndex]
                movelistIndex = movelistIndex + 1

                for m in dicts.keys():
                    if Xinterv[m][0] <= x < Xinterv[m][1] \
                       and Yinterv[m][0] <= y < Yinterv[m][1]:     # stone played in
                                                                   # relevant region
                        if not dicts[m].has_key((x,y)):       
                            if not (self.finalpos[index*150 + 100 + y/4 + 5*(x/2)] & (1 << (x%2 + 2*(y%4)))):
                                                                   # can't become a match
                                del dicts[m]
                                continue
                            else:
                                dicts[m][(x,y)] = '.'
                                dictsNO[m] = dictsNO[m] + 1
                        elif dicts[m][(x,y)] == 'X':
                            del dicts[m][(x,y)]
                            dictsNO[m] = dictsNO[m] - 1
                
        # AW's

        if movelist[movelistIndex] == (25,4):
            movelistIndex = movelistIndex + 1
            while movelist[movelistIndex][0] != 25:
                x, y = movelist[movelistIndex]
                movelistIndex = movelistIndex + 1

                for m in dicts.keys():
                    if Xinterv[m][0] <= x < Xinterv[m][1] \
                       and Yinterv[m][0] <= y < Yinterv[m][1]:
                        if not dicts[m].has_key((x,y)):
                            if not (self.finalpos[index*150 + 100 + y/4 + 5*(x/2)] & (1 << (x%2 + 2*(y%4)))):
                                del dicts[m]
                                continue
                            else:
                                dicts[m][(x,y)] = '.'
                                dictsNO[m] = dictsNO[m] + 1
                        elif dicts[m][(x,y)] == 'O':
                            del dicts[m][(x,y)]
                            dictsNO[m] = dictsNO[m] - 1

        for m in dicts.keys():
            if dictsNO[m] == 0:
                dicts[m]['found'] = 0
        
        # B's and W's

        if caplist: 
            caplistIndex = 0        
            caps = 1
        else:
            caps = 0
            
        if movelist[movelistIndex] == (25,1):
            co = 'B'
        elif movelist[movelistIndex] == (25,2):
            co = 'W'
        else:
            showwarning('Error', 'No moves in this file')
            return
        
        moveLimit = self.moveLimit.get()
        if moveLimit == 250:
            moveLimit = 1000          # practically no limit
        else:
            moveLimit = moveLimit + movelistIndex


        counter = -1
        while movelistIndex < len(movelist)-1 and movelistIndex < moveLimit:
            movelistIndex = movelistIndex + 1
            counter = counter + 1
            
            # man koennte hier noch mal movelist[movelistIndex][0] == 25 checken
            # (i.e. ob noch AB's oder so mitten im SGF file)

            if co == 'B':
                co = 'W'
                c = 'X'
                cinv = 'O' 
            else:
                co = 'B'
                c = 'O'
                cinv = 'X'

            x, y = movelist[movelistIndex]

            if caps and counter == caplist[caplistIndex][0]:
                captures = caplist[caplistIndex][1]
                if caplistIndex < len(caplist)-1:
                    caplistIndex = caplistIndex + 1
                else:
                    caps = 0
            else:
                captures = []
                
            for m in dicts.keys():
                if Xinterv[m][0] <= x < Xinterv[m][1] \
                   and Yinterv[m][0] <= y < Yinterv[m][1]:

                    if dicts[m].has_key('found'):   # so (x,y) is the continuation
                        
                        xx, yy = Pattern.flips[PatternInvFlip(patternList[m[0]].flip)]\
                                 (x, y)
                                
                        XX1, YY1 = Pattern.flips[PatternInvFlip(patternList[m[0]].flip)]\
                                   (Xinterv[m][0], Yinterv[m][0])
                        XX2, YY2 = Pattern.flips[PatternInvFlip(patternList[m[0]].flip)]\
                                   (Xinterv[m][1]-1, Yinterv[m][1]-1)
                        XX = min(XX1, XX2)
                        YY = min(YY1, YY2)
                                
                        xx = xx - XX
                        yy = yy - YY
                      
                        if (patternList[m[0]].colorSwitch and co == 'B') or \
                           (not patternList[m[0]].colorSwitch and co == 'W'):
                            cc = 'B'
                        else: cc = 'W'

                        (xx, yy), cSymm = self.symmetries[(xx,yy)]

                        if (cc == 'B' and not cSymm) or (cc=='W' and cSymm): cc = 'B'
                        else: cc = 'W'

                        if self.nextMoveVar.get():
                            if ((self.nextMoveVar.get() == 2 and cc == 'B') \
                                or (self.nextMoveVar.get() == 1 and cc == 'W')):
                                if self.symmetries[-1] != -1 and not fixedColor:

                                    if cc == 'B': cc='W'
                                    else: cc = 'B'

                                    
                                    xx += XX
                                    yy += YY

                                    xx, yy = Pattern.flips[self.symmetries[-1]](xx, yy)
                                    
                                    XX1, YY1 = Pattern.flips[self.symmetries[-1]](XX1, YY1)
                                    XX2, YY2 = Pattern.flips[self.symmetries[-1]](XX2, YY2)

                                    XX = min(XX1, XX2)
                                    YY = min(YY1, YY2)

                                    xx -= XX
                                    yy -= YY                                    
                                    
                                    (xx1, yy1), cSymm1 = self.symmetries[(xx,yy)]
                                    if not cSymm1:
                                        xx, yy = xx1, yy1

                                    colorSwitch = 1-cSymm
                                    if colorSwitch: noSwitched += 1
                                else:
                                    del dicts[m]
                                    del dictsNO[m]
                                    continue
                            else:
                                colorSwitch = cSymm
                                if colorSwitch: noSwitched += 1
                        else:
                            colorSwitch = cSymm

                        

                        if self.continuations.has_key((xx, yy)):
                            self.continuations[(xx,yy)][cc] = self.continuations[(xx,yy)][cc]+1
                        else:
                            if self.contLabelsIndex >= len(self.contLabels): text = '?'
                            else:
                                text = self.contLabels[self.contLabelsIndex]
                                self.contLabelsIndex += 1
                            if cc == 'B':
                                self.continuations[(xx,yy)] = {'B':1, 'W':0,           # B/W continuations at this point
                                                               'tB':0, 'tW':0,         # tenukis among the above
                                                               'wB':0, 'lB':0, 'wW':0, 'lW':0, # B wins/losses
                                                                                               # among B/W continuations
                                                               'N': text}
                            else:
                                self.continuations[(xx,yy)] = {'B':0, 'W':1, 'tB':0, 'tW':0, 'wB':0, 'lB':0,
                                                               'wW':0, 'lW':0, 'N': text}
                          
                        if not dicts[m]['found'] in [counter, counter+1]:
                            self.continuations[(xx,yy)]['t'+cc] += 1

                        if db['data'][index][4] == 'B':
                            if not (patternList[m[0]].colorSwitch or colorSwitch):
                                self.continuations[(xx,yy)]['w'+cc] += 1
                            else:
                                self.continuations[(xx,yy)]['l'+cc] += 1
                        elif db['data'][index][4] == 'W':
                            if not (patternList[m[0]].colorSwitch or colorSwitch):
                                self.continuations[(xx,yy)]['l'+cc] += 1
                            else:
                                self.continuations[(xx,yy)]['w'+cc] += 1

                        noSwitched += patternList[m[0]].colorSwitch

                        if (patternList[m[0]].colorSwitch or colorSwitch) and self.options.showColorSwap.get():
                            result.append((dicts[m]['found'], self.continuations[(xx,yy)]['N']+'-'))
                        else:
                            result.append((dicts[m]['found'], self.continuations[(xx,yy)]['N']))
                        del dicts[m]
                        del dictsNO[m]
                        continue     # with for m in dicts.keys()
                    
                    if not dicts[m].has_key((x,y)):
                        if not (self.finalpos[index*150 + 100 + y/4 + 5*(x/2)] & (1 << (x%2 + 2*(y%4)))):
                                                                  # this stone won't be captured later,
                                                                  # so this cannot become a match
                            del dicts[m]
                            continue
                        else:
                            dicts[m][(x,y)] = '.'
                            dictsNO[m] = dictsNO[m] + 1
                    elif dicts[m][(x,y)] == c:
                        del dicts[m][(x,y)]
                        dictsNO[m] = dictsNO[m] - 1

                if not dicts[m].has_key('found'):
                    for cap in captures:
                        xx = cap[0]-1
                        yy = cap[1]-1
                        if Xinterv[m][0] <= xx < Xinterv[m][1] \
                           and Yinterv[m][0] <= yy < Yinterv[m][1]:
                            if not dicts[m].has_key((xx,yy)):
                                dicts[m][(xx,yy)] = cinv
                                dictsNO[m] = dictsNO[m] + 1
                            elif dicts[m][(xx,yy)] == '.':
                                del dicts[m][(xx,yy)]
                                dictsNO[m] = dictsNO[m] - 1

                    if dictsNO[m] == 0:
                        dicts[m]['found'] = counter + 1

            if dicts=={}: break

        if not self.nextMoveVar.get():
            for m in dicts.keys():
                if dicts[m].has_key('found'):
                    if patternList[m[0]].colorSwitch: noSwitched += 1
                    if patternList[m[0]].colorSwitch and self.options.showColorSwap.get():
                        result.append((dicts[m]['found'], '-'))
                    else:
                        result.append((dicts[m]['found'], ''))

        if result:
            self.noMatches = self.noMatches + len(result)
            self.noSwitched = self.noSwitched + noSwitched
            
            if db['data'][index][4] == 'B':
                self.Bwins = self.Bwins + len(result) - noSwitched
                self.Wwins = self.Wwins + noSwitched
            elif db['data'][index][4] == 'W':
                self.Bwins = self.Bwins + noSwitched
                self.Wwins = self.Wwins + len(result) - noSwitched

            result.sort()
            self.gamelist.appendC(dbIndex, index, join([str(x[0])+x[1] for x in result], ', '))
            return 1


    def match(self, patternList):
        """This is part of the search engine. See search for a description. """ 
        
        ctr = 0
        noGames = reduce(lambda x,y:x+y, [len(db['helplist']) for db in self.gamelist.DBlist if not db['disabled']], 0)

        moveLimit = self.moveLimit.get()

        hashable = ()

        if self.sel[0] == (1,1) and self.sel[1] == (19,19):
            hashable = ((0,0,18,18),)
        if self.sel[0] == (1,1) and self.sel[1][0] >= 7 and self.sel[1][1] >= 7:
            hashable += ((0,0,6,6),)
        if self.sel[1] == (19,19) and self.sel[0][0] <= 13 and self.sel[0][1] <= 13:
            hashable += ((12,12,18,18),)
        if self.sel[0][0] == 1 and self.sel[1][1] == 19 and self.sel[1][0] >= 7 and self.sel[0][1] <= 13:
            hashable += ((12,0,18,6),)
        if self.sel[0][1] == 1 and self.sel[1][0] == 19 and self.sel[0][0] <= 13 and self.sel[1][1] >= 7:
            hashable += ((0,12,6,18),)
                
        if hashable:
            for pattern in patternList:
                if not pattern.compute_hashkey(hashable):      # wildcards in the pattern, or too many stones?
                    hashable = ()
                    break

        if not hashable:
            if self.Csearch.get():
                patternStrL = []
                PSizeL = []
                PInvFlipsL = []
                PColorSwitchL = []
                PAnchorsL = []
                PBitsTable = array('I')
                PBitsL = array('B')
                
                for pattern in patternList:
                    patternStrL.append(`pattern`)
                    PSizeL.append(chr(pattern.sizeX) + chr(pattern.sizeY))
                    PInvFlipsL.append(chr(PatternInvFlip(pattern.flip)))
                    PColorSwitchL.append(chr(pattern.colorSwitch))
                    PAnchorsL.append(chr(pattern.anchors[0][0]) + chr(pattern.anchors[0][1]) \
                                     + chr(pattern.anchors[1][0]) + chr(pattern.anchors[1][1]))
                    
                    for pb in pattern.bits:
                        PBitsTable.append(len(PBitsL))
                        PBitsL.extend(pb)

                patternStr = join(patternStrL, '')
                PSize = join(PSizeL, '')
                PInvFlips = join(PInvFlipsL, '')
                PColorSwitch = join(PColorSwitchL, '')
                PAnchors = join(PAnchorsL, '')
                hash_indices = array('I')
                hashC = array('B')
                
        for dbIndex in range(len(self.gamelist.DBlist)):
            db = self.gamelist.DBlist[dbIndex]
            
            if db['disabled'] or len(db['helplist']) == 0: continue

            self.mainlistArr, self.posTable, self.finalpos, hashT, hash = self.gamelist.readDB(db['name'])
            if not self.finalpos: continue

            if hashable:
                possiblePatterns = []
                for N in range(len(patternList)):

                    occurs = 1
                    noOccurrences = len(db['data'])+1 +1 # 'safety margin' because of bug in 0.4b

                    for hk in range(len(patternList[N].hashkey)):
                        start = 0
                        end = len(hashT)/2
                    
                        while start < end:
                            help = hashT[2*((start+end)/2)]
                            if help == patternList[N].hashkey[hk]:
                                hash_start = hashT[2*((start+end)/2)+1]
                                if 2*((start+end)/2)+3 >= len(hashT):
                                    hash_end = len(hash)
                                else: hash_end = hashT[2*((start+end)/2)+3]

                                if (hash_end - hash_start)/2 < noOccurrences:
                                    noOccurrences = (hash_end - hash_start)/2
                                    patternList[N].hash_start = hash_start
                                    patternList[N].hash_end = hash_end
                                break
                            elif help > patternList[N].hashkey[hk]:
                                end = (start+end)/2
                            elif help < patternList[N].hashkey[hk]:
                                start = (start+end)/2 + 1

                        if start >= end: occurs = 0

                    if occurs: possiblePatterns.append(N)

                if self.Csearch.get():
                    patternStrL = []
                    PSizeL = []
                    PInvFlipsL = []
                    PColorSwitchL = []
                    PAnchorsL = []
                    PBitsTable = array('I')
                    PBitsL = array('B')
                    hash_indices = array('I')
                    hashC = array('B')
                    
                    for N in possiblePatterns:
                        pattern = patternList[N]
                        patternStrL.append(`pattern`)
                        PSizeL.append(chr(pattern.sizeX) + chr(pattern.sizeY))
                        PInvFlipsL.append(chr(PatternInvFlip(pattern.flip)))
                        PColorSwitchL.append(chr(pattern.colorSwitch))
                        PAnchorsL.append(chr(pattern.anchors[0][0]) + chr(pattern.anchors[0][1]) \
                                         + chr(pattern.anchors[1][0]) + chr(pattern.anchors[1][1]))

                        hash_indices.append(len(hashC))
                        hashC.extend(hash[pattern.hash_start:pattern.hash_end])
                        hash_indices.append(len(hashC))

                        for pb in pattern.bits:
                            PBitsTable.append(len(PBitsL))
                            PBitsL.extend(pb)
                            
                    patternStr = join(patternStrL, '')
                    PSize = join(PSizeL, '')
                    PInvFlips = join(PInvFlipsL, '')
                    PColorSwitch = join(PColorSwitchL, '')
                    PAnchors = join(PAnchorsL, '')
                else:
                    hashInd = {}
                    for N in possiblePatterns: hashInd[N] = patternList[N].hash_start
                    
            else:
                possiblePatterns = range(len(patternList))


            search_start = time.time()

            if self.Csearch.get():

                pB_start = ctr * 1.0 / noGames
                pB_end = (ctr + len(db['helplist'])*0.9) * 1.0 / noGames

                try:

                    games, results, noSwitched, self.contLabelsIndex, Wwins, Bwins, noMatches = \
                           matchC.match(patternStr, PSize, PInvFlips, PColorSwitch,
                                        PAnchors, PBitsTable.tostring(), PBitsL.tostring(), moveLimit,
                                        db['helplist'].tostring(),
                                        self.finalpos.tostring(), 
                                        self.mainlistArr.tostring(), self.posTable.tostring(),
                                        hash_indices.tostring(), hashC.tostring(),
                                        self.progBar.redraw, pB_start, pB_end,
                                        join(self.contLabels, ''), self.contLabelsIndex,
                                        self.continuations, self.symmetries, db['data'],
                                        self.options.showColorSwap.get(), self.nextMoveVar.get(),
                                        self.fixedColorVar.get())
                    
                    db['current'].fromstring(games)
                    db['results'] = results
                    
                    self.noMatches += noMatches
                    self.noSwitched += noSwitched
                    self.Wwins += Wwins
                    self.Bwins += Bwins
                    
                    self.gamelist.appendClist += [ (dbIndex, db['current'][j], j) for j in range(len(db['current'])) ]
                
                except:
                    showwarning('Error', 'An error occurred during the pattern search.')
                    pass
                
            else:
 
                for index in db['helplist']:
                    matchList = []
                    
                    for N in possiblePatterns:
                       
                        pattern = patternList[N]
                        
                        if hashable:
                            while hashInd[N]+1 < pattern.hash_end and index > 256*hash[hashInd[N]] + hash[hashInd[N]+1]:
                                hashInd[N] += 2

                            if index < 256*hash[hashInd[N]] + hash[hashInd[N]+1]: continue
                            elif index == 256*hash[hashInd[N]] + hash[hashInd[N]+1]:
                                hashInd[N] += 2
                                matchList.append((N,(pattern.anchors[0][0],pattern.anchors[0][1])))

                        else:
                            
                            for a0 in range(pattern.anchors[0][0], pattern.anchors[1][0]+1):
                                for a1 in range(pattern.anchors[0][1], pattern.anchors[1][1]+1):
                                    matches = 1
                                    
                                    pIndex = 2*(a1%2) + (a0%2)
                                    patternBits = pattern.bits[pIndex]
                                
                                    pbIndex = 0
                                    fpIndex = index*150 + a1/2 + (a0/2)*10

                                    for x in range(patternBits[0]):
                                        start = patternBits[pbIndex+1]
                                        length = patternBits[pbIndex+2]
                                        fpIndex += start
                                        pbIndex += 2
                                        for y in range(length):
                                            pbIndex += 1
                                            if (patternBits[pbIndex] &
                                                self.finalpos[fpIndex]):
                                                matches = 0
                                                break
                                            fpIndex += 1
                                        if not matches: break
                                        fpIndex += 10 - start - length
                            
                                    if matches:                       # found candidate
                                        matchList.append((N,(a0,a1)))

                    if matchList:
                        self.checkCandidate(dbIndex, index, patternList, matchList, self.fixedColorVar.get())

            self.searchtime += time.time() - search_start

            ctr = ctr + len(db['helplist'])
            # self.gamelist.update()
            self.progBar.redraw(ctr*1.0/noGames)
            self.searchWindow.update()

            del self.finalpos
            del self.mainlistArr
            del self.posTable
            

    def search(self):
        """ The search function. After some preparation (setting up the list of
            relevant patterns etc.), the search proceeds as follows:

            - the match function looks at each game, and produces a list of
              candidates, i.e. of positions on the board, where the pattern could have
              occured, considering the final position of this game.
            - if there are any candidates, checkCandidate is invoked. There, the game
              is played through until all candidates have been checked.

            In the C++ search function, the match and checkCandidate methods are
            combined. """

        if not self.gamelist.noOfGames(): return

        self.configButtons(DISABLED)
        self.gamelist.clearGameInfo()

        boardData = self.board.snapshot()
        boardData2 = (copy(self.board.status), copy(self.board.wildcards), self.board.selection)

        currentTime = time.time()
                
        self.progBar.clear()

        for db in self.gamelist.DBlist:
            if db['disabled']: continue
            db['helplist'] = db['current']
        self.gamelist.clear()
        
        self.board.delLabels()

        d = {}

        if self.board.selection[0][0] > self.board.selection[1][0] or \
           self.board.selection[0][1] > self.board.selection[1][1]:
            self.board.selection = ((1,1), (19, 19))

        self.sel = self.board.selection # copy this because the selection on the board may
                                        # be changed by the user although the search is not yet
                                        # finished

        stoneCount = 0

        for i in range(self.sel[0][1], self.sel[1][1]+1):
            for j in range(self.sel[0][0], self.sel[1][0]+1):
                if not self.board.status.has_key((j,i)): d[(i-1,j-1)] = '.'
                elif self.board.status[(j,i)] == 'black':
                    d[(i-1,j-1)] = 'X'
                    stoneCount += 1
                else:
                    d[(i-1,j-1)] = 'O'
                    stoneCount += 1
                if self.board.wildcards.has_key((j,i)): d[(i-1,j-1)] = '*'

        # anchors:

        if self.sel[0] == (1,1) or \
           (self.sel[0][0], self.sel[1][1]) == (1,19) or \
           (self.sel[1][0], self.sel[0][1]) == (19,1) or \
           self.sel[1] == (19,19) or \
           self.fixedAnchorVar.get():                          # corner, or fixedAnchor
                                                               # so no translation occurs
            anchors = [(self.sel[0][1]-1, self.sel[0][0]-1),
                       (self.sel[0][1]-1, self.sel[0][0]-1)]
        elif 1 in self.sel[0] or 19 in self.sel[1]:            # translation along the edge only
            anchors = []
            if self.sel[0][0] == 1:
                anchors = [(1,0), (17 - self.sel[1][1] + self.sel[0][1] ,0)]
            elif self.sel[0][1] == 1:
                anchors = [(0,1), (0,17 - self.sel[1][0] + self.sel[0][0])]
            elif self.sel[1][0] == 19:
                anchors = [(1,self.sel[0][0]-1),
                           (17-self.sel[1][1]+self.sel[0][1],self.sel[0][0]-1)]
            elif self.sel[1][1] == 19:
                anchors = [(self.sel[0][1]-1,1),
                           (self.sel[0][1]-1,17-self.sel[1][0]+self.sel[0][0])]
        else:                                                        # full translation
            anchors = [(1, 1), (17-self.sel[1][1]+self.sel[0][1],
                                17-self.sel[1][0]+self.sel[0][0])]

            if self.options.askBeforeLongSearch.get() and \
               ( (stoneCount <= 1) or \
                 (stoneCount <= 2 and (self.sel[1][1]-self.sel[0][1]+1)*(self.sel[1][0]-self.sel[0][0]+1) < 30) or\
                 (stoneCount <= 3 and (self.sel[1][1]-self.sel[0][1]+1)*(self.sel[1][0]-self.sel[0][0]+1) < 16) ):
                if not askokcancel('Time consuming search', 'This search may take quite some time. Continue?'):
                    return

        self.currentSearchPattern = Pattern(anchors, d)

        pList, self.symmetries = self.currentSearchPattern.patternList(self.fixedColorVar.get(),
                                                                       self.nextMoveVar.get())

        self.continuations = {}
        self.statistics = {}
        
        if self.options.uppercaseLabels.get():
            self.contLabels = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L' ,'M' ,'N',
                               'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'Y', 'Z', 'a', 'b', 'c', 'd',
                               'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r',
                               's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '1', '2', '3', '4', '5', '6',
                               '7', '8', '9']
        else:
            self.contLabels = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
                               'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B',
                               'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'P', 'Q',
                               'R', 'S', 'T', 'U', 'V', 'W', 'Y', 'Z', '1', '2', '3', '4', '5', '6',
                               '7', '8', '9']

        self.contLabelsIndex = 0

        try:
            if self.cursor and self.cursor.currentNode().has_key('LB'):
                labelsInSGFFile = 1
                for p1 in self.cursor.currentNode()['LB']:
                    p, text = split(p1, ':')
                    x,y = self.convCoord(p)
                    if self.sel[0][0] <= x <= self.sel[1][0] and self.sel[0][1] <= y <= self.sel[1][1]:
                        self.continuations[(y-self.sel[0][1], x-self.sel[0][0])] = \
                                                              {'B':0, 'W':0, 'tB':0, 'tW':0,
                                                               'wB':0, 'lB':0, 'wW':0, 'lW':0,
                                                               'N': text}
                        if text in self.contLabels: self.contLabels.remove(text)
            else: labelsInSGFFile = 0
        except:
            showwarning('Error', 'SGF Error')
            labelsInSGFFile = 0
            
        self.noMatches = 0
        self.noSwitched = 0
        self.Bwins = 0
        self.Wwins = 0

        self.searchtime = 0
        self.match(pList)

        if not labelsInSGFFile:

            contList = [ (self.continuations[k]['B']+self.continuations[k]['W'], k) \
                         for k in self.continuations.keys()
                         if self.continuations[k]['N'] != '?' ]
            contList.sort()
            contList.reverse()
        
            self.contLabelsIndex = 0
            t = []
            for dummy, k in contList:
                t.append(self.continuations[k]['N'])
                self.continuations[k]['N'] = self.contLabels[self.contLabelsIndex]
                self.contLabelsIndex += 1

            trans = maketrans(join(t, ''), join(self.contLabels[:self.contLabelsIndex], ''))

            for db in self.gamelist.DBlist:
                s1 = join(db['results'], '#')
                s2 = translate(s1, trans)
                db['results'] = split(s2, '#')
            
        self.gamelist.update()

        if self.showContinuation.get(): self.showCont()
        elif self.cursor:
            try:
                self.leaveNode()
                self.displayLabels(self.cursor.currentNode())
            except: pass
            
        self.displayStatistics()

        self.progBar.redraw(1)
        self.progBar.write('%1.1f seconds' % (time.time() - currentTime))

        # append the result of this search to self.prevSearches
        
        games = [l['current'] for l in self.gamelist.DBlist]
            # This should really be copy (resp. deepcopy), but that would
            # be too slow, and it hopefully works this way too. One has
            # to be careful with self.gamelist.appendC
            # which changes these lists in place, though
        results = [join(l['results'], '|') for l in self.gamelist.DBlist]
        
        cursorSn = [self.cursor, self.cursor.currentGame, self.cursor.currentN.pathToNode()]

        self.prevSearches.append([boardData, games, results, copy(self.continuations),
                               self.noMatches, self.noSwitched, self.Bwins, self.Wwins,
                               self.modeVar.get(), self.fixedColorVar.get(), self.fixedAnchorVar.get(),
                               self.moveLimit.get(), self.nextMoveVar.get(), cursorSn], boardData2)

        for db in self.gamelist.DBlist:
            if not db['disabled']: del db['helplist']

        self.configButtons(NORMAL)


    def displayStatistics(self):
        """ Display statistical information on the last search on
        self.statisticsCanv. """
        
        self.statisticsCanv.delete('stat')

        if self.noMatches:

            font = (self.options.statFont0.get(), self.options.statFont1.get(), self.options.statFont2.get())
            smallfont = (self.options.statFont0.get(), self.options.statFont3.get(),
                         self.options.statFont2.get())
            
            Bperc = self.Bwins * 100.0 / self.noMatches
            Wperc = self.Wwins * 100.0 / self.noMatches

            self.statistics['total'] = '%d matches (%d/%d), B: %1.1f%%, W: %1.1f%%' % \
                                       (self.noMatches, self.noMatches-self.noSwitched,
                                        self.noSwitched, Bperc, Wperc)

            self.statisticsCanv.create_text(5, 5, text=self.statistics['total'],
                                            font=font, anchor='nw', tags='stat')
               
            used = []

            for i in range(len(self.continuations.keys())):
                max = -1
                maxN = ''
                index = ()
                for x,y in self.continuations.keys():
                    if not (x,y) in used and \
                       (self.continuations[(x,y)]['B'] + self.continuations[(x,y)]['W'] > max or
                        (self.continuations[(x,y)]['B'] + self.continuations[(x,y)]['W'] == max and
                         self.continuations[(x,y)]['N'] < maxN)):
                        index = (x,y)
                        max = self.continuations[(x,y)]['B'] + self.continuations[(x,y)]['W']
                        maxN = self.continuations[(x,y)]['N']
                         
                if i == 0: maxHeight = max

                x, y = index
                used.append(index)

                self.statistics[i] = [(x,y), self.continuations[(x,y)] ]

                if i >= 10: continue

                heightB = int((self.continuations[(x,y)]['B']-self.continuations[(x,y)]['tB']) * 1.0/maxHeight * 125)
                heighttB = int(self.continuations[(x,y)]['tB']*1.0/maxHeight * 125)
                heightW = int((self.continuations[(x,y)]['W']-self.continuations[(x,y)]['tW']) * 1.0/maxHeight * 125)
                heighttW = int(self.continuations[(x,y)]['tW']*1.0/maxHeight * 125)
                
                self.statisticsCanv.create_text(22 + i*22, 173, text=self.continuations[(x,y)]['N'],
                                                font=font, tags='stat')

                if self.continuations[(x,y)]['W']:
                    Wp = '%1.1f' % (self.continuations[(x,y)]['wW']*100.0/self.continuations[(x,y)]['W'])
                else: Wp = '-'
                self.statisticsCanv.create_text(22 + i*22, 185, text=Wp, font=smallfont, tags='stat')

                if self.continuations[(x,y)]['B']:
                    Bp = '%1.1f' % (self.continuations[(x,y)]['wB']*100.0/self.continuations[(x,y)]['B'])
                else: Bp = '-'
                self.statisticsCanv.create_text(22 + i*22, 193, text=Bp, font=smallfont, tags='stat')

                if heighttB:
                    self.statisticsCanv.create_rectangle(12+i*22, 165-heighttB, 28+i*22, 165,
                                                         fill=self.options.Btenuki.get(),
                                                         outline='', tags='stat')
                if heightB:
                    self.statisticsCanv.create_rectangle(12+i*22, 165-heightB-heighttB, 28+i*22, 165-heighttB,
                                                         fill='black', outline='', tags='stat')
                if heightW:
                    self.statisticsCanv.create_rectangle(12+i*22, 165-heightB-heighttB-heightW, 28+i*22,
                                                         165-heightB-heighttB,
                                                         fill='white', outline='', tags='stat')
                if heighttW:
                    self.statisticsCanv.create_rectangle(12+i*22, 165-heighttB-heightB-heightW-heighttW, 28+i*22,
                                                         165-heightB-heighttB-heightW,
                                                         fill=self.options.Wtenuki.get(),
                                                         outline='', tags='stat')
                    
                self.statisticsCanv.create_text(22 + i*22, 157-heightB-heighttB-heightW-heighttW,
                                                font = font,
                                                text=`self.continuations[(x,y)]['B']+self.continuations[(x,y)]['W']`,
                                                tags='stat')


    def clearGI(self):
        self.pbVar.set('')
        self.pwVar.set('')
        self.pVar.set('')
        self.evVar.set('')
        self.frVar.set('')
        self.toVar.set('')
        self.awVar.set('')
        self.referencedVar.set(0)


    def historyGI_back(self):
        if not self.history_GIsearch: return
        self.history_GIS_index -= 1
        if self.history_GIS_index == -1: self.history_GIS_index = len(self.history_GIsearch)-1

        varList = [self.pbVar, self.pwVar, self.pVar, self.evVar, self.frVar, self.toVar, self.awVar, self.referencedVar]

        for i in range(len(varList)):
            var = varList[i]
            var.set(self.history_GIsearch[self.history_GIS_index][i])


    def historyGI_fwd(self):
        if not self.history_GIsearch: return
        self.history_GIS_index += 1
        if self.history_GIS_index == len(self.history_GIsearch): self.history_GIS_index = 0
        varList = [self.pbVar, self.pwVar, self.pVar, self.evVar, self.frVar, self.toVar, self.awVar, self.referencedVar]

        for i in range(len(varList)):
            var = varList[i]
            var.set(self.history_GIsearch[self.history_GIS_index][i])


    def doGISearch(self):
        """ Carry out the search for the parameters in *Var. All non-empty Var's
        have to match at the same time. In the case of pbVar, pwVar, pVar, evVar,
        the string has to occur somewhere in PB[ ...] (...), not necessarily at the
        beginning. For the date, the first four-digit number in DT[] is compared
        to frVar and toVar."""

        if not self.gamelist.noOfGames(): return

        self.statistics = {}

        pbVar = self.pbVar.get().encode('utf-8')
        pwVar = self.pwVar.get().encode('utf-8')
        pVar  = self.pVar.get().encode('utf-8')
        evVar = self.evVar.get().encode('utf-8')
        frVar = self.frVar.get()
        toVar = self.toVar.get()
        awVar = self.awVar.get().encode('utf-8')
        refVar = self.referencedVar.get()

        self.history_GIsearch.append((pbVar, pwVar, pVar, evVar, frVar, toVar, awVar, refVar))
        self.history_GIS_index = len(self.history_GIsearch) - 1

        if self.options.caseInsensitive.get():
            pbVar = lower(pbVar)
            pwVar = lower(pwVar)
            pVar = lower(pVar)
            evVar = lower(evVar)
            awVar = lower(awVar)
        
        if frVar:
            if not (re.match('\d\d\d\d-\d\d-\d\d', frVar) or re.match('\d\d\d\d-\d\d', frVar) \
                    or re.match('\d\d\d\d', frVar)):
                frVar = ''

        if toVar:
            if re.match('\d\d\d\d-\d\d-\d\d', toVar) or re.match('\d\d\d\d-\d\d', toVar) or re.match('\d\d\d\d', toVar):
                toVar += 'Z'
            else:
                toVar = ''

        if not (pbVar or pwVar or pVar or evVar or frVar or toVar or awVar or refVar): return
        
        self.gamelist.clearGameInfo()
        self.configButtons(DISABLED)
        ctr = 0
        noGames = self.gamelist.noOfGames()
        self.progBar.clear()
        currentTime = time.time()

        self.board.delLabels()
        if self.cursor:
            try:
                self.leaveNode()
                self.displayLabels(self.cursor.currentNode())
            except:
                showwarning('Error', 'SGF Error')
                
        for db in self.gamelist.DBlist:
            if not db['disabled']:
                db['helplist'] = db['current']
        self.gamelist.clear() 
        
        for dbIndex in range(len(self.gamelist.DBlist)):
            db = self.gamelist.DBlist[dbIndex]
            if db['disabled']: continue

            oldfilename = ''
            
            for gIndex in range(len(db['helplist'])):
                g = db['helplist'][gIndex]
                
                ctr = ctr + 1
                if ctr % 10 == 0:
                    # if ctr % 1000 == 500:
                    #    self.gamelist.update()
                    self.progBar.redraw(ctr*1.0/noGames)
                    self.searchWindow.update()

                if self.options.caseInsensitive.get():
                    
                    if pwVar and find(lower(db['data'][g][3]), pwVar) == -1: continue
                    if pbVar and find(lower(db['data'][g][2]), pbVar) == -1: continue
                    if pVar and find(lower(db['data'][g][3]), pVar) == -1 \
                       and find(lower(db['data'][g][2]), pVar) == -1: continue

                    if frVar or toVar:
                        date = db['data'][g][6]
                    
                        if frVar and not date >= frVar: continue
                        if toVar and not date <= toVar: continue

                    if refVar and not db['data'][g][7]: continue

                    if evVar or awVar:
                        f1 = os.path.join(db['sgfpath'], db['data'][g][0])
                        if find(f1, '[') != -1:
                            f1, f2 = split(f1, '[')
                            gameNumber = int(strip(f2)[:-1])
                        else: gameNumber = -1
                
                        filename = getFilename(f1)

                        if not oldfilename == filename:
                            oldfilename = filename
                            try:
                                file = open(filename)
                                sgf = file.read()
                                file.close()
                            except IOError:
                                showwarning('IO Error', 'File not found: ' + filename)
                                continue
                            newfile = 1
                        else: newfile = 0
                            
                        
                        if gameNumber == -1:         # no collection, so we get away without parsing the SGF file

                            # is there an encoding tag?

                            enc_begin = find(sgf, 'CA[')
                            if enc_begin != -1:
                                encoding = lower(strip(sgf[enc_begin+3: find(sgf, ']', enc_begin)]))
                            else:
                                encoding = db['encoding']

                            sgf = lower(sgf)
                            if encoding != 'utf-8':
                                try: sgf = sgf.decode(encoding, 'ignore').encode('utf-8', 'ignore')
                                except: pass
                                
                            if awVar and find(sgf, awVar) == -1: continue

                            ok = 1
                            for var,tag in [(evVar,'ev[')]:          # note that we have already lower'ed sgf!
                                if var:
                                    v_begin = find(sgf, tag)
                                    if v_begin == -1: ok = 0
                                    else:
                                        v_text = sgf[v_begin: find(sgf, ']', v_begin)]
                                        if find(v_text, var) == -1: ok = 0
                            if not ok: continue

                        else:
                            if newfile:
                                try:
                                    c = Cursor(sgf, 1)
                                except:
                                    showwarning('Error', 'SGF error in file ' + filename)
                                    continue
                            
                            node = c.getRootNode(gameNumber)

                            encoding = c.encoding or db['encoding']
                            
                            ok = 1
                            for var,tag in [(evVar,'EV')]:
                                try: var = var.decode('utf-8', 'ignore').encode(encoding, 'ignore')
                                except: pass
                                if var and not (node.has_key(tag) and find(lower(node[tag][0]), var) != -1):
                                    ok = 0
                            if not ok: continue

                            if awVar:
                                try: awVar = awVar.decode('utf-8', 'ignore').encode(encoding, 'ignore')
                                except: pass
                                c.game(gameNumber)
                                s = lower(c.outputVar(c.currentN))
                                if find(s, awVar) == -1: continue

                else:
                    
                    if pwVar and find(db['data'][g][3], pwVar) == -1: continue
                    if pbVar and find(db['data'][g][2], pbVar) == -1: continue
                    if pVar and find(db['data'][g][3], pVar) == -1 \
                       and find(db['data'][g][2], pVar) == -1: continue

                    if frVar or toVar:
                        date = db['data'][g][6]
                    
                        if frVar and not date >= frVar: continue
                        if toVar and not date <= toVar: continue

                    if refVar and not db['data'][g][7]: continue

                    if evVar or awVar:
                        f1 = os.path.join(db['sgfpath'], db['data'][g][0])
                        if find(f1, '[') != -1:
                            f1, f2 = split(f1, '[')
                            gameNumber = int(strip(f2)[:-1])
                        else: gameNumber = -1
                        
                        filename = getFilename(f1)

                        if not oldfilename == filename:
                            oldfilename = filename
                            try:
                                file = open(filename)
                                sgf = file.read()
                                file.close()
                            except IOError:
                                showwarning('IO Error', 'File not found: ' + filename)
                                continue
                            newfile = 1
                        else:
                            newfile = 0
                            
                        if gameNumber == -1:   

                            # is there an encoding tag?

                            enc_begin = find(sgf, 'CA[')
                            if enc_begin != -1:
                                encoding = lower(strip(sgf[enc_begin+3: find(sgf, ']', enc_begin)]))
                            else:
                                encoding = db['encoding']
                                
                            if encoding != 'utf-8':
                                try: sgf = sgf.decode(encoding, 'ignore').encode('utf-8', 'ignore')
                                except: pass

                            if awVar and find(sgf, awVar) == -1: continue

                            ok = 1
                            for var,tag in [(evVar,'EV[')]:
                                if var:
                                    v_begin = find(sgf, tag)
                                    if v_begin == -1: ok = 0
                                    else:
                                        v_text = sgf[v_begin: find(sgf, ']', v_begin)]
                                        if find(v_text, var) == -1: ok = 0
                            if not ok: continue

                        else:
                            if newfile:
                                try: c = Cursor(sgf, 1)
                                except:
                                    showwarning('Error', 'SGF error in file ' + filename)
                                    continue
                                
                            node = c.getRootNode(gameNumber)

                            encoding = c.encoding or db['encoding']
                            
                            ok = 1
                            for var,tag in [(evVar,'EV')]:
                                try: var = var.decode('utf-8', 'ignore').encode(encoding, 'ignore')
                                except: pass
                                if var and not (node.has_key(tag) and find(node[tag][0], var) != -1):
                                    ok = 0
                            if not ok: continue

                            if awVar:
                                try: awVar = awVar.decode('utf-8', 'ignore').encode(encoding, 'ignore')
                                except: pass
                                c.game(gameNumber)
                                s = c.outputVar(c.currentN)
                                if find(s, awVar) == -1: continue
                                
                self.gamelist.appendC(dbIndex, g)

        self.gamelist.update()
  
        self.progBar.redraw(1)
        self.progBar.write('%1.1f seconds' % (time.time() - currentTime))

        for db in self.gamelist.DBlist:
            if not db['disabled']: del db['helplist']
        self.configButtons(NORMAL)


    def sigSearch(self):
        """ Search a game by its Dyer signature (sgf coord. of moves 20, 40, 60,
        31, 51, 71)."""

        if not self.gamelist.noOfGames(): return

        window = Toplevel(takefocus=0)
        window.title('Signature Search')

        m20 = StringVar()
        m40 = StringVar()
        m60 = StringVar()
        m31 = StringVar()
        m51 = StringVar()
        m71 = StringVar()
        
        l1 = Label(window, text='Move 20')
        e1 = Entry(window, width=4, textvariable=m20)
        l1.grid(row=1, column=0)
        e1.grid(row=1, column=1)
        l2 = Label(window, text='Move 40')
        e2 = Entry(window, width=4, textvariable=m40)
        l2.grid(row=1, column=2)
        e2.grid(row=1, column=3)
        l3 = Label(window, text='Move 60')
        e3 = Entry(window, width=4, textvariable=m60)
        l3.grid(row=1, column=4)
        e3.grid(row=1, column=5)
        l4 = Label(window, text='Move 31')
        e4 = Entry(window, width=4, textvariable=m31)
        l4.grid(row=2, column=0)
        e4.grid(row=2, column=1)
        l5 = Label(window, text='Move 51')
        e5 = Entry(window, width=4, textvariable=m51)
        l5.grid(row=2, column=2)
        e5.grid(row=2, column=3)
        l6 = Label(window, text='Move 71')
        e6 = Entry(window, width=4, textvariable=m71)
        l6.grid(row=2, column=4)
        e6.grid(row=2, column=5)

        e1.focus()
        
        bs = Button(window, text='Search',
                    command = lambda self=self, window=window,
                    m20=m20, m40=m40, m60=m60, m31=m31, m51=m51, m71=m71:
                    self.doSigSearch(window, m20, m40, m60, m31, m51, m71))
        bq = Button(window, text='Cancel', command = window.destroy)

        window.protocol('WM_DELETE_WINDOW', window.destroy)

        bo = v.Board(window, 19, (5, 18), 0, None, 0, None, self.boardImg, None, None)
        bo.state('normal', lambda pos, self=self, window=window, 
                 e1=e1, e2=e2, e3=e3, e4=e4, e5=e5, e6=e6,
                 m20=m20, m40=m40, m60=m60, m31=m31, \
                 m51=m51, m71=m71: self.sigSearchGetCoord(pos, window, e1, e2, e3, e4, e5, e6,
                                                          m20, m40, m60, m31, m51, m71))
        bo.shadedStoneVar.set(1)
        bo.grid(row=0, column=0, columnspan=6)
        bq.grid(row=3, column=4, columnspan = 2)
        bs.grid(row=3, column=2, columnspan = 2)
        
        window.update_idletasks()  
        # window.focus()
        window.grab_set()
        window.wait_window()


    def sigSearchGetCoord(self, pos, window, e1, e2, e3, e4, e5, e6, m20, m40, m60, m31, m51, m71):
        """ This writes the coordinates of a clicked-at point on the board to
        the currently focused entry. """

        f = window.focus_get()
        m = None
        
        if f is e1:
            e2.focus()
            m = m20
        if f is e2:
            e3.focus()
            m = m40
        if f is e3:
            e4.focus()
            m = m60
        if f is e4:
            e5.focus()
            m = m31
        if f is e5:
            e6.focus()
            m = m51
        if f is e6:
            e1.focus()
            m = m71

        p = chr(pos[0]-1+ord('a')) + chr(pos[1]-1+ord('a'))
        if m: m.set(p)


    def doSigSearch(self, window, m20, m40, m60, m31, m51, m71):

        window.destroy()
        self.configButtons(DISABLED)
        self.gamelist.clearGameInfo()

        self.statistics = {}
        
        for m in [m20, m40, m60, m31, m51, m71]:
            if len(m.get()) != 2: m.set('..')
        
 
        l = []
        oa = ord('a')
        
        for m in [m20, m40, m60, m31, m51, m71]: l.append((ord(m.get()[0])-oa, ord(m.get()[1])-oa))

        # print join([m20.get(), m40.get(), m60.get(), m31.get(), m51.get(), m71.get()], '')

        sig_list = []
        for f in Pattern.flips:
            k1 = [ f(x[0], x[1]) for x in l] 
            k2 = []
            for i in range(6):
                if chr(l[i][0]+oa) == '.': k2.append('..')
                else:
                    k2.append(chr(k1[i][0]+oa))
                    k2.append(chr(k1[i][1]+oa))

            s = join(k2, '')
            if not s in sig_list: sig_list.append(s)

        sp_list = [re.compile(s) for s in sig_list]

        ctr = 0
        noGames = self.gamelist.noOfGames()
        self.progBar.clear()
        currentTime = time.time()

        self.board.delLabels()
        if self.cursor:
            try: self.displayLabels(self.cursor.currentNode())
            except: pass
            
        for db in self.gamelist.DBlist:
            if not db['disabled']: db['helplist'] = db['current']
        self.gamelist.clear() 
        
        for dbIndex in range(len(self.gamelist.DBlist)):
            db = self.gamelist.DBlist[dbIndex]
            if db['disabled']: continue
            for g in db['helplist']:
                ctr = ctr + 1
                if ctr % 100 == 0:
                    # if ctr % 1000 == 500:
                    #     self.gamelist.update()
                    self.progBar.redraw(ctr*1.0/noGames)
                    self.searchWindow.update()
                for sp in sp_list:
                    if sp.match(db['data'][g][5]):
                        self.gamelist.appendC(dbIndex, g)
                        break
                    
        self.gamelist.update()

        self.progBar.redraw(1)
        self.progBar.write('%1.1f seconds' % (time.time() - currentTime))

        for db in self.gamelist.DBlist:
            if not db['disabled']: del db['helplist']
        self.configButtons(NORMAL)


    def back(self, prev=None, selected=None):
        """ Go back to previous search (restore board, game list etc.)
        If an SGF file is currently loaded (and was loaded at the time of
        the previous search too), the position of its cursor is
        restored too."""

        self.leaveNode()

        if not prev:
            try:
                prev, selected = self.prevSearches.pop()
            except:
                self.reset()
                return
                
        try:
            d, g, r, c, nm, ns, bw, ww, mv, fc, fa, ml, nextM, cu = prev
        except:
            self.reset()
            return

        self.comments.delete('1.0', END)

        self.gamelist.clear()
        self.gamelist.clearGameInfo()
        for i in range(len(g)):
            self.gamelist.DBlist[i]['current'] = g[i]
            self.gamelist.DBlist[i]['results'] = [''] * len(g[i])
            
        self.noMatches = 0
        self.noSwitched = 0
        self.Bwins = 0
        self.Wwins = 0
        
        if d:
            if cu:
                found = 0
                for i in range(len(self.filelist)):          
                    if self.filelist[i][3] == cu[0]:
                        if i != self.currentFileNum:
                            self.changeCurrentFile(None, i)
                        found = 1
                if found:
                    if cu[1] != self.cursor.currentGame:
                        self.changeCurrentGame(None, cu[1])
                    else:
                        try: self.cursor.game(self.cursor.currentGame)
                        except:
                            showwarning('Error', 'SGF Error')
                            return
                    self.board.newPosition()
                    self.moveno.set('0')
                    self.capB, self.capW = 0, 0 
                    self.displayNode(self.cursor.getRootNode(self.cursor.currentGame))
                    for i in cu[2]:
                        self.next(i,0)
                    self.cursor.seeCurrent()
                    
                    self.board.restore(d)
            
                    self.sel = self.board.selection      # used in self.showCont()                    
                    self.capVar.set('Cap - B: ' + str(self.capB) + ', W: ' + str(self.capW))
                else:
                    showwarning('Error', 'SGF File not found')

            self.modeVar.set(mv)
            self.fixedColorVar.set(fc)
            self.fixedAnchorVar.set(fa)
            self.moveLimit.set(ml)
            self.nextMoveVar.set(nextM)
            
            for i in range(len(r)):
                self.gamelist.DBlist[i]['results'] = split(r[i], '|')

            self.continuations = c
        
            self.noMatches = nm
            self.noSwitched = ns
            self.Bwins = bw
            self.Wwins = ww
        
            if self.showContinuation.get(): self.showCont()
            self.board.changed.set(0)

            # restore self.currentSearchPattern

            pd = {}

            for i in range(self.sel[0][1], self.sel[1][1]+1):
                for j in range(self.sel[0][0], self.sel[1][0]+1):
                    if not self.board.status.has_key((j,i)):  pd[(i-1,j-1)] = '.'
                    elif self.board.status[(j,i)] == 'black': pd[(i-1,j-1)] = 'X'
                    else:                                     pd[(i-1,j-1)] = 'O'
                    if self.board.wildcards.has_key((j,i)):   pd[(i-1,j-1)] = '*'
            
            if self.sel[0] == (1,1) or \
               (self.sel[0][0], self.sel[1][1]) == (1,19) or \
               (self.sel[1][0], self.sel[0][1]) == (19,1) or \
               self.sel[1] == (19,19) or \
               self.fixedAnchorVar.get():                       
                anchors = [(self.sel[0][1]-1, self.sel[0][0]-1),
                           (self.sel[0][1]-1, self.sel[0][0]-1)]
            elif 1 in self.sel[0] or 19 in self.sel[1]:            # translation along the edge only
                anchors = []
                if self.sel[0][0] == 1:
                    anchors = [(1,0), (17 - self.sel[1][1] + self.sel[0][1] ,0)]
                elif self.sel[0][1] == 1:
                    anchors = [(0,1), (0,17 - self.sel[1][0] + self.sel[0][0])]
                elif self.sel[1][0] == 19:
                    anchors = [(1,self.sel[0][0]-1),
                               (17-self.sel[1][1]+self.sel[0][1],self.sel[0][0]-1)]
                elif self.sel[1][1] == 19:
                    anchors = [(self.sel[0][1]-1,1),
                               (self.sel[0][1]-1,17-self.sel[1][0]+self.sel[0][0])]
            else:                                                        # full translation
                anchors = [(1, 1), (17-self.sel[1][1]+self.sel[0][1],
                                    17-self.sel[1][0]+self.sel[0][0])]

            self.currentSearchPattern = Pattern(anchors, pd)

        self.statistics = {}
        self.displayStatistics()

        self.gamelist.appendClist = []
        for i in range(len(g)):
            self.gamelist.appendClist += [(i, self.gamelist.DBlist[i]['current'][j], j) \
                                          for j in range(len(self.gamelist.DBlist[i]['current']))]
        self.gamelist.update()        
        self.prevSearches.select(selected)


    def completeReset(self):
        self.reset()
        self.board.newPosition()
        self.changeCurrentFile(None, 0)
        for i in range(len(self.filelist)):
            self.delFile()
        
        self.gotoVar.set('')

        self.clearGI()
        self.history_GIsearch = []
        self.history_GIS_index = -1

        self.moveLimit.set(250)
        if not self.options.smartFixedColor.get(): self.fixedColorVar.set(0)
        self.fixedAnchorVar.set(0)
        self.nextMoveVar.set(0)
        
        self.showContinuation.set(0)
        self.oneClick.set(0)
        

    def reset(self):
        """ Reset the game list. """

        self.gamelist.reset()

        self.progBar.clear()
        self.statisticsCanv.delete('stat')
        self.board.delLabels()
        if self.cursor:
            try:
                self.leaveNode()
                self.displayLabels(self.cursor.currentNode())
            except: pass
        self.continuations = {}
        self.statistics = {}


    def toggleSorting(self):
        self.gamelist.sortlist(self.options.sortPerDatabase.get(),
                               self.options.sortCriterion.get())


    def loadDBs(self):

        # first we parse the 'references' file and store the results in a dictionary

        try:
            ref_file = open(os.path.join(self.basepath, 'references'))
            r = ref_file.readlines()
            ref_file.close()
        except:
            r = ''
            
        self.gamelist.references = {}
        abbr = {}
        found_abbr = 0
        found_data = 0
        for l in r:

            if l and l[0] == '#': continue
        
            if not found_abbr:
                if l[:7] != '---Abbr': continue
                else:
                    found_abbr = 1
                    continue
                
            if found_abbr and not found_data:
                if strip(l):
                    s = split(l)
                    abbr[s[0]] = join(s[1:], ' ')

            if found_abbr and not found_data:
                if l[:7] != '---Data': continue
                else:
                    found_data = 1
                    continue
                
            if found_data and strip(l):
                s = split(l)
                if abbr.has_key(s[1]):
                    if self.gamelist.references.has_key(s[0]):
                        self.gamelist.references[s[0]] += ', ' + abbr[s[1]] + ' ' + join(s[2:], ' ')
                    else:
                        self.gamelist.references[s[0]] = abbr[s[1]] + ' ' + join(s[2:], ' ')

        # now load the databases

        DBlistIndex = 0
        for i in range(len(self.gamelist.DBlist)):
            db = self.gamelist.DBlist[DBlistIndex]

            try:
                file = open(os.path.join(db['name'][0], 'namelist' + db['name'][1] + '.db'))
                namelist = file.readlines()
                file.close()
            except: 
                showwarning('IOError', 'Could not open database ' +\
                            os.path.join(db['name'][0], 'namelist' + db['name'][1] + '.db'))
                del self.gamelist.DBlist[DBlistIndex]
                continue

            # first check for right format: (... should really check every line in the for loop ...)

            try:
                filename, PB, PW, res, sig, date = split(namelist[0][:-1], '|%')
            except:
                showwarning('Error', 'The database file ' + os.path.join(db['name'][0], 'namelist' + db['name'][1] + '.db') +\
                            ' does not have the right format. '
                            + 'Please reprocess it.')
                del self.gamelist.DBlist[DBlistIndex]
                continue

            try:
                for namelistIndex in range(len(namelist)):
                    g = namelist[namelistIndex][:-1]
                    filename, PB, PW, res, sig, date = split(g, '|%')

                    if self.gamelist.references.has_key(sig): ref = '*'
                    else: ref = ''
                        
                    self.gamelist.append(DBlistIndex, (filename, namelistIndex,
                                                       PB, PW, res, sig, date, ref))
            except ImportError:
                showwarning('Error', 'The database file ' + os.path.join(db['name'][0], 'namelist' + db['name'][1] + '.db') +\
                            ' does not have the right format. '
                            + 'Please reprocess it.')
                del self.gamelist.DBlist[DBlistIndex]
                continue

            DBlistIndex = DBlistIndex + 1

        self.gamelist.addAppendedGames()
        

    def showCont(self):
        """ Toggle 'show continuations'. """
        
        if self.showContinuation.get():
            self.board.delLabels()
            for x,y in self.continuations.keys():
                if not self.continuations[(x,y)]['B']:
                    color = 'white'
                elif not self.continuations[(x,y)]['W']:
                    color = 'black'
                else:
                    color = self.options.labelColor.get()
                self.board.placeLabel((y+self.sel[0][0], x+self.sel[0][1]),
                                      '+LB', self.continuations[(x,y)]['N'], color)

        else:
            self.board.delLabels()
            if self.cursor:
                try: self.displayLabels(self.cursor.currentNode())
                except: pass

    def doubleClick(self, event):
        if not self.oneClick.get(): self.search()


    # ---- putting stones on the board, and navigation in SGF file ----------------------

    def nextMove(self, pos):
        self.board.delLabels()
        self.board.delMarks()
        self.continuations = {}
        
        if self.board.wildcards.has_key(pos):
            self.board.delete(self.board.wildcards[pos])
            del self.board.wildcards[pos]
            
        v.Viewer.nextMove(self, pos)

        if self.oneClick.get() and (self.board.selection[1] == (0,0) or \
                                    (self.board.selection[0][0] <= pos[0] <= self.board.selection[1][0] \
                                     and self.board.selection[0][1] <= pos[1] <= self.board.selection[1][1])):
            self.search()

                   
    def next(self, n=0, update=1):
        if not self.cursor or self.cursor.atEnd: return
        self.board.delLabels()
        self.continuations = {}
        self.board.delWildcards()
        
        if v.Viewer.next(self, n, update): return 1
        return 0

    
    def prev(self, markCurrent=1):
        if not self.cursor or self.cursor.atStart: return
        self.board.delLabels()
        self.continuations = {}
        
        self.board.delWildcards()
        
        v.Viewer.prev(self, markCurrent)
 
                
    def start(self, update=1):
        self.board.newPosition()
        v.Viewer.start(self, update)

        self.statisticsCanv.delete('stat')
        self.continuations = {}
        
        self.board.delWildcards()

        
    def end(self):
        v.Viewer.end(self)
            

    def delStone(self, event):
        x, y = self.board.getBoardCoord((event.x, event.y), 1)
        if not x*y: return

        if self.board.wildcards.has_key((x,y)):
            self.board.delete(self.board.wildcards[(x,y)])
            del self.board.wildcards[(x,y)]
        else: Viewer.delStone(self, event)


    def delVar(self):
        self.prevSearches.deleteNode(self.cursor, self.cursor.currentGame, self.cursor.currentN.pathToNode())
        v.Viewer.delVar(self)


    def delGame(self):
        self.prevSearches.deleteGame(self.cursor, self.cursor.currentGame)
        v.Viewer.delGame(self)


    def delFile(self):
        self.prevSearches.deleteFile(self.cursor)
        return v.Viewer.delFile(self)


    # -------------------------------------------------

    def exportCurrentPos(self):

        numberOfMoves = IntVar()
        exportMode = IntVar()

        dialog = Toplevel()
        dialog.title('Export position')
        dialog.protocol('WM_DELETE_WINDOW', lambda: None)

        f1 = Frame(dialog)
        f2 = Frame(dialog)
        f3 = Frame(dialog)

        for f in [f1, f2, f3]: f.pack(side=TOP, fill=BOTH, expand=YES, pady=5)
        
        Label(f1, text='Number of moves to be shown (0-9):').pack(side=TOP)
        Entry(f1, textvariable=numberOfMoves).pack(side=TOP)

        Label(f2, text='Export mode:').pack(side=LEFT)
        Radiobutton(f2, text='ASCII', variable=exportMode, value=0, highlightthickness=0).pack(side=LEFT)
        Radiobutton(f2, text='Wiki', variable=exportMode, value=1, highlightthickness=0).pack(side=LEFT)

        Button(f3, text='OK', command=dialog.destroy).pack(anchor=E)

        dialog.update_idletasks()  
        dialog.focus()
        dialog.grab_set()
        dialog.wait_window()
        
        n = numberOfMoves.get()

        l = []
        t = []

        for i in range(19): l.append(['. ']*19)

        for i in range(1,20):
            for j in range(1,20):
                if not self.board.status.has_key((j,i)): pass
                elif self.board.status[(j,i)] == 'black': l[i-1][j-1] = 'X '
                else:                                    l[i-1][j-1] = 'O '
                if self.board.wildcards.has_key((j,i)): l[i-1][j-1] = '* '

        # mark hoshis with ,'s

        for i in range(3):
            for j in range(3):
                ii = 3 + 6*i 
                jj = 3 + 6*j 
                if l[ii][jj] == '. ': l[ii][jj] = ', '

        remarks = []
        nextMove = 'B'

        node = self.cursor.currentN
        
        for i in range(1, min(n, 10)+1):
            if i == 10: ii = '0 '
            else: ii = `i` + ' '

            node = node.next
            if not node: break
            try:
                if node.getData().has_key('B'): color = 'B'
                elif node.getData().has_key('W'): color = 'W'
                else: continue

                if i == 1: nextMove = color

                pos = self.convCoord(node.getData()[color][0])
                pos = (pos[1]-1, pos[0]-1)
                if not pos: continue
            
                if l[pos[0]][pos[1]] in ['. ', ', ']: l[pos[0]][pos[1]] = ii
                elif l[pos[0]][pos[1]] in [`i`+' ' for i in range(10)]:
                    remarks.append(ii + 'at ' + l[pos[0]][pos[1]] + '\n')
                elif l[pos[0]][pos[1]] in ['X ', 'O ']:
                    remarks.append(ii + 'at ' + 'ABCDEFGHJKLMNOPQRST'[pos[1]] + `pos[0]+1` + '\n')

            except: break
            
        if not exportMode.get():
            for i in range(19):
                l[i].insert(0, '%2d  ' % (19-i))

            l.insert(0, '')
            l.insert(0, '    A B C D E F G H J K L M N O P Q R S T')

            if n:
                if nextMove == 'B': remarks.append('Black = 1\n')
                elif nextMove == 'W': remarks.append('White = 1\n')
        else:
            for i in range(19):
                l[i].insert(0, '$$ | ')
                l[i].append('|')
        
            l.insert(0, '$$  ---------------------------------------')
            l.insert(0, '$$' + nextMove)
            l.append('$$  ---------------------------------------')

        for line in l:
            t.append(join(line, '') + '\n')

        t.append('\n')
        t.extend(remarks)

        ESR_TextEditor(self, exportMode.get(), join(t, ''), self.sgfpath,
                       (self.options.exportFont0, self.options.exportFont1, self.options.exportFont2))


    def exportSGF(self):
        self.leaveNode()
        try:
            n = self.cursor.root.next
            for i in range(self.cursor.currentGame): n = n.down
            
            t = '('+self.cursor.outputVar(n)+')'
            t = replace(t, '\r', '')
        
            TextEditor(t, self.sgfpath,
                       (self.options.exportFont0, self.options.exportFont1, self.options.exportFont2))
        except: showwarning('Error', 'SGF error')


    def exportText(self):
        """Export some information on the previous search in a small text editor,
        where it can be edited and saved to a file. """

        exportMode = IntVar()

        dialog = Toplevel()
        dialog.title('Export position')
        dialog.protocol('WM_DELETE_WINDOW', lambda: None)

        f1 = Frame(dialog)
        f2 = Frame(dialog)
        f3 = Frame(dialog)

        for f in [f1, f2, f3]: f.pack(side=TOP, fill=BOTH, expand=YES, pady=5)
        
        Label(f1, text='Export mode:').pack(side=LEFT)
        Radiobutton(f1, text='ASCII', variable=exportMode, value=0, highlightthickness=0).pack(side=LEFT)
        Radiobutton(f1, text='Wiki', variable=exportMode, value=1, highlightthickness=0).pack(side=LEFT)

        showAllCont = IntVar()
        Checkbutton(f2, text='Show all continuations', variable=showAllCont, highlightthickness=0).pack(side=LEFT)

        Button(f3, text='OK', command=dialog.destroy).pack(anchor=E)

        dialog.update_idletasks()  
        dialog.focus()
        dialog.grab_set()
        dialog.wait_window()

        t = []
        if self.currentSearchPattern: 

            d = {}
            for i in range(self.currentSearchPattern.sizeX):
                for j in range(self.currentSearchPattern.sizeY):
                    d[(i,j)] = self.currentSearchPattern.data[i][j] + ' '

            # borders

            if self.currentSearchPattern.anchors[0][0] == 0:
                for j in range(self.currentSearchPattern.sizeY):
                    d[(-1,j)] = '--'
            if self.currentSearchPattern.anchors[0][1] == 0:
                for i in range(self.currentSearchPattern.sizeX):
                    d[(i,-1)] = '| '
            if self.currentSearchPattern.anchors[1][0] + self.currentSearchPattern.sizeX == 19:
                for j in range(self.currentSearchPattern.sizeY):
                    d[(self.currentSearchPattern.sizeX,j)] = '--'
            if self.currentSearchPattern.anchors[1][1] + self.currentSearchPattern.sizeY == 19:
                for i in range(self.currentSearchPattern.sizeX):
                    d[(i, self.currentSearchPattern.sizeY)] = '| '

            if self.currentSearchPattern.anchors[0] == (0,0):
                d[(-1,-1)] = ' -'
            if (self.currentSearchPattern.anchors[1][0] + self.currentSearchPattern.sizeX,
                self.currentSearchPattern.anchors[1][1] + self.currentSearchPattern.sizeY) == (19,19):
                d[(self.currentSearchPattern.sizeX, self.currentSearchPattern.sizeY)] = ' '
            if (self.currentSearchPattern.anchors[0][0],
                self.currentSearchPattern.anchors[1][1] + self.currentSearchPattern.sizeY) == (0,19):
                d[(-1, self.currentSearchPattern.sizeY)] = ' '
            if (self.currentSearchPattern.anchors[1][0] + self.currentSearchPattern.sizeX,
                self.currentSearchPattern.anchors[0][1]) == (19,0):
                d[(self.currentSearchPattern.sizeX, -1)] = ' -'

            if self.currentSearchPattern.anchors[0][0] == 0 and not \
               (self.currentSearchPattern.anchors[0][0],
                self.currentSearchPattern.anchors[1][1] + self.currentSearchPattern.sizeY) == (0,19):
                d[(-1, self.currentSearchPattern.sizeY-1)] = '- '

            if self.currentSearchPattern.anchors[1][0] + self.currentSearchPattern.sizeX == 19 \
               and not (self.currentSearchPattern.anchors[1][0] + self.currentSearchPattern.sizeX,
                        self.currentSearchPattern.anchors[1][1] + self.currentSearchPattern.sizeY) == (19,19):
                d[(self.currentSearchPattern.sizeX, self.currentSearchPattern.sizeY-1)] = '- '

            # mark hoshis with ,'s

            if self.currentSearchPattern.anchors[0] == self.currentSearchPattern.anchors[1]:
                for i in range(3):
                    for j in range(3):

                        ii = 3 + 6*i - self.currentSearchPattern.anchors[0][0]
                        jj = 3 + 6*j - self.currentSearchPattern.anchors[0][1]

                        if d.has_key((ii,jj)) and d[(ii, jj)] == '. ' :
                            d[(ii, jj)] = ', '

            # ----------------------
            
            l1 = []
            for i in range(-1, self.currentSearchPattern.sizeX+1):
                l1.append( ' ' + join([d[(i,j)] for j in range(-1, self.currentSearchPattern.sizeY+1) \
                                       if d.has_key((i,j))], ''))

            if l1[0] == ' ': l1 = l1[1:]
            if l1[-1] == ' ': l1 = l1[:-1]

            if showAllCont.get():
                N = 400
            else: N = 10

            for i in range(N):
                if not self.statistics.has_key(i): break

                if d[self.statistics[i][0]] in ['. ', ', ']:
                    d[self.statistics[i][0]] = self.statistics[i][1]['N'] + ' '

            l2 = []
            for i in range(-1, self.currentSearchPattern.sizeX+1):
                l2.append(' ' + join([d[(i,j)] for j in range(-1, self.currentSearchPattern.sizeY+1) \
                                      if d.has_key((i,j))], '') )

            if l2[0] == ' ': l2 = l2[1:]
            if l2[-1] == ' ': l2 = l2[:-1]
        
            if not exportMode.get():
                s1 = join(l1, '\n')
                s2 = join(l2, '\n')

            else:
                s1 = '$$B Search Pattern\n$$' + join(l1, '\n$$') + '\n' 
                s2 = '$$B Continuations\n$$' + join(l2, '\n$$') + '\n'
        
            if exportMode.get(): t.append('!')
            t.append('Search results\n\n')
            if not exportMode.get(): t.append('Pattern:\n')

            t.append(s1)

            if not exportMode.get(): t.append('\n\nContinuations:\n')
            else: t.append('\n\n')
        
            t.append(s2)
            t.append('\n')

            if self.statistics:

                if exportMode.get(): t.append('%%%%\n!')
                else: t.append('\n')

                t.append('Statistics:\n')

                t.append(self.statistics['total'])
                if exportMode.get(): t.append(' %%%\n')
                t.append('\n')

                for i in range(N):
                    if not self.statistics.has_key(i): break
                    else: sd = self.statistics[i][1]
                    if sd['B']:
                        t.append('B'+sd['N'] + ':    ' + `sd['B']` + ' (' + `sd['B']-sd['tB']` + '), ')
                        Bp = '%1.1f' % (sd['wB']*100.0/sd['B'])
                        Wp = '%1.1f' % (sd['lB']*100.0/sd['B'])
                        t.append('B'+Bp+'% - W'+Wp+'%')
                        if exportMode.get(): t.append(' %%%')
                        t.append('\n')
                    if sd['W']:
                        t.append('W'+sd['N'] + ':    ' + `sd['W']` + ' (' + `sd['W']-sd['tW']` + '), ')
                        Bp = '%1.1f' % (sd['wW']*100.0/sd['W'])
                        Wp = '%1.1f' % (sd['lW']*100.0/sd['W'])
                        t.append('B'+Bp+'% - W'+Wp+'%')
                        if exportMode.get(): t.append(' %%%')
                        t.append('\n')

                t.append('\n')
                if exportMode.get(): t.append('!')
                t.append('Hits per database\n')

                for db in self.gamelist.DBlist:
                    if db['disabled']: continue
                    t.append(db['sgfpath'] + ': ' + `len(db['current'])` + ' games (of ' + `len(db['data'])` + ')')
                    if exportMode.get(): t.append(' %%%')
                    t.append('\n')
        
        ESR_TextEditor(self, exportMode.get(), join(t, ''), self.sgfpath,
                   (self.options.exportFont0, self.options.exportFont1, self.options.exportFont2))

    # ----------------------------------------------------------------------------------

    def getFilename(self, no, label):
        dbindex, index = self.gamelist.getIndex(int(no[0]))
        if dbindex == -1: return
        
        moveno = 0
        if self.gamelist.DBlist[dbindex]['results'][index]:
            s = self.gamelist.DBlist[dbindex]['results'][index]
            i=0
            while i<len(s) and s[i] in digits: i += 1
            
            if i:
                moveno = int(s[:i]) 


        f1 = strip(os.path.join(self.gamelist.DBlist[dbindex]['sgfpath'],
                                self.gamelist.DBlist[dbindex]['data']
                                [self.gamelist.DBlist[dbindex]['current'][index]][0]))
        if find(f1, '[') != -1:
            f1, f2 = split(f1, '[')
            gameNumber = int(strip(f2)[:-1])

        else: gameNumber = 0
                
        filename = getFilename(f1)        

        return filename, gameNumber, moveno

    def getEncoding(self, no, label):
        dbindex, index = self.gamelist.getIndex(int(no[0]))
        if dbindex == -1: return
        else: return self.gamelist.DBlist[dbindex]['encoding']



    def openViewer_external(self, filename, gameNumber, moveno, encoding = ''):
        if self.options.altViewerVar1.get():
                
            if sys.platform[:3] == 'win': filenameQU = '"' + filename + '"'
            else: filenameQU = filename
            
            s1 = self.options.altViewerVar1.get()
            if os.name == 'posix':
                s1 = s1.replace('~', os.getenv('HOME'))
                    
            s2 = replace(self.options.altViewerVar2.get(), '%f', filenameQU)
            s2 = replace(s2, '%F', filename)
            s2 = replace(s2, '%n', `moveno`)
            s2 = replace(s2, '%g', `gameNumber`)

            try:
                    
                if sys.platform[:3] == 'win':
                    os.spawnv(os.P_DETACH, s1, ('"'+s1+'"',)+tuple(split(s2)))
                        # it is necessary to quote the
                        # path if it contains blanks

                elif os.path.isfile(s1):
                    pid = os.fork()
                    if pid == 0:
                        os.execv(s1, (s1,)+tuple(split(s2)))
                        showwarning('Error', 'Error starting SGF viewer')
                else: showwarning('Error', s1 + ' not found.')
            except OSError:
                showwarning('Error', 'Error starting SGF viewer')
                
        else:
            window = Toplevel()
            window.withdraw()
            
            viewer = v.Viewer(window)
            viewer.frame.focus_force()

            window.protocol('WM_DELETE_WINDOW', viewer.quit)

            apply(viewer.openFile, os.path.split(filename) + (encoding,))
            viewer.changeCurrentFile(None, 1)
            viewer.delFile()

            if gameNumber:
                viewer.dataWindow.gamelist.list.select_clear(0)
                viewer.dataWindow.gamelist.list.select_set(gameNumber)
                viewer.dataWindow.gamelist.list.see(gameNumber)
                viewer.changeCurrentGame(None, gameNumber)

            if self.options.jumpToMatchVar.get():
                try:
                    if moveno:
                        while not (viewer.cursor.currentNode().has_key('B') or viewer.cursor.currentNode().has_key('W')):
                            viewer.next(0,0)
                    i=1
                    while i < moveno:
                        if viewer.cursor.atEnd:
                            break
                    
                        if viewer.cursor.currentNode().has_key('B') or viewer.cursor.currentNode().has_key('W'):
                            i += 1
                        viewer.next(0,0)
                    viewer.cursor.seeCurrent()
                except:
                    showwarning('Error', 'SGF error')


    def openViewer_internal(self, filename, gameNumber, moveno, encoding):

        apply(self.openFile, os.path.split(filename) + (encoding,))

        if gameNumber:
            self.dataWindow.gamelist.list.select_clear(0)
            self.dataWindow.gamelist.list.select_set(gameNumber)
            self.dataWindow.gamelist.list.see(gameNumber)
            self.changeCurrentGame(None, gameNumber)

        if self.options.jumpToMatchVar.get():
            try:
                if moveno:
                    while not (self.cursor.currentNode().has_key('B') or self.cursor.currentNode().has_key('W')):
                        self.next(0,0)
                i=1
                while i < moveno:
                    if self.cursor.atEnd:
                        break
                    
                    if self.cursor.currentNode().has_key('B') or self.cursor.currentNode().has_key('W'):
                        i += 1
                    self.next(0,0)    
            except:
                showwarning('Error', 'SGF error')
            self.cursor.seeCurrent()

            
    def altOpenViewer(self, no, label):
        """ Open game from game list in SGF viewer - "alternative mode", i.e. if the default is to
        open an external viewer, this method will load the game into the internal list, and vice versa."""
        
        filename, gameNumber, moveno = self.getFilename(no, label)
        encoding = self.getEncoding(no, label)
        if not self.options.externalViewer.get():
            self.openViewer_external(filename, gameNumber, moveno, encoding)
        else:
            self.openViewer_internal(filename, gameNumber, moveno, encoding)


    def openViewer(self, no, label):
        """ Open game from game list in SGF viewer. """

        filename, gameNumber, moveno = self.getFilename(no, label)
        encoding = self.getEncoding(no, label)
        if self.options.externalViewer.get():
            self.openViewer_external(filename, gameNumber, moveno, encoding)           
        else: # don't open external viewer
            self.openViewer_internal(filename, gameNumber, moveno, encoding)

                    
    def altViewer(self):
        """ Ask for alternative SGF viewer. """
        
        window = Toplevel()
        window.title('Alternative SGF viewer')

        f = Frame(window)
        f.pack()
        
        l1 = Label(f, text='Enter the command to launch the SGF viewer')
        e1 = Entry(f, width=40, textvariable=self.options.altViewerVar1)
        l2 = Label(f, text='Enter the command line options, with %f for the filename')
        e2 = Entry(f, width=40, textvariable=self.options.altViewerVar2)

        b = Button(f, text='OK', command = window.destroy) 

        window.protocol('WM_DELETE_WINDOW', lambda: None)
        
        l1.pack(side=TOP, anchor=W)
        e1.pack(side=TOP)
        l2.pack(side=TOP, anchor=W)
        e2.pack(side=TOP)
        b.pack(side=RIGHT)

        window.update_idletasks()  
        window.focus()
        window.grab_set()
        window.wait_window()

    # ---- administration of DBlist ----------------------------------------------------

    def addDB(self, window):
        self.editDB_OK.config(state=DISABLED)
        self.saveProcMess.config(state=DISABLED)
        self.stopAddDBButton.config(state=NORMAL)

        self.stopAddDBVar.set(0)

        if self.options.oldTkVar.get():
            dbp = tkFileDialog.askopenfilename(initialdir = self.datapath)

            if dbp and os.path.isfile(dbp):
                dbp = os.path.split(dbp)[0]
                dbp = os.path.normpath(dbp)
            else:
                self.editDB_OK.config(state=NORMAL)
                self.saveProcMess.config(state=NORMAL)
                self.stopAddDBButton.config(state=DISABLED)
                return
        else:
            try:
                dbp = str(askdirectory(initialdir = self.datapath))
            except TclError:
                self.options.oldTkVar.set(1)
                dbp = tkFileDialog.askopenfilename(initialdir = self.datapath)

                if dbp and os.path.isfile(dbp): dbp = os.path.split(dbp)[0]
                else:
                    self.editDB_OK.config(state=NORMAL)
                    self.saveProcMess.config(state=NORMAL)
                    self.stopAddDBButton.config(state=DISABLED)
                    return
                
            if not dbp:
                self.editDB_OK.config(state=NORMAL)
                self.saveProcMess.config(state=NORMAL)
                self.stopAddDBButton.config(state=DISABLED)
                return
            else:
                dbp = os.path.normpath(dbp)
            
        self.datapath = os.path.split(dbp)[0]

        if self.options.storeDatabasesSeparately.get() and self.options.whereToStoreDatabases.get():
            datap = (self.options.whereToStoreDatabases.get(), '')
            if os.path.exists(datap[0]) and not os.path.isdir(datap[0]):
                showwarning('Error', datap[0] + ' is not a directory.')
                self.editDB_OK.config(state=NORMAL)
                self.saveProcMess.config(state=NORMAL)
                self.stopAddDBButton.config(state=DISABLED)
                return
            elif not os.path.exists(datap[0]):
                askyesno_result =  askokcancel('Error', 'Directory ' + datap[0] + ' does not exist. Create it?')
                if askyesno_result:
                    try:
                        os.makedirs(datap[0])
                    except:
                        showwarning('Error', datap[0] + ' could not be created.')
                        self.editDB_OK.config(state=NORMAL)
                        self.saveProcMess.config(state=NORMAL)
                        self.stopAddDBButton.config(state=DISABLED)
                        return
                else:
                    self.editDB_OK.config(state=NORMAL)
                    self.saveProcMess.config(state=NORMAL)
                    self.stopAddDBButton.config(state=DISABLED)
                    return          
        else:
            datap = ('', '#') # this means: same as dbpath

        if not self.options.recProcess.get():
            self.addOneDB((self.processMessages, datap), dbp, None)
        else:
            os.path.walk(dbp, self.addOneDB, (self.processMessages, datap))

        self.editDB_OK.config(state=NORMAL)
        self.saveProcMess.config(state=NORMAL)
        self.stopAddDBButton.config(state=DISABLED)

        
    def addOneDB(self, arguments, dbpath, dummy):        # dummys needed for os.path.walk

        messages, datap = arguments 
        if datap[1] == '#': datap = (dbpath, '')
        else:
            try:
                flist = glob.glob(os.path.normpath(os.path.join(datap[0], 'namelist*.db')))
                i = 1
                while 1:
                    if not os.path.normpath(os.path.join(datap[0], 'namelist' + `i` + '.db')) in flist \
                       and not os.path.isfile(os.path.join(datap[0], 'namelist' + `i` + '.db')):
                        break
                    i += 1
            except IOError:
                messages.insert(END, 'Error: ' + datap[0] + 'cannot be used as database path.')
                return

            datap = (datap[0], `i`)
            
        if self.stopAddDBVar.get(): return
        
        for db in self.gamelist.DBlist:
            if dbpath == db['sgfpath']:
                messages.insert(END, 'Error: The database ' + dbpath + ' is already in the list.\n')
                return

        try:
            file = open(os.path.join(datap[0], 'namelist' + datap[1] + '.db'))
            namelist = file.readlines()
            file.close()

            g = namelist[0][:-1]
            filename, PB, PW, res, sig, date = split(g, '|%')    # check that this is a namelist for
                                                                 # version 0.5
        except:
            try:
                self.process(dbpath, datap, messages)
            except:
                showwarning('Error', 'A fatal error occured when processing ' + dbpath + '. Please send a bug report.')
                return
            try:
                file = open(os.path.join(datap[0], 'namelist' + datap[1] + '.db'))
                namelist = file.readlines()
                file.close()
            except:
                return

        if not namelist: return
        
        self.gamelist.DBlist.append({'name': datap, 'sgfpath':dbpath, 'encoding': self.encodingVar.get(),
                                     'current':array('I'), 'results':[], 'numNewGames':0,
                                     'data':[], 'disabled': 0})
        db = self.gamelist.DBlist[-1]
        db_date = getDateOfFile(os.path.join(db['name'][0],'namelist'+db['name'][1]+'.db'))

        self.db_list.insert(END, dbpath + ' ('+db_date+')' )
        self.db_list.list.see(END)
        
        DBlistIndex = len(self.gamelist.DBlist)-1

        for namelistIndex in range(len(namelist)):
            g = namelist[namelistIndex][:-1]
            
            filename, PB, PW, res, sig, date = split(g, '|%')
            if self.gamelist.references.has_key(sig): ref = '*'
            else: ref = ''

            self.gamelist.append(DBlistIndex, (filename, namelistIndex,
                                               PB, PW, res, sig, date, ref))

        self.prevSearches.clear()
        self.currentSearchPattern = None
        self.gamelist.addAppendedGames()
        
        messages.insert(END, 'Added ' + dbpath + '.\n')

     
    def removeDB(self):
        while self.db_list.list.curselection():
            index = self.db_list.list.curselection()[0]
            self.db_list.delete(index)
            i = int(index)
            datap = self.gamelist.DBlist[i]['name']
            dbpath = self.gamelist.DBlist[i]['sgfpath']
            del self.gamelist.DBlist[i]

            try:
                for filename in ['namelist', 'lists', 'finalpos', 'posTable', 'hash', 'hashT']:
                    os.remove(os.path.join(datap[0], filename+datap[1]+'.db'))
            except:
                showwarning('I/O Error', 'Could not delete the database files.')
            self.processMessages.insert(END, 'Removed ' + dbpath + '\n')
            
        self.gamelist.reset()
        self.prevSearches.clear()
        self.currentSearchPattern = None
                                 

    def reprocessDB(self):
        for index in self.db_list.list.curselection():
            i = int(index)
        
            self.editDB_OK.config(state=DISABLED)
            self.prevSearches.clear()
            self.currentSearchPattern = None
        
            dbpath = self.gamelist.DBlist[i]['sgfpath']
            datap = self.gamelist.DBlist[i]['name']
            del self.gamelist.DBlist[i]

            try:
                for filename in ['namelist', 'lists', 'finalpos', 'posTable', 'hash', 'hashT']:
                    os.remove(os.path.join(datap[0], filename+datap[1]+'.db'))
            except:
                showwarning('I/O Error', 'Could not delete the database files.')

            try:
                self.process(dbpath, datap, self.processMessages)
            except:
                showwarning('Error', 'A fatal error occurred when processing ' + dbpath + '.')
                continue
            try:
                file = open(os.path.join(datap[0], 'namelist'+datap[1]+'.db'))
                namelist = file.readlines()
                file.close()
            except:
                self.db_list.delete(index)
                self.gamelist.reset()
                self.editDB_OK.config(state=NORMAL)
                continue
        
            self.gamelist.DBlist[i:i] = [{'name':datap, 'sgfpath':dbpath, 'current':array('I'),
                                          'encoding': self.encodingVar.get(), 'results':[], 'numNewGames':0,
                                          'data':[], 'disabled': 0}]

            db = self.gamelist.DBlist[i]
            db_date = getDateOfFile(os.path.join(db['name'][0],'namelist'+db['name'][1]+'.db'))

            self.db_list.delete(index)
            self.db_list.insert(index, dbpath + ' ('+db_date+')' )
            self.db_list.list.select_set(i)
        
            for namelistIndex in range(len(namelist)):
                g = namelist[namelistIndex][:-1] 
                filename, PB, PW, res, sig, date = split(g, '|%')
                if self.gamelist.references.has_key(sig): ref = '*'
                else: ref = ''
                self.gamelist.DBlist[i]['data'].append((filename, namelistIndex, PB, PW, res, sig, date, ref))

        self.gamelist.sortlist(self.options.sortPerDatabase.get(), self.options.sortCriterion.get())
        self.gamelist.reset()
        self.editDB_OK.config(state=NORMAL)


    def toggleDisabled(self):
        for index in self.db_list.list.curselection():
            i = int(index)
        
            db = self.gamelist.DBlist[i]

            db_date = getDateOfFile(os.path.join(db['name'][0], 'namelist'+db['name'][1]+'.db'))

            if db['disabled']:
                db['disabled'] = 0
                self.db_list.delete(i)
                self.db_list.insert(i, db['sgfpath'] + ' ('+db_date+')')
                self.db_list.list.select_set(i)
                self.db_list.list.see(i)
            else:
                db['disabled'] = 1
                self.db_list.delete(i)
                self.db_list.insert(i, 'DISABLED - ' + db['sgfpath'] + ' ('+db_date+')')
                self.db_list.list.select_set(i)
                self.db_list.list.see(i)
        self.gamelist.sortlist(self.options.sortPerDatabase.get(), self.options.sortCriterion.get())
        self.gamelist.reset()
        self.prevSearches.clear()
        self.currentSearchPattern = None
        

    def saveMessagesEditDBlist(self):
        filename = tkFileDialog.asksaveasfilename(initialdir = os.curdir)
        try:
            file = open(filename, 'w')
            file.write(self.processMessages.get('1.0', END))
            file.close()
        except IOError:
            showwarning('Error', 'Could not write to ' + filename)


    def DBlistClick(self, event):
        self.db_list.clickedLast = self.db_list.list.nearest(event.y)
        self.db_list.dragLast = -1


    def DBlistDrag(self, event):
        i = self.db_list.list.nearest(event.y)
        if self.db_list.dragLast == -1:
            if self.db_list.clickedLast == i: return
            else: self.db_list.dragLast = self.db_list.clickedLast        
        if self.db_list.dragLast != i:
            s = self.db_list.list.get(self.db_list.dragLast)
            self.db_list.delete(self.db_list.dragLast)
            self.db_list.insert(i, s)
            self.db_list.list.select_set(i)
            self.db_list.dragLast = i 
        return 'break'


    def DBlistRelease(self, event):
        if self.db_list.dragLast == -1:
            return
        
        i = self.db_list.list.nearest(event.y)

        if self.db_list.dragLast != i:
            s = self.db_list.list.get(self.db_list.dragLast)
            self.db_list.delete(self.db_list.dragLast)
            self.db_list.insert(i, s)
            self.db_list.list.select_set(i)
            self.db_list.dragLast = i

        if self.db_list.clickedLast != i:
            db = self.gamelist.DBlist[self.db_list.clickedLast]
            del self.gamelist.DBlist[self.db_list.clickedLast]
            self.gamelist.DBlist.insert(i, db)
            self.gamelist.sortlist(self.options.sortPerDatabase.get(), self.options.sortCriterion.get())
            self.gamelist.reset()
            self.prevSearches.clear()
            self.currentSearchPattern = None


    def browseDatabases(self):
        initdir = self.options.whereToStoreDatabases.get() or os.curdir
        filename = askdirectory(initialdir = initdir)
        if filename: filename = str(filename)
        self.options.whereToStoreDatabases.set(filename)
        

    def toggleWhereDatabases(self):
        if self.options.storeDatabasesSeparately.get():
            self.whereDatabasesEntry.config(state=NORMAL)
            if not self.options.whereToStoreDatabases.get(): self.browseDatabases()
        else:
            self.whereDatabasesEntry.config(state=DISABLED)

        
    def editDBlist(self):
        self.gamelist.clearGameInfo()

        self.stopAddDBVar = IntVar()
        
        window = Toplevel()
        window.transient(self.master)

        window.title('Edit database list')

        f1 = Frame(window)
        f1.grid(row=0, sticky=NSEW)
        f2 = Frame(window)
        f2.grid(row=1, sticky=NSEW)
        for i in range(4): f2.columnconfigure(i, weight=1)

        f2a = Frame(window)
        f2a.grid(row=2, sticky=NSEW)
        f3 = Frame(window)
        f3.grid(row=3, sticky=NSEW)
        f3.columnconfigure(0, weight=1)
        f4 = Frame(window)
        f4.grid(row=4, sticky=NSEW)

        window.rowconfigure(0, weight=1)
        window.rowconfigure(4, weight=2)
        window.columnconfigure(0, weight=1)
                
        self.db_list = v.ScrolledList(f1)
        self.db_list.list.config(width=60, selectmode = EXTENDED)
        self.db_list.pack(side=LEFT, expand=YES, fill=BOTH)

        self.db_list.list.bind('<1>', self.DBlistClick)
        self.db_list.list.bind('<B1-Motion>', self.DBlistDrag)
        self.db_list.list.bind('<ButtonRelease-1>', self.DBlistRelease)

        for db in self.gamelist.DBlist:
            db_date = getDateOfFile(os.path.join(db['name'][0],'namelist'+db['name'][1]+'.db'))

            if db['disabled']:
                self.db_list.insert(END, 'DISABLED - ' + db['sgfpath'] + ' (' + db_date  + ')' )
            else:
                self.db_list.insert(END, db['sgfpath'] + ' (' + db_date  + ')')

        i = 0
        for text, command in [('Add DB', lambda self=self, window=window: self.addDB(window)),
                              ('Toggle normal/disabled', self.toggleDisabled),
                              ('Remove DB', self.removeDB),
                              ('Reprocess DB', self.reprocessDB)]:
            Button(f2, text=text, command=command).grid(row=0, column=i, sticky=NSEW)
            i += 1
            
        Label(f3, text='Processing options', justify=LEFT).grid(row=0, column=0, sticky=W)

        self.filenamesVar = StringVar()
        filenamesMenu = Pmw.OptionMenu(f3, labelpos='w', label_text='Files:',
                                       menubutton_textvariable = self.filenamesVar,
                                       items = ['*.sgf', '*.sgf, *.mgt', 'All files'], initialitem = 0)
        filenamesMenu.grid(row=1, column=0, sticky=W)

        self.encodingVar = StringVar()

        enclist = ['utf-8', 'latin1', 'iso8859_2', 'iso8859_3', 'koi8_r']
        if cjkcodecs_available:
            enclist += ['gb2312', 'gbk', 'gb18030', 'hz',
                       'big5', 'cp950', 'cp932',  'shift-jis', 'shift-jisx0213',
                       'euc-jp', 'euc-jisx0213', 'iso-2022-jp', 'iso-2022-jp-1',
                       'iso-2022-jp-2', 'iso-2022-jp-3', 'iso-2022-jp-ext',
                       'cp949', 'euc-kr', 'johab', 'iso-2022-kr']

        encodingMenu = Pmw.OptionMenu(f3, labelpos='w', label_text='Encoding:',
                                      menubutton_textvariable = self.encodingVar,
                                      items = enclist, initialitem = 0)
        encodingMenu.grid(row=1, column=1, sticky=W)

        self.encoding1Var = StringVar()
        encoding1Menu = Pmw.OptionMenu(f3, labelpos='w', label_text='',
                                       menubutton_textvariable = self.encoding1Var,
                                       items = ['Do not change SGF', 'Add CA tag', 'Transcode SGF to utf-8'],
                                       initialitem = 0)
        encoding1Menu.grid(row=1, column=2, sticky=W)

        recursionButton = Checkbutton(f3, text = "Recursively add subdir's", highlightthickness=0,
                                      variable = self.options.recProcess)
        recursionButton.grid(row=2, column=0, columnspan=2, sticky=W)

        whereDatabasesButton = Checkbutton(f3, text = 'Store databases separately from SGF files', highlightthickness=0,
                                           command = self.toggleWhereDatabases, variable = self.options.storeDatabasesSeparately)
        whereDatabasesButton.grid(row=3, column=0, columnspan=2, sticky=W)

        self.whereDatabasesEntry = Entry(f3, textvariable = self.options.whereToStoreDatabases)
        self.whereDatabasesEntry.grid(row=4, column=0, columnspan=2, sticky=NSEW)
        if not self.options.storeDatabasesSeparately.get(): self.whereDatabasesEntry.config(state=DISABLED)

        browseButton = Button(f3, text ='Browse', command = self.browseDatabases)
        browseButton.grid(row=4, column=2)

        self.processMessages = Message(f4)
        self.processMessages.pack(side=TOP, expand=YES, fill=BOTH)

        self.editDB_OK = Button(f2a, text='OK', command = window.destroy)
        self.editDB_OK.pack(side=RIGHT)

        self.saveProcMess = Button(f2a, text='Save messages', command = self.saveMessagesEditDBlist)
        self.saveProcMess.pack(side=RIGHT)

        self.stopAddDBButton = Button(f2a, text='Stop processing', command = lambda self=self: self.stopAddDBVar.set(1),
                                      state=DISABLED)
        self.stopAddDBButton.pack(side=RIGHT)

        window.update_idletasks()
        window.focus()
        window.grab_set()
        window.wait_window()

        del self.db_list
        del self.processMessages
        del self.saveProcMess
        del self.stopAddDBButton
        del self.editDB_OK
        del self.whereDatabasesEntry
        
    # --------------------------------------------------------------------


    def copyCurrentGamesToFolder(self):
        dir = askdirectory(initialdir=self.datapath)
        if not dir: return
        dir = str(dir)

        if not os.path.exists(dir) and askokcancel('Error', 'Directory ' + dir + ' does not exist. Create it?'):
            try:
                os.makedirs(dir)
            except:
                return
        
        l = self.gamelist.listOfCurrentSGFFiles()
        for f in l:
            inf = open(f)
            s = inf.read()
            inf.close()

            outfile = os.path.join(dir, os.path.split(f)[-1])
            if os.path.exists(outfile): continue
            out = open(outfile, 'w')
            out.write(s)
            out.close()



    def askMaxLengthPrevSearchStack(self):
        s = askstring('Configure Back button', 'Enter maximal number of searches to remember:\n' + \
                      '(Current value: ' + `self.options.maxLengthSearchesStack.get()` + '; 0 = no limit)')
        try:
            i = int(s)
        except (TypeError, ValueError):
            pass
        else:
            self.options.maxLengthSearchesStack.set(i)    
        
    def openFile(self, path=None, filename=None, encoding = ''):
        self.board.newPosition()
        v.Viewer.openFile(self, path, filename, encoding)


    def customizeAppReset(self):
        self.options.labelFont0.set('Helvetica')
        self.options.labelFont1.set(5)
        self.options.labelFont2.set('bold')
        self.options.labelColor.set('grey50')
        self.options.statFont0.set('Helvetica')
        self.options.statFont1.set(8)
        self.options.statFont2.set('')
        self.options.statFont3.set(6)
        self.options.Btenuki.set('grey45')
        self.options.Wtenuki.set('grey77')
        self.options.progBarFont0.set('Helvetica')
        self.options.progBarFont1.set(8)
        self.options.progBarFont2.set('')
        self.options.progBarColBG.set('black')
        self.options.progBarColFG.set('white')
        self.options.exportFont0.set('Courier')
        self.options.exportFont1.set(10)
        self.options.exportFont2.set('')

        
    def customizeApp(self):
        window = Toplevel()
        window.transient(self.master)

        window.title('Customize appearance')

        Label(window, text='Labels on the board',
              font=('Helvetica', 10, 'bold'), anchor=W).pack(side=TOP, anchor=W)
        
        f1 = Frame(window)
        f1.pack(expand=YES,fill=BOTH, anchor=W)

        i = 0
        for text, var in [('Font', self.options.labelFont0),
                          ('Font size', self.options.labelFont1),
                          ('Font attribute', self.options.labelFont2),
                          ("Color of 'grey' labels", self.options.labelColor)]:
            
            Label(f1, text=text, anchor=W, justify=LEFT).grid(row=i, column=0, sticky=W)
            Entry(f1, width=15, textvariable=var).grid(row=i, column=1)
            
            i += 1

        Label(window, text='Statistics display', font=('Helvetica', 10, 'bold'),
              anchor=W, justify=LEFT).pack(side=TOP, anchor=W)

        f2 = Frame(window)
        f2.pack(expand=YES,fill=BOTH, anchor=W)

        i = 0
        for text, var in [('Font', self.options.statFont0),
                          ('Font size', self.options.statFont1),
                          ('Small font size', self.options.statFont3),
                          ('Font attribute', self.options.statFont2),
                          ('Black tenuki color', self.options.Btenuki),
                          ('White tenuki color', self.options.Wtenuki)]:
            Label(f2, text=text, anchor=W, justify=LEFT).grid(row=i, column=0, sticky=W)
            Entry(f2, width=15, textvariable=var).grid(row=i, column=1)
            
            i += 1

        Label(window, text='Progress bar', font=('Helvetica', 10, 'bold'),
              anchor=W, justify=LEFT).pack(side=TOP, anchor=W)
        
        f3 = Frame(window)
        f3.pack(expand=YES,fill=BOTH, anchor=W)

        i = 0
        for text, var in [('Font', self.options.progBarFont0),
                          ('Font size', self.options.progBarFont1),
                          ('Font attribute', self.options.progBarFont2),
                          ('Background color', self.options.progBarColBG),
                          ('Font color', self.options.progBarColFG)]:
            
            Label(f3, text=text, anchor=W, justify=LEFT).grid(row=i, column=0, sticky=W)
            Entry(f3, width=15, textvariable=var).grid(row=i, column=1)
            
            i += 1
            
        Label(window, text='Export results editor', font=('Helvetica', 10, 'bold'),
              anchor=W, justify=LEFT).pack(side=TOP, anchor=W)
        
        f4 = Frame(window)
        f4.pack(expand=YES,fill=BOTH, anchor=W)

        i = 0
        for text, var in [('Font', self.options.exportFont0),
                          ('Font size', self.options.exportFont1),
                          ('Font attribute', self.options.exportFont2)]:
            
            Label(f4, text=text, anchor=W, justify=LEFT).grid(row=i, column=0, sticky=W)
            Entry(f4, width=15, textvariable=var).grid(row=i, column=1)
            
            i += 1
            
        Button(window, text='Reset', command=self.customizeAppReset).pack(side=LEFT)

        Button(window, text='OK', command=window.destroy).pack(side=RIGHT)
        
        window.update_idletasks()
        window.focus()
        window.grab_set()
        window.wait_window()

        # check values

        for intVar, default in [(self.options.labelFont1, 5),
                                (self.options.statFont2, 8),
                                (self.options.statFont3, 6),
                                (self.options.progBarFont1, 8),
                                (self.options.exportFont1, 10)]:
            try:
                intVar.get()
            except ValueError:
                intVar.set(default)
        
        for fontVars, default, errorMesg in [((self.options.labelFont0, self.options.labelFont1, self.options.labelFont2),
                                              ('Helvetica', 5, 'bold'), 'label'),
                                             ((self.options.statFont0, self.options.statFont1, self.options.statFont2),
                                              ('Helvetica', 8, ''), 'statistics'),
                                             ((self.options.statFont0, self.options.statFont3, self.options.statFont2),
                                              ('Helvetica', 6, ''), 'small statistics'),
                                             ((self.options.progBarFont0, self.options.progBarFont1, self.options.progBarFont2),
                                              ('Helvetica', 8, ''), 'progress bar'),
                                             ((self.options.exportFont0, self.options.exportFont1, self.options.exportFont2),
                                              ('Courier', 10, ''), 'export results editor')]:
            try:
                l = Label(None, text='just a test',
                          font=(fontVars[0].get(),fontVars[1].get(),fontVars[2].get()))
            except TclError:
                for i in range(3):
                    fontVars[i].set(default[i])
                showwarning('Error', 'Invalid values for the ' + errorMesg + ' font')

        for color, default, errorMesg in [(self.options.labelColor, 'grey50', 'label'),
                                          (self.options.Btenuki, 'grey45', 'black tenuki'),
                                          (self.options.Wtenuki, 'grey77', 'white tenuki'),
                                          (self.options.progBarColBG, 'black', 'progress bar background'),
                                          (self.options.progBarColFG, 'white', 'progress bar font')]:
            try:
                l = Label(None, text='just a test', bg = color.get())
            except TclError:
                color.set(default)
                showwarning('Error', 'Invalid value for the ' + errorMesg + ' color')
        
                                          
    def quit(self):
        for i in range(len(self.filelist)):
            s = self.dataWindow.filelist.list.get(i)
            if self.options.confirmDelete.get() and s[0:2] == '* ':
                if not askokcancel('Confirm deletion', 'There are unsaved changes. Delete?'):
                    return
                else: break
        
        self.saveOptions()
        try:
            f = open(os.path.join(self.optionspath,'kombilo.def'),'w')
            f.write('kombilo05\n')
            f.write('s|%'+self.sgfpath+'\n')
            f.write('p|%'+self.datapath+'\n')
            for db in self.gamelist.DBlist:
                if db['disabled']:
                    f.write('dD|%'+ db['sgfpath'] + '|%' + db['name'][0] +\
                            '|%' + db['name'][1] + '|%' + db['encoding'] + '\n')
                else:
                    f.write('d|%' + db['sgfpath'] + '|%' + db['name'][0] +\
                            '|%' + db['name'][1] + '|%' + db['encoding']  +'\n')
            f.close()
        except IOError:
            showwarning('IOError', 'Could not write kombilo.def')

        self.master.quit()
        

    def configButtons(self, state):
        """ Disable buttons and board during search, reset them afterwards. """

        for b in [self.resetButtonS, self.backButtonS, 
                  self.searchButtonS, 
                  self.nextMove1S, self.nextMove2S, self.nextMove3S]:
            b.config(state=state)
        
        if state==NORMAL:
            self.board.state('normal', self.nextMove)
        elif state == DISABLED:
            self.board.state('disabled')


    def helpAbout(self):
        """ Display the 'About ...' window with the logo and some basic information. """

        t = []
        
        t.append('Kombilo 0.5m - written by Ulrich Goertz (u@g0ertz.de)\n\n')
        t.append('Kombilo is a go database program.\n')
        t.append('You can find more information on Kombilo and the newest ')
        t.append('version at http://www.u-go.net/kombilo/\n\n')
        
        t.append('Kombilo is free software; for more information ')
        t.append('see the file license.txt.\n\n')
        t.append('It is written in Python (see http://www.python.org/). The images of '+\
                 'the board and the stones were supplied by Patrice Fontaine.\n\n')

        t.append('I am grateful to everybody who sent/will send me comments ')
        t.append('or suggestions on Kombilo.')
        
        window = Toplevel()
        window.title('About Kombilo ...')

        if self.logo:
            canv = Canvas(window, width=75,height=23)
            canv.pack()
            canv.create_image(0,0,image=self.logo, anchor=NW)

        text = Text(window, height=15, width=60, relief=FLAT, wrap=WORD)
        text.insert(1.0, join(t, ''))
 
        text.config(state=DISABLED)
        text.pack()

        b = Button(window, text="OK", command = window.destroy)
        b.pack(side=RIGHT)
        
        window.update_idletasks()
        
        window.focus()
        window.grab_set()
        window.wait_window()


    def helpLicense(self):
        """ Display the GNU General Public License. """
        try:
            file = open(os.path.join(self.basepath, 'doc', 'license.txt'))
            t = file.read()
            file.close()
        except:
            t = 'Kombilo was written by Ulrich Goertz (u@g0ertz.de).\n' 
            t = t + 'It is published under the GNU General Public License. ' 
            t = t + 'See the file license.txt for more information. ' 
            t = t + 'The GPL is also available at http://www.gnu.org/copyleft/gpl.html\n'
            t = t + 'This program is distributed WITHOUT ANY WARRANTY!\n\n'
        self.textWindow(t,'Kombilo license')


    def saveOptions(self):
        """ Save options to disk (to file 'kombilo.opt'). """
        self.options.windowGeom = StringVar()
        self.options.windowGeom.set(self.master.geometry())
        self.options.windowGeom1 = StringVar()
        self.options.windowGeom1.set(self.searchWindow.geometry())
        self.options.dataWindowGeometry = StringVar()
        self.options.dataWindowGeometry.set(self.dataWindow.get_geometry())
        self.options.saveToDisk(os.path.join(self.optionspath,'kombilo.opt'),
                                lambda: showwarning('Save options', 'IO Error'))


    def loadOptions(self):
        """ Load options from disk. """
        self.options.windowGeom = StringVar()
        self.options.windowGeom1 = StringVar()
        self.options.dataWindowGeometry = StringVar()
        self.options.loadFromDisk(os.path.join(self.optionspath,'kombilo.opt'))
        if self.options.windowGeom.get():
            self.master.geometry(self.options.windowGeom.get())
        if self.options.windowGeom1.get():
            self.searchWindow.geometry(self.options.windowGeom1.get())
        if self.options.dataWindowGeometry.get():
            self.dataWindow.set_geometry(self.options.dataWindowGeometry.get())
        else:
            self.dataWindow.set_geometry('1|%65|%0|%65|%1|%75|%1|%30|%1|%90|%1|%80|%1|%135|%415x515+10+10')


    def showFilenameInGamelist(self):
        self.gamelist.showFilename = self.options.showFilename.get()
        self.reset()

        
    def showDateInGamelist(self):
        self.gamelist.showDate = self.options.showDate.get()
        self.reset()


    def gotoChange(self, event):
        if event.char == '': return
        if ord(event.char[0]) < 32: return

        t = self.gotoVar.get() + event.char

        i = self.gamelist.list.curselection()
        if not i: start = 0
        else: start = int(i[0])

        if self.options.sortCriterion.get() == 'PW': criterion = 3
        elif self.options.sortCriterion.get() == 'PB': criterion = 2
        elif self.options.sortCriterion.get() == 'DT': criterion = 6
        else: criterion = 0          # criterion == 'filename'

        if self.gamelist.sortCrit(start, criterion)[:len(t)] > t:
            increment = 2*self.options.sortReverse.get() - 1              # (= -1 or 1)
        else:
            increment = -2*self.options.sortReverse.get() + 1

        if increment == 1:
            if self.options.sortReverse.get():
                while start < len(self.gamelist.gameIndex)-1 and self.gamelist.sortCrit(start, criterion)[:len(t)] > t: start += 1
            else:
                while start < len(self.gamelist.gameIndex)-1 and self.gamelist.sortCrit(start, criterion) < t: start += 1
        else:
            if self.options.sortReverse.get():
                while start > 0 and self.gamelist.sortCrit(start, criterion) < t: start -= 1
            else:
                while start > 0 and self.gamelist.sortCrit(start, criterion)[:len(t)] > t: start -= 1
        
        if i: self.gamelist.list.select_clear(i)
        self.gamelist.list.select_set(start)
        self.gamelist.list.see(start)
        self.gamelist.printGameInfo(None, start)


    def rebindMouseButtons(self):
        self.board.rebindMouseButtons(self.options.onlyOneMouseButton.get())




    def initMenus(self):
        """ Initialize the menus, and a few options variables. """

        v.Viewer.initMenus(self)

        self.filemenu.insert_command(5, label='Complete reset', command = self.completeReset)


        self.options.fuzzy.set(0)
        self.options.showNextMoveVar.set(1)

        # a few option variables used in the 'edit DB list' window:

        self.options.recProcess = IntVar() # "recursively add subdir's"
        self.options.recProcess.set(1)
        
        self.options.whereToStoreDatabases = StringVar()
        self.options.storeDatabasesSeparately = IntVar()

        # ----------------------------

        self.options.labelFont0, self.options.labelFont1, self.options.labelFont2 = self.lF

        self.options.labelColor = StringVar()

        self.options.statFont0 = StringVar()
        self.options.statFont1 = IntVar()
        self.options.statFont2 = StringVar()
        self.options.statFont3 = IntVar()

        self.options.Btenuki = StringVar()
        self.options.Wtenuki = StringVar()

        self.options.progBarFont0 = StringVar()
        self.options.progBarFont1 = IntVar()
        self.options.progBarFont2 = StringVar()

        self.options.progBarColBG = StringVar()
        self.options.progBarColFG = StringVar()

        self.options.exportFont0 = StringVar()
        self.options.exportFont1 = IntVar()
        self.options.exportFont2 = StringVar()

        self.customizeAppReset() # set default values of the above options
        
        self.filemenu.insert_separator(4)
        self.filemenu.insert_command(5, label='Edit DB list', underline = 0, command=self.editDBlist)
        self.filemenu.insert_command(6, label='Export search results', command=self.exportText)
        self.filemenu.insert_command(7, label='Export current position', command=self.exportCurrentPos)
        self.filemenu.insert_command(8, label='Export SGF', command=self.exportSGF)

        self.filemenu.insert_command(9, label='Copy current SGF files to folder',
                                     command=self.copyCurrentGamesToFolder)

        self.customMenus = customMenus(self)

        self.optionsmenu.insert_command(1, label='Custom Menus', command=self.customMenus.change)


        # ------ game list submenu ------------

        gamelistMenu = Menu(self.optionsmenu)
        self.optionsmenu.add_cascade(label='Game list', underline = 0, menu = gamelistMenu)

        self.options.sortPerDatabase = IntVar()
        self.options.sortPerDatabase.set(1)
        gamelistMenu.add_checkbutton(label='Sort per database', variable = self.options.sortPerDatabase,
                                     command=self.toggleSorting)

        self.options.sortCriterion = StringVar()
        self.options.sortCriterion.set('filename')
        gamelistMenu.add_radiobutton(label='Sort by white player', variable = self.options.sortCriterion,
                                     value = 'PW', command=self.toggleSorting)
        gamelistMenu.add_radiobutton(label='Sort by black player', variable = self.options.sortCriterion,
                                     value = 'PB', command=self.toggleSorting)
        gamelistMenu.add_radiobutton(label='Sort by filename', variable = self.options.sortCriterion,
                                     value = 'filename', command=self.toggleSorting)
        gamelistMenu.add_radiobutton(label='Sort by date', variable = self.options.sortCriterion,
                                     value = 'DT', command=self.toggleSorting)
        

        self.options.sortReverse = IntVar()
        gamelistMenu.add_checkbutton(label='Reverse order', variable=self.options.sortReverse,
                                     command=self.toggleSorting)


        gamelistMenu.add_separator()

        self.options.showFilename = IntVar()
        self.options.showFilename.set(1)
        self.options.showDate = IntVar()
        gamelistMenu.add_checkbutton(label='Show filename', variable = self.options.showFilename,
                                     command = self.showFilenameInGamelist)
        gamelistMenu.add_checkbutton(label='Show date', variable = self.options.showDate,
                                     command = self.showDateInGamelist)

        # -------------------------------------

        self.options.invertSelection = IntVar()
        self.optionsmenu.add_checkbutton(label='Invert selection',
                                            variable = self.options.invertSelection)
        
        self.options.jumpToMatchVar = IntVar()
        self.options.jumpToMatchVar.set(1)
        self.optionsmenu.add_checkbutton(label='Jump to match', underline = 0,
                                            variable = self.options.jumpToMatchVar)

        self.options.smartFixedColor = self.sFC
        self.options.smartFixedColor.set(1)
        self.optionsmenu.add_checkbutton(label='Smart FixedColor', underline = 1,
                                            variable = self.options.smartFixedColor)

        self.options.uppercaseLabels = IntVar()
        self.optionsmenu.insert_checkbutton(8, label='Uppercase labels', variable=self.options.uppercaseLabels)

        self.options.showColorSwap = IntVar()
        self.options.showColorSwap.set(1)
        self.optionsmenu.add_checkbutton(label='Show color swap', variable=self.options.showColorSwap)

        advOptMenu = Menu(self.optionsmenu)
        self.optionsmenu.add_cascade(label='Advanced', underline=0, menu=advOptMenu)

        self.options.maxLengthSearchesStack = IntVar()
        self.options.maxLengthSearchesStack.set(10)
        advOptMenu.add_command(label='Number of previous searches remembered',
                               command = self.askMaxLengthPrevSearchStack)

        self.options.externalViewer = IntVar()
        advOptMenu.add_checkbutton(label='Open games in external viewer',
                                   variable = self.options.externalViewer)

        self.options.altViewerVar1 = StringVar()
        self.options.altViewerVar2 = StringVar()
        self.options.altViewerVar1.set('')
        
        advOptMenu.add_command(label='Alternative SGF viewer', underline=0,
                               command=self.altViewer)

        self.Csearch = IntVar()
        advOptMenu.add_checkbutton(label='Use C search routine', variable = self.Csearch)
        if CimportSucceeded:
            self.Csearch.set(1)
        else:
            advOptMenu.entryconfig(4, state=DISABLED)

        self.options.sloppySGF = IntVar()
        self.options.sloppySGF.set(1)
        advOptMenu.add_checkbutton(label='Sloppy SGF parsing', variable = self.options.sloppySGF)

        advOptMenu.add_separator()

        self.options.duplicateCheck = IntVar()
        self.options.duplicateCheck.set(1)
        advOptMenu.add_radiobutton(label='No duplicate check',
                                   variable = self.options.duplicateCheck, value=0)
        advOptMenu.add_radiobutton(label='Always accept duplicates',
                                   variable = self.options.duplicateCheck, value=1)
        advOptMenu.add_radiobutton(label='Ask before accepting duplicates',
                                   variable = self.options.duplicateCheck, value=2)
        advOptMenu.add_radiobutton(label='Always reject duplicates',
                                   variable = self.options.duplicateCheck, value=3)

        self.options.strictDuplicateCheck = IntVar()
        advOptMenu.add_checkbutton(label='Use strict duplicate check',
                                   variable = self.options.strictDuplicateCheck)

        advOptMenu.add_separator()

        self.options.oldTkVar = IntVar()
        advOptMenu.add_checkbutton(label='Old Tk version', underline=0,
                                   variable=self.options.oldTkVar)

        advOptMenu.add_checkbutton(label='Do not change window title',
                                   variable=self.options.doNotChangeWindowTitle)

        self.options.onlyOneMouseButton = IntVar()
        advOptMenu.add_checkbutton(label='Only one mouse button',
                                   variable=self.options.onlyOneMouseButton,
                                   command=self.rebindMouseButtons)

        self.options.askBeforeLongSearch = IntVar()
        advOptMenu.add_checkbutton(label='Ask before time consuming search',
                                   variable=self.options.askBeforeLongSearch)

        advOptMenu.add_command(label='Customize appearance', underline=0,
                               command=self.customizeApp)

        self.helpmenu.insert_command(2, label='Tutorial', underline=0, command=self.showTutorial)


    def showTutorial(self):
        try:
            webbrowser.open('file:'+os.path.abspath(os.path.join(self.basepath,'doc','tutorial.html')),new=1)
        except:
            showwarning('Error', 'Failed to open the web browser.\nYou can find the tutorial as\n'+\
                        'file:'+os.path.abspath(os.path.join(self.basepath,'doc','tutorial.html')))
        

    def balloonHelp(self):
        v.Viewer.balloonHelp(self)
        
        self.balloon.bind(self.resetButtonS, 'Reset game list')
        self.balloon.bind(self.backButtonS, 'Back to previous search pattern')
        self.balloon.bind(self.showContButtonS, 'Show continuations')
        self.balloon.bind(self.oneClickButtonS, '1-click mode')
        self.balloon.bind(self.colorButtonS, "(Don't) swap colors in search pattern")
        self.balloon.bind(self.anchorButtonS, "(Don't) translate search pattern")
        self.balloon.bind(self.nextMove1S, 'Black or white plays next (or no continuation)')
        self.balloon.bind(self.nextMove2S, 'Black plays next')
        self.balloon.bind(self.nextMove3S, 'White plays next')
        
        self.balloon.bind(self.scaleS, 'Pattern has to occur before move n (250=no limit)')
        self.balloon.bind(self.searchButtonS, 'Start pattern search')

        self.balloon.bind(self.GI_reset, 'Reset game list')
        self.balloon.bind(self.GIstart, 'Start game info search')
        self.balloon.bind(self.GIclear, 'Clear all entries')
        self.balloon.bind(self.GI_bwd, 'Restore entries of previous game info search')
        self.balloon.bind(self.GI_fwd, 'Restore entries of next game info search')
        self.balloon.bind(self.sigsearchB, 'Start signature search')

                        
    def __init__(self, master):

        # Initialization of the viewer
        # (we cannot simply call v.Viewer.__init__ because we need BoardWC instead of Board

        self.options = v.BunchTkVar()

        self.master = master
        self.frame = Frame(master)
        self.frame.pack(expand=YES, fill=BOTH)

        navFrame = Frame(self.frame)
        self.boardFrame = Frame(self.frame)
        labelFrame = Frame(self.frame)
        
        navFrame.pack(side=TOP)
        self.boardFrame.pack(side=TOP, expand=YES, fill=BOTH)
        labelFrame.pack(side=TOP)

        # The board 

        self.fixedColorVar = IntVar()
        self.sFC = IntVar()
        self.lF = (StringVar(), IntVar(), StringVar())

        if sys.path[0].endswith('library.zip'): SYSPATH = os.path.split(sys.path[0])[0]
        else: SYSPATH = sys.path[0]

        gifpath = os.path.join(SYSPATH,'gifs')
        try:
            self.boardImg = PhotoImage(file=os.path.join(gifpath, 'board.gif'))
        except (TclError, IOError):
            self.boardImg = None
        try:
            self.blackStone = Image.open(os.path.join(gifpath, 'black.gif')).convert('RGBA')
            self.whiteStone = Image.open(os.path.join(gifpath, 'white.gif')).convert('RGBA')
        except (TclError, IOError, AttributeError):
            self.blackStone = None
            self.whiteStone = None
            
        self.board = BoardWC(self.boardFrame, 19, (5, 22), 1, self.lF, self.fixedColorVar, self.sFC,
                             self.boardImg, self.blackStone, self.whiteStone)
        self.board.bind('<Control-1>', self.labelClick)
        self.board.bind('<Double-1>', self.doubleClick)
        self.board.bind('<Shift-3>', self.gotoMove)
        
        self.board.state('normal', self.nextMove)
        self.board.shadedStoneVar.set(1)
      
        self.initButtons(navFrame, labelFrame)
        self.initMenus()
        self.moveno.set(0)
        self.capB, self.capW = 0, 0
        self.board.invertSelection = self.options.invertSelection
        
        self.datapath = self.basepath

        self.board.pack(expand=YES, fill=BOTH)
        self.board.update_idletasks()

        # the data window

        self.dataWindow = DataWindow(self)
        self.comments = self.dataWindow.comments

        self.cursor = v.EnhancedCursor('(;GM[1]FF[4]SZ[19]AP[Kombilo])', self.comments, self.dataWindow.SGFtreeC, self)

        self.dataWindow.updateGameInfo(self.cursor)
        self.dataWindow.window.update()

        self.filelist = []
        self.currentFileNum = -2
        self.newFile()

        # ------------ the search window

        self.searchWindow = Toplevel()
        # self.searchWindow.withdraw()
        self.searchWindow.geometry('400x550')
        self.searchWindow.protocol('WM_DELETE_WINDOW', lambda: 0)
        self.searchWindow.title('Kombilo: game list')

        self.frameS = Frame(self.searchWindow)   # suffix S means 'in search/results window'
        self.frameS.pack(fill=BOTH, expand=YES)

        self.topFrameS = Frame(self.frameS)
        self.listFrameS = Frame(self.frameS)
        self.notebookFrameS = Frame(self.frameS, height=400)
        self.notebook = Pmw.NoteBook(self.notebookFrameS, borderwidth=2, pagemargin=0)
        
        self.bottomFrameS = self.notebook.add('Pattern search')
        self.giSFS = self.notebook.add('Game info search')
        self.gameinfoSearchFS = Frame(self.giSFS)
        self.dummyGISFS = Frame(self.giSFS)

        self.gameinfoSearchFS.pack(side=LEFT)
        self.dummyGISFS.pack(expand=YES, fill=BOTH)
                
        self.buttonFrameS = Frame(self.bottomFrameS, height=200)
        self.emptyC = Canvas(self.bottomFrameS, height=200, width=10, highlightthickness=0)
        self.statisticsFrameS = Frame(self.bottomFrameS, height=200)
        
        self.notebook.pack(fill=BOTH, expand=YES, pady=0)

        noGamesLabel = Label(self.topFrameS, text=' ', width=18, height=1)
        noGamesLabel.grid(row=0, column=0)

        winPercLabel = Label(self.topFrameS, text=' ', width=18, height=1)
        winPercLabel.grid(row=0, column=1)

        self.gameinfoS = v.ScrolledText_HSB(self.frameS, height = 4, width=40, wrap=WORD, 
                                            relief=SUNKEN, state = DISABLED) # , font=('Arial', 11))
        self.gameinfoS.tag_config('blue', foreground='blue')

        self.gamelist = GameList(self.listFrameS, self, noGamesLabel,
                                 winPercLabel, self.gameinfoS)
        self.gamelist.pack(expand=YES, fill=BOTH, side=TOP)

        # ------------ read the kombilo.def file (default databases etc.)
        # (do this as soon as possible, because setting the geometry properties of
        # the searchWindow (in loadOptions) under Windows doesn't work properly otherwise

        try:
            f = open(os.path.join(self.basepath,'kombilo.def'))
            s = split(f.read(),'\n')
            f.close()
        except:
            pass
        else:
            if s[0]=='kombilo05':      # otherwise this is an old .def file which should be ignored
                counter = 1
                while 1:
                    try:
                        line = split(s[counter], '|%')

                        if line and line[0] == 'i':            # refer to individual def-file
                            l = line[1]
                            self.optionspath = os.path.join(os.getenv('HOME',l),'.kombilo')
                       
                            if not os.path.exists(self.optionspath):
                                os.mkdir(self.optionspath)
                            try:
                                f1 = open(os.path.join(self.optionspath, 'kombilo.def'))
                                s1 = split(f1.read(), '\n')
                                for l1 in s1: s.append(l1)
                            except IOError:
                                pass
                        elif line and line[0] == 's':          # new sgfpath
                            self.sgfpath = line[1]
                        elif line and line[0] == 'd':
                            if len(line)==4:
                                enc = 'utf-8'
                            else:
                                enc = line[4]
                            self.gamelist.DBlist.append({'sgfpath':line[1], 'name':(line[2], line[3]), 'data':[],
                                                         'encoding': enc, 'results':[], 'current':array('I'),
                                                         'numNewGames':0, 'disabled': 0})
                    
                        elif line and line[0] == 'dD':
                            if len(line)==3:
                                enc = 'utf-8'
                            else:
                                enc = line[4]
                            self.gamelist.DBlist.append({'sgfpath':line[1], 'name':(line[2], line[3]), 'data':[],
                                                         'encoding': enc, 'results':[], 'current':array('I'),
                                                         'numNewGames':0, 'disabled': 1})
                            
                        elif line and line[0] == 'p':          # new datapath
                            self.datapath = line[1]
                    except: pass
                    
                    if counter >= len(s)-1: break
                    counter += 1

        # --------------------------------

        self.customMenus.path = self.optionspath
        self.customMenus.buildMenus(1, self.basepath)

        self.options.caseInsensitive = IntVar()
        
        self.loadOptions()
        self.toggleDatawindow()
        self.master.deiconify()
        self.searchWindow.deiconify()

        if self.options.showCoordinates.get():
            self.board.coordinates = 1
            self.board.resize()

        # -------------------------

        # self.notebookFrameS.grid(row=4, sticky=NSEW)

        self.buttonFrameS.pack(side=LEFT)
        self.emptyC.create_line(5,0,5,200, fill='black', width=2)
        self.emptyC.pack(side=LEFT)
        self.statisticsFrameS.pack(side=LEFT, expand=YES, fill=BOTH)
        
        self.statisticsCanv = Canvas(self.statisticsFrameS, width=244, height=200, highlightthickness=0)
        self.statisticsCanv.pack(side=BOTTOM, expand=YES, fill=BOTH)

        try:
            im = PhotoImage(file=os.path.join(self.basepath, 'gifs', 'stat_bg.gif'))
            self.tkImages.append(im)
            self.statisticsCanv.create_image(0,0,anchor=NW, image=im)
        except TclError:
            pass

        self.statistics = {}

        self.buttonFrame1S = Frame(self.buttonFrameS)
        self.buttonFrame1S.grid(row=0, column=0, columnspan=4)

        self.resetButtonS = Button(self.buttonFrame1S, text = 'Reset', command = self.reset)
        self.resetButtonS.grid(row=0, column=1)

        self.backButtonS = Button(self.buttonFrame1S, text = 'Back', command = self.back)
        self.backButtonS.grid(row=0, column=0)

        self.showContinuation = IntVar()
        self.showContButtonS = Checkbutton(self.buttonFrame1S, text = 'Cont',
                                          variable = self.showContinuation, 
                                          indicatoron=0, command=self.showCont)

        self.oneClick = IntVar()
        self.oneClickButtonS = Checkbutton(self.buttonFrame1S, text = '1 click',
                                           variable = self.oneClick,
                                           indicatoron=0)

        # pack everything
        self.oneClickButtonS.grid(row=0, column=3)
        self.showContButtonS.grid(row=0, column=2)

        self.colorButtonS = Checkbutton(self.buttonFrameS, text='Fixed Color', highlightthickness=0,
                                        variable = self.fixedColorVar)
        self.colorButtonS.grid(row=1, column=0, columnspan=4)

        self.fixedAnchorVar = IntVar()
        self.anchorButtonS = Checkbutton(self.buttonFrameS, text='Fixed Anchor', highlightthickness=0,
                                        variable = self.fixedAnchorVar)
        self.anchorButtonS.grid(row=2, column=0, columnspan=4)

        l = Label(self.buttonFrameS, text='Next:')
        l.grid(row=4, column=0)

        self.nextMoveVar = IntVar() # 0 = either player, 1 = black, 2 = white
        self.nextMove1S = Radiobutton(self.buttonFrameS, text='B/W', highlightthickness=0, indicatoron=0,
                                      variable = self.nextMoveVar, value=0)
        self.nextMove1S.grid(row=4, column=1)
        self.nextMove2S = Radiobutton(self.buttonFrameS, text='B', highlightthickness=0, indicatoron=0,
                                      variable = self.nextMoveVar, value=1)
        self.nextMove2S.grid(row=4, column=2)
        self.nextMove3S = Radiobutton(self.buttonFrameS, text='W', highlightthickness=0, indicatoron=0,
                        variable = self.nextMoveVar, value=2)
        self.nextMove3S.grid(row=4, column=3)


        self.mvLimLabel = Label(self.buttonFrameS, text='MvLim')
        self.mvLimLabel.grid(row=5, column=0)
        self.moveLimit = IntVar()
        self.moveLimit.set(250)
        self.scaleS = Scale(self.buttonFrameS, highlightthickness=0,
                            length=80, variable = self.moveLimit, from_=1, to=250,
                            tickinterval = 149, showvalue=YES, orient='horizontal')
        self.scaleS.grid(row=5, column=1, columnspan=3)
        
        self.searchButtonS = Button(self.buttonFrameS, text = 'Pattern ?', command = self.search)
        self.searchButtonS.grid(row=6, column=0)

        self.progBar = ProgressBar(self.frameS, 390,
                                   (self.options.progBarFont0, self.options.progBarFont1, self.options.progBarFont2),
                                   (self.options.progBarColBG, self.options.progBarColFG))

        self.topFrameS.pack(side=TOP, fill=X, expand=NO) 
        self.listFrameS.pack(side=TOP, fill=BOTH, expand=YES) 
        self.gameinfoS.pack(side=TOP, fill=X, expand=NO) 
        self.notebookFrameS.pack(side=TOP, fill=X, expand=NO) 
        self.progBar.pack(side=TOP, fill=X, expand=NO)

        # game info search frame

        f1 = Frame(self.gameinfoSearchFS)
        f1.pack()
        f3 = Frame(self.gameinfoSearchFS)
        f3.pack(expand=YES, fill=X)
        
        self.pbVar = StringVar()
        self.pwVar = StringVar()
        self.pVar = StringVar()
        self.evVar = StringVar()
        self.frVar = StringVar()
        self.toVar = StringVar()
        self.awVar = StringVar()
        
        l1 = Label(f1, text='White', anchor=W)
        e1 = Entry(f1, width=16, textvariable=self.pwVar)
        l2 = Label(f1, text='Black', anchor=W)
        e2 = Entry(f1, width=16, textvariable=self.pbVar)
        l3 = Label(f1, text='Player', anchor=W)
        e3 = Entry(f1, width=33, textvariable=self.pVar)
        l4 = Label(f1, text='Event', anchor=W)
        e4 = Entry(f1, width=33, textvariable=self.evVar)
        l5 = Label(f1, text='From', anchor=W)
        e5 = Entry(f1, width=16, textvariable=self.frVar)
        l6 = Label(f1, text='To', anchor=W)
        e6 = Entry(f1, width=16, textvariable=self.toVar)

        l7 = Label(f1, text='Anywhere', anchor=W)
        e7 = Entry(f1, width=33, textvariable=self.awVar)

        self.referencedVar = IntVar()
        b1 = Checkbutton(f3, text='Referenced', variable = self.referencedVar, highlightthickness=0)
        
        b2 = Checkbutton(f3, text='Case insensitive', variable=self.options.caseInsensitive, highlightthickness=0)
        self.GIstart = Button(f3, text='Start',
                              command = self.doGISearch)
        self.GIclear = Button(f3, text='Clear', command = self.clearGI)

        self.GI_bwd = Button(f3, text = '<-', command = self.historyGI_back)
        self.GI_fwd = Button(f3, text = '->', command = self.historyGI_fwd)

        for e in [e1, e2, e3, e4, e5, e6, e7]:
            e.bind('<Return>', lambda event, bs = self.GIstart: bs.invoke())
        
        l1.grid(row=0, column=0, sticky=E)
        e1.grid(row=0, column=1, sticky=NSEW)
        l2.grid(row=0, column=2, sticky=E)
        e2.grid(row=0, column=3, sticky=NSEW)
        l3.grid(row=2, column=0, sticky=E)
        e3.grid(row=2, column=1, columnspan = 3, sticky=NSEW)
        l4.grid(row=3, column=0,  sticky=E)
        e4.grid(row=3, column=1, columnspan = 3, sticky=NSEW)
        l5.grid(row=4, column=0, sticky=E)
        e5.grid(row=4, column=1, sticky=NSEW)
        l6.grid(row=4, column=2, sticky=E)
        e6.grid(row=4, column=3, sticky=NSEW)

        l7.grid(row=5, column=0, sticky=E)
        e7.grid(row=5, column=1, columnspan=3, sticky=NSEW)
        
        b1.pack(side=LEFT, padx=10) 
        b2.pack(side=LEFT) 
        self.GI_fwd.pack(side=RIGHT)
        self.GI_bwd.pack(side=RIGHT)
        self.GIstart.pack(side=RIGHT)
        self.GIclear.pack(side=RIGHT)

        f4 = Canvas(self.gameinfoSearchFS, height=3, width=390, highlightthickness=0)
        f4.pack(expand=NO)
        f4.create_line(1,1, 400,1)
        
        f5 = Frame(self.gameinfoSearchFS)
        f5.pack(side=LEFT, expand=YES, fill=X)

        self.GI_reset = Button(f5, text='Reset', command = self.reset)
        self.GI_reset.pack(side=LEFT, padx=10)
        l1 = Label(f5, text='Go to:', anchor=W)
        l1.pack(side=LEFT)
        self.gotoVar = StringVar()
        e1 = Entry(f5, width=20, textvariable=self.gotoVar)
        e1.pack(side=LEFT)
        e1.bind('<Key>', self.gotoChange)
        e1.bind('<Up>', self.gamelist.up)
        e1.bind('<Down>', self.gamelist.down)
        e1.bind('<Prior>', self.gamelist.pgup)
        e1.bind('<Next>', self.gamelist.pgdown)

        self.sigsearchB = Button(f5, text='Signature search', command = self.sigSearch)
        self.sigsearchB.pack(side=RIGHT)

        self.history_GIsearch = []
        self.history_GIS_index = -1

        # icons for the buttons

        for button, filename in [(self.showContButtonS, 'ab'), (self.backButtonS, 'bwd'),
                                 (self.resetButtonS, 'reset'), (self.searchButtonS, 'psearch'),
                                 (self.oneClickButtonS, 'oneclick'),
                                 (self.nextMove1S, 'bw'), (self.nextMove2S, 'b'), (self.nextMove3S, 'w'),
                                 (self.GI_reset, 'reset'), (self.sigsearchB, 'ssearch'), (self.GIstart, 'isearch'),
                                 (self.GIclear, 'clear'),
                                 (self.GI_bwd, 'bwd'), (self.GI_fwd, 'fwd')]:
            try:
                im = PhotoImage(file=os.path.join(self.basepath, 'gifs', filename + '.gif'))
                self.tkImages.append(im)
                button.config(image=im)
            except TclError:
                pass

        # logo

        emptyFrame = Frame(self.topFrameS, width=50)
        emptyFrame.grid(row=0, column=2)
        self.topFrameS.columnconfigure(2, weight=1)

        self.logoCanv = Canvas(self.topFrameS, height=23, width=75, highlightthickness=0)
        self.logoCanv.grid(column=3, row=0, sticky=E)

        try:
            self.logo = PhotoImage(file=os.path.join(self.basepath,'gifs/logok.gif'))
            self.logoCanv.create_image(0,0,
                                    image=self.logo, tags='logo', anchor=NW)
        except TclError:
            self.logo = None

        if self.options.smartFixedColor.get():
            self.fixedColorVar.set(1)

        self.board.update()
        self.notebook.setnaturalsize()
        self.searchWindow.update()
        
        self.showFilenameInGamelist()
        self.showDateInGamelist()
        self.loadDBs()
        self.continuations = {}
        self.prevSearches = PrevSearchesStack(self.options.maxLengthSearchesStack, self.board.changed,
                                              self.dataWindow.prevSF, self)
        self.board.callOnChange = self.prevSearches.select_clear
        self.currentSearchPattern = None
        self.balloonHelp()


# ---------------------------------------------------------------------------------------

root = Tk()
root.withdraw()

if sys.path[0].endswith('library.zip'): SYSPATH = os.path.split(sys.path[0])[0]
else: SYSPATH = sys.path[0]

try:
    root.option_readfile(os.path.join(SYSPATH, 'kombilo.app'))
except TclError:
    showwarning('Error', 'Error reading kombilo.app')
    
app = App(root)

root.protocol('WM_DELETE_WINDOW', app.quit)
root.title('Kombilo: board')

app.boardFrame.focus_force()

for filename in sys.argv[1:]:              # load sgf files given as arguments
    app.openFile(os.path.split(filename)[0], os.path.split(filename)[1])

root.mainloop()
root.destroy()



